/*
 * assoc_ctrl.c - implements association control logic and util functions
 *
 * Copyright (C) 2025 IOPSYS Software Solutions AB. All rights reserved.
 *
 */
#include <easymesh.h>

#include "agent.h"
#include "agent_map.h"
#include "agent_cmdu.h"
#include "agent_tlv.h"
#include "assoc_ctrl.h"
#include "config.h"
#include "wifi.h"

#include "utils/debug.h"
#include "utils/utils.h"

static void assoc_ctrl_update_config(struct agent *a, uint8_t *mac,
		struct netif_apcfg *cfg, int add)
{
	char stastr[18] = {0};

	hwaddr_ntoa(mac, stastr);
	config_update2("mapagent", &a->cfg, "ap",
		"ifname", cfg->name,
		"assoc_ctrl", add, stastr, 18);

	add ? stax_add_entry(&cfg->assoc_ctrllist, mac)
		: stax_del_entry(&cfg->assoc_ctrllist, mac);
}

static void agent_disconnect_stas(struct agent *a, int num_sta,
		uint8_t sta_macs[][6], struct netif_ap *ap)
{
	trace("agent: %s: --->\n", __func__);

	int i;

	if (!sta_macs)
		return;

	for (i = 0; i < num_sta; i++) {
		struct netif_ap *s_ap = agent_get_ap_with_sta(a, sta_macs[i]);

		if (!s_ap || s_ap != ap)
			continue;

		dbg("%s: Disconnect STA: " MACFMT " from %s\n",
		    __func__, MAC2STR(sta_macs[i]), ap->ifname);
		wifi_disconnect_sta(ap->ifname, sta_macs[i],
				WIFI_REASON_AP_INITIATED);
	}
}

static int assoc_ctrl_restrict_sta(struct agent *a, uint32_t num_sta,
		uint8_t sta_list[][6], struct netif_ap *ap, uint8_t mode)
{
	trace("agent: %s: --->\n", __func__);

	int ret = 0;
	int i;

	if (!sta_list || !num_sta)
		return -1;

	for (i = 0; i < num_sta; i++)
		ret |= WARN_ON(wifi_restrict_sta(ap->ifname, ap->band,
					sta_list[i], mode != ASSOC_CTRL_UNBLOCK));

	if (mode == ASSOC_CTRL_TIMED_BLOCK || mode == ASSOC_CTRL_INDEF_BLOCK)
		/* Disassociate any associated STA(s) specified in the req */
		agent_disconnect_stas(a, num_sta, sta_list, ap);

	return ret;
}

int assoc_ctrl_reapply(struct netif_ap *ap)
{
	trace("agent: %s: --->\n", __func__);

	uint8_t client_sta[ASSOC_CTRL_MAX_STA][6] = {0};
	struct restrict_sta_entry *s = NULL;
	int ret = 0;
	int i, num_sta = 0;

	list_for_each_entry(s, &ap->restrict_stalist, list) {
		memcpy(client_sta[num_sta++], s->macaddr, 6);
	}

	for (i = 0; i < num_sta; i++)
		ret |= wifi_restrict_sta(ap->ifname, ap->band,
					client_sta[i], s->mode != ASSOC_CTRL_UNBLOCK);

	return ret;
}

static void wifi_sta_restrict_timeout_cb(atimer_t *t)
{
	trace("agent: %s: --->\n", __func__);

	struct restrict_sta_entry *s =
		container_of(t, struct restrict_sta_entry, restrict_timer);
	struct netif_ap *ap;

	ap = agent_get_ap(s->agent, s->bssid);
	if (!ap) {
		trace("[%s:%d] Error BSSID not present", __func__, __LINE__);
		return;
	}

	/* Unblock STA */
	wifi_restrict_sta(ap->ifname, ap->band, s->macaddr, false);
	list_del(&s->list);
	free(s);
}

static struct restrict_sta_entry *find_restricted_sta(struct netif_ap *ap, uint8_t *mac)
{
	struct restrict_sta_entry *s = NULL;

	list_for_each_entry(s, &ap->restrict_stalist, list) {
		if (!memcmp(s->macaddr, mac, 6))
			return s;
	}
	return NULL;
}

int assoc_ctrl_add_sta(struct agent *a, struct netif_ap *ap,
		uint32_t num_sta, uint8_t sta_list[][6],
		uint16_t validity_period, uint8_t mode,
		bool update_config)
{
	trace("agent: %s: --->\n", __func__);

	int i = 0;

	if (!sta_list)
		return -1;

	for (i = 0; i < num_sta; i++) {
		struct restrict_sta_entry *s;

		s = find_restricted_sta(ap, sta_list[i]);
		if (!s) {
			/* Add new entry to restricted list */
			s = calloc(1, sizeof(struct restrict_sta_entry));
			if (!s)
				return -1;

			memcpy(s->macaddr, sta_list[i], 6);
			memcpy(s->bssid, ap->bssid, 6);
			s->agent = a;
			s->mode = mode;
			list_add_tail(&s->list, &ap->restrict_stalist);

			timer_init(&s->restrict_timer, wifi_sta_restrict_timeout_cb);
		}

		if (validity_period > 0)
			/* If timeout is positive, create a timed sta assoc-control
			 * entry but don't update config. At the timer expiry, remove
			 * sta from assoc-control list.
			 */
			timer_set(&s->restrict_timer, validity_period * 1000);
		else if (update_config)
			/* If timeout is 0 or negative, permanently add sta to the
			 * AP's assoc control list by creating a config entry.
			 */
			assoc_ctrl_update_config(a, sta_list[i], ap->cfg, 1);
	}

	return 0;
}

int assoc_ctrl_block_sta(struct agent *a, struct netif_ap *ap,
		uint16_t validity_period, uint8_t mode,
		uint32_t num_sta, uint8_t sta_list[][6])
{
	trace("agent: %s: --->\n", __func__);
	int ret = 0;

	/* Add STA to restricted list */
	ret = assoc_ctrl_add_sta(a, ap, num_sta, sta_list,
			validity_period, mode, true);
	if (!ret)
		/* Block STA in WiFi */
		ret = assoc_ctrl_restrict_sta(a, num_sta, sta_list, ap, mode);

	return ret;
}

int assoc_ctrl_del_sta(struct agent *a, struct netif_ap *ap,
		uint32_t num_sta, uint8_t sta_list[][6], bool update_config)
{
	trace("agent: %s: --->\n", __func__);

	struct restrict_sta_entry *s = NULL, *tmp = NULL;
	int i = 0;

	if (sta_list == NULL)
		return -1;

	for (i = 0; i < num_sta; i++) {
		/* Stop validity timer and remove entry from restricted list */
		list_for_each_entry_safe(s, tmp, &ap->restrict_stalist, list) {
			if (!memcmp(s->macaddr, sta_list[i], 6)) {
				if (update_config)
					/* Update config to remove the STA */
					assoc_ctrl_update_config(a, sta_list[i], ap->cfg, 0);

				timer_del(&s->restrict_timer);
				list_del(&s->list);
				free(s);
			}
		}
	}

	return 0;
}

int assoc_ctrl_unblock_sta(struct agent *a, struct netif_ap *ap,
		uint32_t num_sta, uint8_t sta_list[][6])
{
	trace("agent: %s: --->\n", __func__);
	int ret = 0;

	/* Remove STA from restricted list */
	ret = assoc_ctrl_del_sta(a, ap, num_sta, sta_list, true);
	if (!ret)
		/* Unblock STA in WiFi */
		ret = assoc_ctrl_restrict_sta(a, num_sta, sta_list, ap, ASSOC_CTRL_UNBLOCK);

	return ret;
}

int assoc_ctrl_sync_from_config(struct agent *a, struct netif_ap *ap)
{
	trace("agent: %s: --->\n", __func__);

	struct stax *x = NULL;
	struct restrict_sta_entry *s = NULL;
	uint8_t stalist[MAX_STA][6] = {0};
	int num_sta = 0;
	int ret = 0;

	/* 1. Add STAs missing in runtime list */
	list_for_each_entry(x, &ap->cfg->assoc_ctrllist, list) {
		memcpy(stalist[num_sta++], x->macaddr, 6);
	}

	if (num_sta > 0)
		/* Update the runtime list and restart timer */
		ret |= assoc_ctrl_add_sta(a, ap, num_sta, stalist,
				0 /* FIXME tmo from cfg */,
				ASSOC_CTRL_INDEF_BLOCK /* FIXME mode from cfg */,
				false);

	/* 2. Remove STAs that are not in config list */
	list_for_each_entry(s, &ap->restrict_stalist, list) {
		bool found = false;
		int i = 0;

		for (i = 0; i < num_sta; i++) {
			if (!memcmp(stalist[i], x->macaddr, 6)) {
				found = true;
				break;
			}
		}

		if (!found) {
			/* Remove STA from runtime list */
			ret |= assoc_ctrl_del_sta(a, ap, 1, &x->macaddr, false);
		}
	}

	return ret;
}

/* Blocks all stas from the ap->restrict_stalist in WiFi */
int assoc_ctrl_apply_restriction(struct agent *a, struct netif_ap *ap)
{
	trace("agent: %s: --->\n", __func__);

	uint8_t wifi_res_stas[ASSOC_CTRL_MAX_STA * 6] = {0};
	int num_wifi_res_stas = ASSOC_CTRL_MAX_STA;
	struct restrict_sta_entry *s = NULL;
	int i;
	int ret = 0;

	if (list_empty(&ap->restrict_stalist))
		return 0;

	/* Get the list of STAs currently blocked in WiFi */
	ret = wifi_get_restricted_stas(ap->ifname, ap->band, wifi_res_stas, &num_wifi_res_stas);
	if (ret) {
		err("[%s:%d] Failed to get restricted STAs\n", __func__, __LINE__);
		return -1;
	}

	/* Invoke block_sta for STAs that are in runtime list but not in blocked_stas */
	list_for_each_entry(s, &ap->restrict_stalist, list) {

		for (i = 0; i < num_wifi_res_stas; i++) {

			if (!memcmp(&wifi_res_stas[i * 6], s->macaddr, 6))
				/* STA already blocked in WiFi */
				continue;

			assoc_ctrl_restrict_sta(a, 1, &s->macaddr, ap, s->mode);
		}
	}

	/* Theoretically, it could be blocked for other reasons, i.e. network admin-issued block
	 * manually or via webGUI. Uncomment this part if we want to unblock STAs anyway.
	 */
#if 0
	/* Invoke unblock_sta for STAs that are in blocked_stas but not in runtime list */
	for (i = 0; i < num_wifi_res_stas; i++) {
		bool found = false;

		list_for_each_entry(s, &ap->restrict_stalist, list) {
			if (!memcmp(&wifi_res_stas[i * 6], s->macaddr, 6)) {
				found = true;
				break;
			}
		}

		if (!found) {
			/* STA not in runtime list, unblock it */
			uint8_t mac[6];

			memcpy(mac, &wifi_res_stas[i * 6], 6);
			assoc_ctrl_restrict_sta(a, 1, &mac, ap, ASSOC_CTRL_UNBLOCK);
		}
	}
#endif

	return 0;
}

int assoc_ctrl_process_request(void *agent, uint8_t *p,
				 struct cmdu_buff *cmdu)
{
	trace("agent: %s: --->\n", __func__);

	struct tlv_client_assoc_ctrl_request *data =
		(struct tlv_client_assoc_ctrl_request *)p;
	struct agent *a = (struct agent *) agent;
	int num_err = 0;
	struct netif_ap *ap;
	struct sta_error_response sta_resp[MAX_STA];
	int offset = 0;
	uint8_t sta_list[ASSOC_CTRL_MAX_STA][6];
	uint16_t validity_period = 0;
	int ret = 0;
	int i;

	if (!data->num_sta || data->num_sta > ASSOC_CTRL_MAX_STA) {
		err("[%s:%d] Invalid number of STAs\n", __func__, __LINE__);
		ret = -1;
		goto send_ack;
	}

	offset = sizeof(*data);
	for (i = 0; i < data->num_sta; i++) {
		memcpy(sta_list[i], &p[offset], 6);
		offset += 6;
	}

	ap = agent_get_ap(a, data->bssid);
	if (!ap) {
		err("[%s:%d] BSSID not present\n", __func__, __LINE__);
		ret = -1;
		goto send_ack;
	}

	/* Special case for bBSS: If bssid is backhaul BSS and STA addr
	 * is ff:ff:ff:ff:ff:ff block/unblock associations for all macs.
	 */
	if ((ap->cfg && ap->cfg->multi_ap & 0x01)
		 && data->num_sta == 1
		 && hwaddr_is_bcast(data->sta[0].macaddr)) {
		bool disallow = false;

		if (data->control == ASSOC_CTRL_BLOCK)
			disallow = true;

		trace("%s: %s assoc. for bBSS: " MACFMT " [%s]\n",
			  __func__, disallow ? "Disallow" : "Allow",
			  MAC2STR(data->bssid), ap->ifname);

		if (!wifi_set_backhaul_bss_association_mode(ap->ifname, disallow)) {
			ap->cfg->disallow_assoc = disallow;
			uci_apply_bss_association_mode(ap->ifname, disallow);
		}

		goto send_ack;
	}

	switch (data->control) {
	case ASSOC_CTRL_BLOCK:
		/* Block - check for an Error Scenario:
		 * Send an error TLV in case STA is already associated
		 * with the BSSID for which the 'Block' mode is being set.
		 */
		for (i = 0; i < data->num_sta; i++) {
			struct netif_ap *s_ap = agent_get_ap_with_sta(a, sta_list[i]);

			if (s_ap == ap) {
				dbg("STA already associated with BSSID!\n");
				/* 11.6: "include an Error Code TLV with the
				 * reason set to 0x01 and the STA MAC address
				 * field included"
				 */
				memcpy(sta_resp[num_err].sta_mac, sta_list[i], 6);
				sta_resp[num_err].response = 0x01;
				num_err++;
			}
		}

		if (num_err == data->num_sta) {
			/* All STAs are already associated with the BSSID */
			warn("All STAs already associated with given BSSID!\n");
			ret = -1;
			goto send_ack;
		}
		/* no break - checkpatch ignore */
	case ASSOC_CTRL_TIMED_BLOCK:
		/* tmo needed for block and timed block modes */
		validity_period = BUF_GET_BE16(data->validity_period);
		if (!validity_period) {
			warn("Validity period unset, while expected!\n");
			ret = -1;
			goto send_ack;
		}
		/* no break - checkpatch ignore */
	case ASSOC_CTRL_INDEF_BLOCK:
		/* Add STA to blocking list for block/timed block/indef block */
		WARN_ON(assoc_ctrl_block_sta(a, ap, validity_period, data->control,
					data->num_sta, sta_list));
		break;
	case ASSOC_CTRL_UNBLOCK:
		/* Rem STA from blocking list and stop the restrict timer */
		WARN_ON(assoc_ctrl_unblock_sta(a, ap, data->num_sta, sta_list));
		break;
	default:
		warn("Reserved Association Control mode!\n");
		ret = -1;
		goto send_ack;
	}

send_ack:
	/* If a Multi-AP Agent receives a Client Association Control
	 * Request message, then it shall respond within one second
	 * with a 1905 ACK message.
	 */
	ret = send_1905_acknowledge(agent, cmdu->origin, cmdu_get_mid(cmdu),
							    sta_resp, num_err);
	return ret;
}
