/*
 * agent.c - Multi-AP easymesh agent
 *
 * Copyright (C) 2019-2023 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for licensing information.
 *
 */
#include "agent.h"

#include <1905_tlvs.h>
#include <cmdu.h>
#if (EASYMESH_VERSION >= 3)
#ifdef USE_LIBDPP
#include <dpp_api.h>
#endif /* USE_LIBDPP */
#endif
#include <easy/timestamp.h>
#include <easy/utils.h>
#include <easymesh.h>
#include <errno.h>
#include <i1905_wsc.h>
#include <inttypes.h>
#include <json-c/json_object.h>
#include <json-c/json_tokener.h>
#include <json-c/json_types.h>
#include <libubox/blob.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <map_module.h>
#include <net/if.h>
#include <netlink/socket.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ubusmsg.h>
#include <uci.h>
#include <unistd.h>
#include <wifiutils.h>
#include <timer.h>

#include "agent_cmdu.h"
#include "agent_map.h"
#include "agent_tlv.h"
#include "agent_ubus.h"
#include "agent_ubus_dbg.h"
#include "assoc_ctrl.h"
#include "backhaul.h"
#include "backhaul_blacklist.h"
#include "config.h"
#if (EASYMESH_VERSION >= 3)
#ifdef USE_LIBDPP
#include "dpp.h"
#include "dpphandler.h"
#endif /* USE_LIBDPP */
#include "qos.h"
#endif
#include "extension.h"
#include "nl.h"
#include "steer_rules.h"
#include "timer_impl.h"
#include "utils/1905_ubus.h"
#include "utils/debug.h"
#include "utils/liblist.h"
#include "utils/utils.h"
#include "wifi.h"
#include "wifi_caps.h"

struct json_object;

#define map_plugin	"ieee1905.map"

#define IFACE_TIMEOUT 1
#define BOOT_UP_SCAN_TIME	(0 * 1000)
#define BOOT_UP_SCAN_ITV	(2 * 1000)
#define BOOT_UP_SCAN_MAX_TRY 5

static struct agent *this_agent;

static int agent_subscribe_for_cmdus(struct agent *a);
static void agent_free_nodes(struct agent *a);
static void agent_free_wifi_assoc_frames(struct agent *a);
static int wifi_topology_notification(struct agent *agent, struct sta *s,
		char *ifname, uint8_t assoc_event, uint16_t reason);
static int wifi_del_sta(struct agent *a, const char *vif,
		unsigned char *macaddr);
static int agent_req_btm_steer(struct agent *a, const char *ifname,
		unsigned char *sta_macaddr, uint8_t pref_num, struct nbr *pref,
		uint32_t validity_int);
static void netif_free(struct netif_ap *ap);
static void netif_reset(struct netif_ap *ap);
#if (EASYMESH_VERSION >= 6)
static void clear_apmld(struct netif_mld *mld);
#endif
void agent_clear_bsta_mld(struct agent *a);

static void bsta_steer_cb(atimer_t *t)
{
	struct netif_bk *bk = container_of(t, struct netif_bk, steer_timeout);
	struct agent *a = bk->agent;
	char fmt[64] = {0};
	uint8_t *prev_bssid = bk->bsta_steer.prev_bssid;

	dbg("|%s:%d| steer timer expired, restoring old bssid (" MACFMT ") for"\
	    "bsta %s\n", __func__, __LINE__, MAC2STR(prev_bssid), bk->ifname);

	wifi_set_iface_bssid(bk, prev_bssid);

	/* Flawfinder: ignore */
	snprintf(fmt, sizeof(fmt), "set_network_bssid %s " MACFMT,
		 bk->ifname, MAC2STR(prev_bssid));
	agent_exec_platform_scripts(fmt);

	bk->bsta_steer.expired = true;

	if (bk->bsta_steer.trigger == BK_STEER_LOCAL) {
		agent_bsta_steer_clear(bk);
	} else if (a->connected) {
		/* if connected we can send response right away */
		bk->bsta_steer.reason = ERR_REASON_BH_STEERING_NOT_FOUND;
		send_backhaul_sta_steer_response(a, bk,
						 bk->bsta_steer.ul_ifname);
		agent_bsta_steer_clear(bk);
	}
}

int agent_exec_platform_scripts(char *arg)
{
	char buf[16] = {0};

	warn("/lib/wifi/multiap %s\n", arg);
	chrCmd(buf, sizeof(buf), "/lib/wifi/multiap %s", arg);
	return atoi(buf);
}

/* find node by macaddress */
struct node *agent_find_node(struct agent *a, uint8_t *almac)
{
	struct node *n = NULL;

	list_for_each_entry(n, &a->nodelist, list) {
		if (!memcmp(n->alid, almac, 6))
			return n;
	}

	return NULL;
}

struct node *agent_alloc_node(struct agent *a, uint8_t *almac)
{
	struct node *n;

	n = calloc(1, sizeof(struct node));
	if (!n) {
		warn("OOM: node malloc failed!\n");
		return NULL;
	}

	n->agent = a;
	memcpy(n->alid, almac, 6);
	n->map_profile = MULTIAP_PROFILE_1;

	list_add(&n->list, &a->nodelist);
	a->num_nodes++;

	dbg("|%s:%d| ------- " MACFMT "\n", __func__, __LINE__, MAC2STR(almac));
	return n;
}

struct node *agent_add_node(struct agent *a, uint8_t *almac)
{
	struct node *n;

	n = agent_find_node(a, almac);
	if (!n) {
		n = agent_alloc_node(a, almac);
		if (!n) {
			err("|%s:%d| failed to allocate node "MACFMT"\n",
			    __func__, __LINE__, MAC2STR(almac));
			return NULL;
		}
	} else {
		return n;
	}

	/* Add actions
	 * 1. Write to UCI?
	 */

#if (EASYMESH_VERSION > 2)
	qos_sync_rules_to_nodes(a);
#endif

	return n;
}

#if (EASYMESH_VERSION >= 6)
bool wifi_is_affiliated_ap(struct agent *a, uint8_t *bssid)
{
	struct netif_ap *ap;

	ap = agent_get_ap(a, bssid);
	if (ap && ap->is_affiliated_ap)
		return true;

	return false;
}


bool agent_radio_support_ap_wifi7(struct wifi7_radio_capabilities *wifi7_caps)
{
	if (wifi7_caps->ap_str_support || wifi7_caps->ap_nstr_support ||
		wifi7_caps->ap_emlsr_support || wifi7_caps->ap_emlmr_support)
		return true;

	return false;
}

bool agent_radio_support_bsta_wifi7(struct wifi7_radio_capabilities *wifi7_caps)
{
	if (wifi7_caps->bsta_str_support || wifi7_caps->bsta_nstr_support ||
		wifi7_caps->bsta_emlsr_support || wifi7_caps->bsta_emlmr_support)
		return true;

	return false;
}
#endif
struct wifi_radio_element *agent_get_radio_by_band(struct agent *a,
						   enum wifi_band band)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->band == band)
			return re;
	}

	return NULL;
}

struct wifi_radio_element *agent_get_radio_by_name(struct agent *a, const char *name)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		if (!memcmp(re->name, name, sizeof(re->name)))
			return re;
	}

	return NULL;
}


struct wifi_radio_element *agent_get_radio(struct agent *a, uint8_t *macaddr)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		if (!memcmp(re->macaddr, macaddr, 6))
			return re;
	}

	return NULL;
}

struct wifi_bss_element *agent_get_wifi_bss(struct agent *a, uint8_t *bssid)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		int i;

		for (i = 0; i < re->num_bss; i++) {
			if (hwaddr_equal(re->bsslist[i].bssid, bssid))
				return &re->bsslist[i];
		}
	}

	return NULL;
}

struct wifi_radio_element *agent_get_radio_with_ap(struct agent *a, uint8_t *bssid)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			if (!memcmp(ap->bssid, bssid, 6))
				return re;
		}
	}

	return NULL;
}

struct netif_ap *agent_get_ap_by_ifname(struct agent *a, const char *ifname)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			if (!memcmp(ap->ifname, ifname, sizeof(ap->ifname)))
				return ap;
		}
	}

	return NULL;
}

struct sta *agent_get_sta(struct agent *a, uint8_t *sta)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap;

		list_for_each_entry(ap, &re->aplist, list) {
			struct sta *s;

			list_for_each_entry(s, &ap->stalist, list) {
				if (memcmp(s->macaddr, sta, 6) == 0) {
					return s;
				}
			}
		}
	}

	return NULL;
}


struct netif_ap *agent_get_ap_with_sta(struct agent *a, uint8_t *sta)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			struct sta *s = NULL;

			trace("Looking for STAs in bssid = " MACFMT "\n",
					MAC2STR(ap->bssid));

			list_for_each_entry(s, &ap->stalist, list) {
				trace("stalist: " MACFMT "\n",
						MAC2STR(s->macaddr));
				if (memcmp(s->macaddr, sta, 6) == 0) {
					return ap;
				}
			}
		}
	}

	return NULL;
}

static const char *wifi_ifname_to_radioname(struct agent *a, char *ifname)
{
	struct agent_config_radio *rcfg;
	const char *name;

	/* do a backup lookup from configs */
	rcfg = get_agent_config_radio_from_iface(&a->cfg, ifname);
	if (rcfg) {
		name = rcfg->name;
		return name;
	}

	trace("|%s %d| couldn't find radio for ifname %s\n",
		   __func__, __LINE__, ifname);

	return NULL;
}

struct wifi_radio_element *agent_get_radio_with_ifname(struct agent *a,
		char *ifname)
{
	const char *radio;

	radio = wifi_ifname_to_radioname(a, ifname);
	if (!radio)
		return NULL;

	return agent_get_radio_by_name(a, radio);
}

static void wifiagent_log_steer(struct agent *a,
				unsigned char *stamac,
				char *vifname,
				char *steer_type,
				unsigned char *to_bss)
{
	char ev[512] = {0};
	unsigned char macaddr[6] = {0};
	char ifname[IFNAMSIZ] = {0};
	char type[32] = {0};

	if (!stamac || !vifname || !steer_type)
		return;

	memcpy(macaddr, stamac, 6);
	strncpy(ifname, vifname, IFNAMSIZ - 1);
	strncpy(type, steer_type, 31);

	snprintf(ev, sizeof(ev), "{\"macaddr\":\""MACFMT"\""
		 ",\"vif\":\"%.16s\",\"action\":\"%.32s\"",
		 MAC2STR(macaddr), ifname, type);
	if (to_bss) {
		snprintf(ev + strlen(ev), sizeof(ev), ",\"to\":\""MACFMT"\"}",
							MAC2STR(to_bss));
	} else {
		snprintf(ev + strlen(ev), sizeof(ev), "%s", "}");
	}

	info("steer: %s\n", ev);
	agent_notify_event(a, "map.agent", ev);
}

static void wifiagent_log_stainfo(struct agent *a, struct sta *s)
{
	int i;
	char ev[512] = {0};
	char rssis[64] = {0};

	if (!a || !s)
		return;

	for (i = 0; i < 4; i++) {
		if (s->rssi[i] > -100 && s->rssi[i] <= -10)
			snprintf(rssis + strlen(rssis), sizeof(rssis), "%d%s",
					s->rssi[i], i == 3 ? "" : ",");
	}

	snprintf(ev, sizeof(ev),
			"{\"macaddr\":\""MACFMT"\""
			",\"rssi_avg\":\"%d\""
			",\"rssi\":\"%s\""
			",\"tx_rate\":%u"
			",\"rx_rate\":%u"
			",\"tx_Bps\":%d"
			",\"rx_Bps\":%d"
			",\"tx_pkts\":%" PRIu64
			",\"rx_pkts\":%" PRIu64
			",\"tx_fpkts\":%u,\"rx_fpkts\":%u,\"rtx_pkts\":%u}",
			MAC2STR(s->macaddr), s->rssi_avg[0], rssis,
			s->tx_rate, s->rx_rate,
			s->tx_thput, s->rx_thput,
			s->tx_pkts,
			s->rx_pkts,
			s->tx_fail_pkts,
			s->rx_fail_pkts,
			s->rtx_pkts);

	//trace("stainfo: %s\n", ev);
	agent_notify_event(a, "map.agent", ev);
}

void wifiagent_log_cntlrinfo(struct agent *a)
{
	char ev[512] = {0};

	dbg("%s: called.\n", __func__);

	if (!a)
		return;

	sprintf(ev,
		"{\"active_cntlr\":%u"
		",\"cntlr_almac\":\""MACFMT"\""
		",\"local_cntlr\":%u"
		",\"multiple_cntlr_found\":%u}",
		a->active_cntlr ? true : false,
		MAC2STR(a->cntlr_almac),
		is_local_cntlr_running() ? 1 : 0,
		a->multiple_cntlr ? 1 : 0);

	trace("cntlrinfo: %s\n", ev);
	agent_notify_event(a, "map.agent", ev);
}

// static
int ubus_call_object(struct agent *a, wifi_object_t wobj,
		const char *method,
		void (*response_cb)(struct ubus_request *, int, struct blob_attr *),
		void *priv)
{
	struct blob_buf bb = {};
	int ret;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(a->ubus_ctx, wobj, method, bb.head,
				response_cb, priv, 2 * 1000);
	if (ret) {
		err("Failed to get '%s' (ret = %d)\n", method, ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
	return 0;
}

uint32_t ubus_get_object(struct ubus_context *ctx, const char *name)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(ctx, name, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", name);
		return WIFI_OBJECT_INVALID;
	}

	return id;
}

struct wifi_assoc_frame *wifi_get_frame(struct agent *a, uint8_t *macaddr)
{
	struct wifi_assoc_frame *f;

	list_for_each_entry(f, &a->framelist, list) {
		trace("frame mac: " MACFMT " input: " MACFMT "\n",
				MAC2STR(f->macaddr),
				MAC2STR(macaddr));
		if (!memcmp(f->macaddr, macaddr, 6))
			return f;
	}

	return NULL;
}

#if (EASYMESH_VERSION >=6)
struct netif_mld *agent_get_mld_by_ifname(struct agent *a, const char *ifname)
{
	struct netif_mld *mld = NULL;

	list_for_each_entry(mld, &a->apmldlist, list) {
		if (!strncmp(ifname, mld->ifname, sizeof(mld->ifname)))
			return mld;
	}

	return NULL;
}

struct netif_mld *agent_alloc_mld(struct agent *a, const char *ifname)
{
	struct netif_mld *mld;

	if (strlen(ifname) == 0) {
		err("|%s:%d| MLD can not have empty ifname\n", __func__, __LINE__);
		return NULL;
	}

	mld = calloc(1, sizeof(*mld));
	if (!mld)
		return NULL;

	dbg("|%s:%d| allocated mld ifname:%s\n", __func__, __LINE__, ifname);

	INIT_LIST_HEAD(&mld->stamldlist);

	strncpy(mld->ifname, ifname, sizeof(mld->ifname));
	list_add_tail(&mld->list, &a->apmldlist);
	a->num_ap_mld++;

	return mld;
}


bool agent_is_ap_in_mld(struct agent *a, struct netif_mld *mld,
			     uint8_t *macaddr)
{
	int i;

	for (i = 0; i < mld->num_affiliated_ap; i++) {
		if (!memcmp(mld->affiliated_ap[i]->bssid, macaddr, 6))
			return true;
	}

	return false;
}

bool agent_is_sta_in_sta_mld(struct agent *a, struct sta_mld *mld,
			     uint8_t *macaddr)
{
	int i;

	for (i = 0; i < mld->num_affiliated_sta; i++) {
		if (!memcmp(mld->affiliated_sta[i]->macaddr, macaddr, 6))
			return true;
	}

	return false;
}


bool agent_is_sta_in_bsta_mld(struct agent *a, struct bsta_mld *mld,
			     uint8_t *macaddr)
{
	int i;

	for (i = 0; i < mld->num_affiliated_bsta; i++) {
		if (!memcmp(mld->affiliated_bsta[i]->macaddr, macaddr, 6))
			return true;
	}

	return false;
}


struct sta_mld *agent_get_stamld(struct agent *a, uint8_t *macaddr)
{
	struct netif_mld *mld;

	list_for_each_entry(mld, &a->apmldlist, list) {
		struct sta_mld *s;

		list_for_each_entry(s, &mld->stamldlist, list) {
			if (!memcmp(macaddr, s->macaddr, 6))
				return s;
		}
	}

	return NULL;
}

struct sta_mld *agent_alloc_stamld(struct agent *a, struct netif_mld *mld,
				    uint8_t *macaddr)
{
	struct sta_mld *s;

	s = calloc(1, sizeof(*s));
	if (!s)
		return NULL;

	trace("|%s:%d| allocate sta mac:"MACFMT"\n", __func__, __LINE__, MAC2STR(macaddr));
	memcpy(s->macaddr, macaddr, 6);

	list_add_tail(&s->list, &mld->stamldlist);

	return s;
}

struct bsta_mld *agent_init_bstamld(struct agent *a, char *ifname,
				     struct wifi7_radio_capabilities *caps)
{
	struct bsta_mld *bsta = &a->bstamld;

	strncpy(bsta->ifname, ifname, sizeof(bsta->ifname));
	dbg("|%s:%d| allocated bsta mld:%s\n", __func__, __LINE__, bsta->ifname);

	bsta->str_enabled = caps->bsta_str_support;
	bsta->nstr_enabled = caps->bsta_nstr_support;
	bsta->emlsr_enabled = caps->bsta_emlsr_support;
	bsta->emlmr_enabled = caps->bsta_emlmr_support;
	a->has_bstamld = true;

	return bsta;
}
#endif

/* lookup netif_bk struct by device */
struct netif_bk *agent_bsta_in_radio(struct agent *a,
		const char *device)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->has_bsta && !strncmp(device, re->name, sizeof(re->name)))
			return &re->bk;
	}

	return NULL;
}

/* get netif_ap based on bssid */
struct netif_ap *agent_get_ap(struct agent *a, uint8_t *bssid)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			if (hwaddr_equal(ap->bssid, bssid))
				return ap;
		}
	}

	return NULL;
}

struct netif_bk *agent_get_bsta_by_band(struct agent *a, enum wifi_band band)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->has_bsta && re->band == band)
			return &re->bk;
	}

	return NULL;
}

struct netif_bk *agent_get_bsta_by_ifname(struct agent *a, const char *ifname)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->has_bsta && !strncmp(re->bk.ifname, ifname, sizeof(re->bk.ifname)))
			return &re->bk;
	}

#if (EASYMESH_VERSION >= 6)
	if (a->has_bstamld) {
		if (!strncmp(a->bstamld.bk.ifname, ifname, sizeof(a->bstamld.bk.ifname)))
			return &a->bstamld.bk;
	}
#endif

	return NULL;
}

/* find bkhaul by hwaddr */
struct netif_bk *agent_get_bsta(struct agent *a, uint8_t *macaddr)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->has_bsta && !memcmp(re->bk.macaddr, macaddr, 6))
			return &re->bk;
	}

#if (EASYMESH_VERSION >= 6)
	if (a->has_bstamld) {
		if (!memcmp(a->bstamld.bk.macaddr, macaddr, 6))
			return &a->bstamld.bk;
	}
#endif
	return NULL;
}

/* find wds by interface name */
struct netif_wds *agent_get_wds_by_ifname(struct agent *a, const char *ifname)
{
	struct netif_wds *p;

	list_for_each_entry(p, &a->wdslist, list) {
		if (!strncmp(p->ifname, ifname, sizeof(p->ifname)))
			return p;
	}

	return NULL;
}

void agent_link_ap_to_cfg(struct agent *a)
{
	struct netif_apcfg *acfg = NULL;

	list_for_each_entry(acfg, &a->cfg.aplist, list) {
		struct netif_ap *f;

		f = agent_get_ap_by_ifname(a, acfg->name);
		if (!f)
			continue;

		f->cfg = acfg;
	}
}

static void agent_config_load_post_action(struct agent *a)
{
	struct wifi_radio_element *re = NULL;

	agent_link_ap_to_cfg(a);

	if (a->cfg.pcfg)
		a->pvid = a->cfg.pcfg->pvid;

	/* Update restrict_stalist from assoc_ctrllist */
	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *p = NULL;

		list_for_each_entry(p, &re->aplist, list) {

			/* Copy restrictions from cfg to runtime */
			assoc_ctrl_sync_from_config(a, p);

			/* Block STAs in WiFi */
			assoc_ctrl_apply_restriction(a, p);
		}
	}
}

int agent_config_reload(struct agent *a)
{
	int ret;
	char ev[128] = { 0 };

	ret = agent_config_load(&a->cfg);
	agent_config_load_post_action(a);

	snprintf(ev, sizeof(ev), "{"\
			"\"event\":\"configuration\","\
			"\"data\": {"\
				"\"action\":\"%s\","\
				"\"status\":\"%s\","\
			"}"\
		"}", "reload", "success");


	agent_notify_event(a, "map.agent", ev);
	return ret;
}

int wifi_radio_scan_req(struct agent *a, struct wifi_radio_element *re,
			       const char *ssid, int num_opclass,
			       uint8_t *opclass, int num_channel,
			       uint8_t *channel)
{
	struct scan_param_ex param = {};
	int res = 0;

	if (ssid) {
		param.flag |= WIFI_SCAN_REQ_SSID;
		param.num_ssid = 1;

		strncpy(param.ssid[0], ssid, strlen(param.ssid[0]) - 1);
	}

	if (re) {
		re->scan_state = SCAN_REQUESTED;
		return wifi_scan(re->name, &param, num_opclass, opclass,
						num_channel, channel);
	}

	list_for_each_entry(re, &a->radiolist, list) {
		re->scan_state = SCAN_REQUESTED;
		res |= wifi_scan(re->name, &param, num_opclass, opclass,
						num_channel, channel);
	}

	return res;
}

int wifi_radio_scan_req_all(struct agent *a)
{
	return wifi_radio_scan_req(a, NULL, NULL, 0, NULL, 0, NULL);
}

int issue_channel_scan(struct agent *a, struct wifi_radio_element *re,
			      struct wifi_scan_request_radio *req)
{
	trace("%s: --->\n", __func__);

	uint8_t opclasses[128] = {};
	uint8_t channels[128] = {};
	int op_idx = 0, ch_idx = 0;
	int i, j;
	int ret = 0;

	if (WARN_ON(!a ||!re || !req))
		return -1;

	trace("Attempting to scan radio %s neighbors\n", re->name);

	/* pass all opclasses & channels */
	for (i = 0; i < req->num_opclass; i++) {

		for (j = 0; j < req->opclass[i].num_channel; j++) {
			if (ch_idx >= ARRAY_SIZE(channels))
				break;

			if (req->opclass[i].channel[j] > 0)
				channels[ch_idx++] =
					req->opclass[i].channel[j];
		}

		/* Channels listed explicitly */
		if (req->opclass[i].num_channel)
			continue;

		if (req->opclass[i].classid == 0)
			/* only channels provided, skip opclass */
			// TODO: revisit
			continue;

		if (op_idx >= ARRAY_SIZE(opclasses))
			break;

		opclasses[op_idx++] = req->opclass[i].classid;

		/* num_channel == 0 indicates that the Multi-AP Agent is
		 * requested to scan on all channels in the Operating Class.
		 */
		if (req->opclass[i].num_channel == 0) {
			ch_idx = ARRAY_SIZE(channels);
			wifi_opclass_get_supported_ctrl_channels(
						&re->opclass,
						req->opclass[i].classid,
						channels,
						&ch_idx);
		}
	}

	/* Pass all the opclasses & channels from the request */
	if (op_idx > 0 || ch_idx > 0)
		ret = wifi_radio_scan_req(a, re, NULL, op_idx, opclasses,
					  ch_idx, channels);
	else
		ret = wifi_radio_scan_req_all(a);

	return ret;
}

uint8_t agent_steer_next_dialog_token(struct agent *a)
{
	a->steer_dialog_token %= 255;

	return ++a->steer_dialog_token;
}

typedef void (*destructor)(void *);

#define delete_expired_entries(vif, type, h, l, ts_member, tmo, func, _nr) \
do {									   \
	type *e, *etmp;							   \
	typeof((destructor) func) fptr = (func);			   \
									   \
	list_for_each_entry_safe(e, etmp, h, l) {			   \
		if (timestamp_expired(&e->ts_member, tmo)) {		   \
			dbg("%s: Entry aged out.. delete.\n", __func__);   \
			if (fptr)					   \
				fptr(e);				   \
			list_del(&e->l);				   \
			(_nr)--;                                           \
			free(e);					   \
		}							   \
	}								   \
} while (0)

#ifdef VENDOR_EXTENSION
static int agent_load_extensions(struct agent *a)
{
#define EXTENSION_PLUGIN_MAX_NUM	8
	char *argv[EXTENSION_PLUGIN_MAX_NUM] = {0};
	struct agent_extension_config *e;
	int argc = 0;
	int ret;


	list_for_each_entry(e, &a->cfg.extlist, list) {
		argv[argc++] = e->name;

		if (argc >= EXTENSION_PLUGIN_MAX_NUM)
			break;
	}

	info("map-agent: load extensions\n");
	INIT_LIST_HEAD(&a->extlist);
	ret = load_extensions(argc, argv, &a->extlist);

	return ret;
}

static void agent_unload_extensions(struct agent *a)
{
	if (a)
		unload_extensions(&a->extlist);
}
#endif

int (*compare)(const void *a, const void *b);

/* Private 'cmp()' function to sort neighbors by RSSI decreasing */
static int cmp_nbr_rssi(void *priv, struct list_head *a, struct list_head *b)
{
	struct pref_neighbor *ea, *eb;

	UNUSED(priv);
	ea = container_of(a, struct pref_neighbor, list);
	eb = container_of(b, struct pref_neighbor, list);

	return eb->rssi - ea->rssi;
}

#define alloc_entry(type)		\
({					\
	type *__n;			\
	__n = calloc(1, sizeof(type));	\
	__n ? &__n->list : NULL;	\
})

#define free_entry(ptr, type)					\
({								\
	if (ptr)						\
		free(container_of(ptr, type, list));		\
})

#define copy_entry(from, to, type)			\
({							\
	type *__s, *__d;				\
	if (from && to)	{				\
		__s = container_of(from, type, list);	\
		__d = container_of(to, type, list);	\
		memcpy(__d, __s, sizeof(type));		\
	}						\
})

static void *alloc_neighbor_entry(void)
{
	struct neighbor *n;

	n = calloc(1, sizeof(struct neighbor));
	if (n)
		return &n->list;

	return NULL;
}

static void free_neighbor_entry(struct list_head *n)
{
	free_entry(n, struct neighbor);
}

static void copy_neighbor_entry(struct list_head *from, struct list_head *to)
{
	copy_entry(from, to, struct neighbor);
}

static void *alloc_sta_neighbor_entry(void)
{
	struct sta_neighbor *n;

	n = calloc(1, sizeof(struct sta_neighbor));
	if (n)
		return &n->list;

	return NULL;
}

static void free_sta_neighbor_entry(struct list_head *n)
{
	free_entry(n, struct sta_neighbor);
}

static void copy_sta_neighbor_entry(struct list_head *from,
		struct list_head *to)
{
	copy_entry(from, to, struct sta_neighbor);
}

/* Function used to match duplicate entries in neighbor list */
static int match_bssid(void *priv, struct list_head *a, struct list_head *b)
{
	struct neighbor *ea;
	struct sta_neighbor *eb;

	UNUSED(priv);
	ea = container_of(a, struct neighbor, list);
	eb = container_of(b, struct sta_neighbor, list);

	return hwaddr_equal(ea->nbr.bssid, eb->nbr.bssid);
}

static struct list_head *create_joined_node(void *priv, struct list_head *a,
		struct list_head *b)
{
	struct pref_neighbor *new;
	struct neighbor *ea = NULL;
	struct sta_neighbor *eb = NULL;
	struct wifi_bss bss;
	struct sta *sta = (struct sta *)priv;
	struct netif_ap *nif = NULL;

	if (a)
		ea = container_of(a, struct neighbor, list);

	if (b)
		eb = container_of(b, struct sta_neighbor, list);

	if (!ea && !eb)
		return NULL;

	new = malloc(sizeof(*new));
	if (!new) {
		warn("OOM: failed to alloc preferred neighbor!\n");
		return NULL;
	}

	memset(new, 0, sizeof(*new));

	if (ea)
		memcpy(new->bssid, ea->nbr.bssid, 6);
	else
		memcpy(new->bssid, eb->nbr.bssid, 6);

	if (sta)
		nif = sta->vif;

	memset(&bss, 0, sizeof(struct wifi_bss));
	if (nif && wifi_scanresults_get_bss(nif->ifname, new->bssid, &bss)) {
		/* new->rssi = bss.rssi; */ /* it is AP's view; meaningless */
		new->ch_util = bss.load.utilization;
		new->num_stas = bss.load.sta_count;
	}

	if (ea) {
		new->bssid_info = ea->nbr.bssid_info;
		new->reg = ea->nbr.reg;
		new->channel = ea->nbr.channel;
		new->phy = ea->nbr.phy;
		new->flags = ea->flags;
		/* TODO: est_bwrate from snr */
	}

	if (eb) {
		new->rssi = eb->nbr.rssi; /* override rssi by sta's view */
		new->rsni = eb->nbr.rsni;
	}

	return &new->list;
}

static void free_joined_node(void *priv, struct list_head *a)
{
	struct pref_neighbor *ea = NULL;

	UNUSED(priv);
	if (a)
		ea = container_of(a, struct pref_neighbor, list);

	free(ea);
}

static void recalc_desired_neighbors(struct sta *s)
{
	struct netif_ap *vif = s->vif;
	struct list_head nbrlist_copy, sta_nbrlist_copy;

	/**
	 * First, does a splice of the two lists -
	 *	11k neighbor list and
	 *	11k STA beacon list.
	 *
	 * It then rearranges the combined list on the basis of decreasing
	 * 'rssi'.
	 * In rule matching modules, walk the pref_nbrlist to select the
	 * next best neighbor Bss based on rssi, bssload, est_dl_throughput etc.
	 */

	INIT_LIST_HEAD(&nbrlist_copy);
	list_dup(&vif->nbrlist, &nbrlist_copy,
			alloc_neighbor_entry,
			free_neighbor_entry,
			copy_neighbor_entry);

	INIT_LIST_HEAD(&sta_nbrlist_copy);
	list_dup(&s->sta_nbrlist, &sta_nbrlist_copy,
			alloc_sta_neighbor_entry,
			free_sta_neighbor_entry,
			copy_sta_neighbor_entry);

	list_flush(&s->pref_nbrlist, struct pref_neighbor, list);

	dbg_list_print("Neighbor nodes", &nbrlist_copy,
			struct neighbor, list, nbr.bssid);

	list_join_uniq(s, &nbrlist_copy, &sta_nbrlist_copy, &s->pref_nbrlist,
			match_bssid, create_joined_node, free_joined_node,
			free_neighbor_entry, free_sta_neighbor_entry);

	dbg_list_print("STA Neighbors", &s->pref_nbrlist,
			struct pref_neighbor, list, bssid);

	list_flush(&nbrlist_copy, struct neighbor, list);
	list_flush(&sta_nbrlist_copy, struct sta_neighbor, list);

	list_sort(NULL, &s->pref_nbrlist, cmp_nbr_rssi);
}

static void agent_sta_add_rssi(struct sta *s, int8_t new_rssi)
{
	int i;

	/* Shift all elements to the right */
	for (i = MAX_RSSI_MEAS - 1; i > 0; i--)
		s->rssi_avg[i] = s->rssi_avg[i - 1];

	/* Insert new RSSI at index 0 */
	s->rssi_avg[0] = new_rssi;

	/* Update the number of RSSI samples */
	if (s->num_rssi < MAX_RSSI_MEAS)
		s->num_rssi++;
}

static int update_sta_entry(struct agent *a, struct netif_ap *vif, struct wifi_sta *e)
{
	struct sta *s;

	list_for_each_entry(s, &vif->stalist, list) {
		if (!memcmp(s->macaddr, e->macaddr, 6)) {
			bool no_traffic = true;

			/* dbg("%s: update STA " MACFMT " entry\n", __func__,
			 *	MAC2STR(s->macaddr));
			 */
			timestamp_update(&s->last_update);
			if (s->rssi[0] != e->rssi[0] ||
					s->rssi[1] != e->rssi[1] ||
					s->rssi[2] != e->rssi[2] ||
					s->rssi[3] != e->rssi[3]) {

				no_traffic = false;

				s->rssi[0] = e->rssi[0];
				s->rssi[1] = e->rssi[1];
				s->rssi[2] = e->rssi[2];
				s->rssi[3] = e->rssi[3];
			}

			if (s->rssi_avg[0] != e->rssi_avg)
				no_traffic = false;
			agent_sta_add_rssi(s, (int8_t)e->rssi_avg);

			s->connected_ms = e->conn_time * 1000;
			s->tx_rate = e->tx_rate.rate;
			s->rx_rate = e->rx_rate.rate;

			s->rx_airtime = e->tx_airtime / 1000;
			s->tx_airtime = e->rx_airtime / 1000;

			s->tx_thput = 0;
			s->rx_thput = 0;
			if (e->stats.tx_bytes != s->tx_bytes ||
					e->stats.rx_bytes != s->rx_bytes) {
				uint64_t tx_bytes_lastsec;
				uint64_t rx_bytes_lastsec;

				no_traffic = false;

				tx_bytes_lastsec = s->tx_bytes;
				rx_bytes_lastsec = s->rx_bytes;
				s->tx_bytes = e->stats.tx_bytes;
				s->rx_bytes = e->stats.rx_bytes;
				s->tx_thput = s->tx_bytes - tx_bytes_lastsec;
				s->rx_thput = s->rx_bytes - rx_bytes_lastsec;
			}

			s->est_rx_thput = e->est_rx_thput;
			s->est_tx_thput = e->est_tx_thput;

			if (e->stats.tx_pkts != s->tx_pkts ||
					e->stats.rx_pkts != s->rx_pkts) {

				no_traffic = false;

				s->tx_pkts = e->stats.tx_pkts;
				s->rx_pkts = e->stats.rx_pkts;
			}

			if (no_traffic)
				s->stale_stats_cnt++;
			else
				s->stale_stats_cnt = 0;

			s->tx_fail_pkts = e->stats.tx_fail_pkts;
			s->rx_fail_pkts = e->stats.rx_fail_pkts;
			s->rtx_pkts = e->stats.tx_retry_pkts;


			memcpy(s->bssid, e->bssid, 6);

		/*loud("STA: " MACFMT " (rssi = %d)   " \
				"Tx-bytes/sec = %u    Rx-bytes/sec = %u\n",
				MAC2STR(s->macaddr), s->rssi[0],
				s->tx_thput, s->rx_thput);*/
			wifiagent_log_stainfo(vif->agent, s);
			return 0;
		}
	}

	return -1;
}

static int agent_req_beacon_metrics(struct agent *a,
	struct netif_ap *ap, uint8_t *sta_addr, uint8_t opclass,
	uint8_t channel, uint8_t *bssid, uint8_t reporting_detail,
	uint8_t ssid_len, char *ssid, uint8_t num_report,
	uint8_t *report, uint8_t num_element, uint8_t *element)
{
	struct wifi_request_neighbor_param param = {};

	trace("agent: %s: --->\n", __func__);

	if (!sta_addr)
		return -1;

	param.opclass = opclass;
	param.channel = channel;
	param.bssid = bssid;
	param.reporting_detail = reporting_detail;
	param.ssid_len = ssid_len;
	param.ssid = ssid;
	param.num_report = num_report;
	param.report = report;
	param.num_element = num_element;
	param.element = element;

	return wifi_req_neighbor(ap->ifname, sta_addr, &param);
}

static void wifi_sta_bcn_req(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_bcn_req_timer);
	struct netif_ap *vif = s->vif;
	struct agent *a = vif->agent;
	struct sta_bcn_req *breq;

	if (s->sta_bcn_req_nr < 1)
		/* No request enqueued */
		return;

	/* LIFO */
	breq = &s->bcn_req_queue[s->sta_bcn_req_nr - 1];

	agent_req_beacon_metrics(a, breq->ap, s->macaddr,
			breq->opclass, breq->channel, breq->bssid,
			breq->reporting_detail,
			breq->ssid_len, breq->ssid, 0, NULL,
			breq->num_element, breq->element);

	/* Dequeue */
	memset(breq, 0, sizeof(struct sta_bcn_req));
	s->sta_bcn_req_nr--;

	if (s->sta_bcn_req_nr >= 1)
		/* requests remainig, send next in 3 sec */
		timer_set(&s->sta_bcn_req_timer, 3 * 1000);
}

static void wifi_sta_finalize(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_finalize_timer);
	struct sta *p, *tmp;
	struct netif_ap *vif = s->vif;

	list_for_each_entry_safe(p, tmp, &vif->stalist, list) {
		if (!memcmp(p->macaddr, s->macaddr, 6)) {
			if (s->legacy_steered)
				s->legacy_steered = false;

			if (!s->ref) {
				dbg("Finalize STA " MACFMT "\n",
						MAC2STR(s->macaddr));
				timer_del(&s->sta_timer);
				timer_del(&s->sta_bcn_req_timer);
				timer_del(&s->sta_steer_timer);
				list_del(&s->list);
				list_flush(&s->pref_nbrlist, struct pref_neighbor, list);
				list_flush(&s->sta_nbrlist, struct sta_neighbor, list);
				if (s->assoc_frame) {
					free(s->assoc_frame->frame);
					free(s->assoc_frame);
				}
#if (EASYMESH_VERSION >= 6)
				if (s->is_affiliated_sta)
					wifi_sta_mld_del_sta(vif->agent, s);
#endif
				free(s);
			}
		}
	}
}

static int steer_sta_legacy(struct sta *s)
{
	struct netif_ap *vif = s->vif;
	uint16_t reason = 0;

	if (s->legacy_steered) {
		dbg("Skip Legacy Steering " MACFMT " again\n",
				MAC2STR(s->macaddr));
		return 0;
	}

	s->legacy_steered = true;
	dbg("Try to Legacy Steer " MACFMT " ------>\n", MAC2STR(s->macaddr));
	wifi_disconnect_sta(vif->ifname, s->macaddr, reason);
	wifiagent_log_steer(vif->agent, s->macaddr, vif->ifname,
					"steer_legacy", NULL);

	timer_set(&s->sta_finalize_timer,
			vif->cfg->steer_legacy_reassoc_secs * 1000);

	return 0;
}

/* Translate internal 'pref_neighbor' to wifi's 'nbr' struct */
void agent_pref_neighbor_to_nbr(struct pref_neighbor *pref_nbr_src,
		struct nbr *wifi_nbr_dst)
{
	memcpy(wifi_nbr_dst->bssid, pref_nbr_src->bssid, 6);
	wifi_nbr_dst->bssid_info = pref_nbr_src->bssid_info;
	wifi_nbr_dst->reg = pref_nbr_src->reg;
	wifi_nbr_dst->channel = pref_nbr_src->channel;
	wifi_nbr_dst->phy = pref_nbr_src->phy;
}

static int steer_sta(struct sta *s, struct pref_neighbor *pref_nbr)
{
	struct netif_ap *vif = s->vif;
	struct agent *a = vif->agent;
	struct pref_neighbor *pref_nbr_el;
	struct nbr wifi_nbr_el = {};
	int ret = 0;

	s->steer_secs++;
	trace("Steer STA " MACFMT " (steer_secs = %u)\n",
				MAC2STR(s->macaddr), s->steer_secs);

	if (pref_nbr) {
		pref_nbr_el = pref_nbr;
	} else {
		if (list_empty(&s->pref_nbrlist)) {
			trace("No better BSS for STA " MACFMT " found\n",
							MAC2STR(s->macaddr));
			return 0;
		}

		pref_nbr_el = list_first_entry(&s->pref_nbrlist,
						struct pref_neighbor, list);
	}

	if (!pref_nbr_el) {
		warn("Unexpected! pref_nbr_el is NULL or empty!\n");
		return -1;
	}

	/* Translate internal 'pref_neighbor' to wifi's 'nbr' struct */
	agent_pref_neighbor_to_nbr(pref_nbr_el, &wifi_nbr_el);

	/* btm steer ? */
	if (!s->steer_btm_cnt) {
		s->steer_btm_cnt++;
		info("Try {%d} to BTM Steer " MACFMT " =======>\n",
				s->steer_btm_cnt, MAC2STR(s->macaddr));

		ret = agent_req_btm_steer(a, vif->ifname, s->macaddr, 1, &wifi_nbr_el, 0);
		if (!ret) {
			wifiagent_log_steer(a, s->macaddr, vif->ifname,
					"steer_btmreq", wifi_nbr_el.bssid);
		} else  {
			warn("Failed to send BTM request to " MACFMT "\n",
					MAC2STR(s->macaddr));
		}

		return ret;
	}

	/* retry btm steer ? */
	if (vif->cfg->steer_btm_retry) {
		if (!((s->steer_secs - 1) % vif->cfg->steer_btm_retry_secs) &&
			(s->steer_btm_cnt < vif->cfg->steer_btm_retry + 1)) {

			s->steer_btm_cnt++;
			info("Try {%d} to BTM Steer " MACFMT " =======>\n",
					s->steer_btm_cnt, MAC2STR(s->macaddr));

			ret = agent_req_btm_steer(a, vif->ifname, s->macaddr,1, &wifi_nbr_el, 0);
			if (!ret)
				wifiagent_log_steer(a, s->macaddr, vif->ifname,
						"steer_btmreq", wifi_nbr_el.bssid);
			else
				warn("Failed to send BTM request to " MACFMT "\n",
						MAC2STR(s->macaddr));

			return ret;
		}
	}

	/* fallback to legacy steer ? */
	if (vif->cfg->fallback_legacy
			&& !s->legacy_steered
			&& (s->steer_secs >= vif->cfg->fallback_legacy))
		return steer_sta_legacy(s);

	return ret;
}

static void rebuild_cntlr_preflist(struct agent *a, struct sta *s,
					int cnt, struct pref_neighbor *pref)
{
	int sz = sizeof(int) + cnt * sizeof(*pref);

	if (s->cntlr_preflist)
		free(s->cntlr_preflist);

	s->cntlr_preflist = malloc(sz);
	if (s->cntlr_preflist) {
		memset(s->cntlr_preflist, 0, sz);
		s->cntlr_preflist->num = cnt;
		memcpy(s->cntlr_preflist->pref, pref, cnt);
	} else
		warn("OOM: cntlr_preflist\n");
}

int calculate_steer_verdict(struct sta *s, struct pref_neighbor **nbr)
{
	struct steer_rule *r;
	steer_verdict_t v = STEER_SKIP;

	list_for_each_entry(r, &regd_steer_rules, list) {
		if (!r->enabled)
			continue;

		if (r->check) {
			loud("STA " MACFMT " Check rule '%s'\n",
					MAC2STR(s->macaddr), r->name);
			v = r->check(r, s, nbr);
		}
		if (v != STEER_SKIP)
			return v;
	}

	return v;
}

/* returns 0 = don't steer, = 1 for steer */
static int should_steer_sta(struct sta *s, unsigned int *res)
{
	*res = 0;

	/* TODO:
	 * Check sta stats viz. active data sessions, retransmits etc. to
	 * decide if should steer sta.
	 * If verdict is OK, steer to cntlr_preflist.
	 */

	return 0;
}

/* returns 0 = don't steer, >= 1 for steer */
static int maybe_steer_sta(struct sta *s, unsigned int *res,
					struct pref_neighbor **nbr)
{
	struct netif_ap *vif = s->vif;
	struct agent *a = vif->agent;
	struct stax *x = NULL;

	if (!a->cfg.pcfg) {
		/* Do nothing */
		return 0;
	}

	/* dbg("%s: vif = %p  cfg = %p ifname = %s\n",
	 *	__func__, vif, cfg, vif->ifname);
	 */

	/* TODO: don't check here ..
	 * flag exclude whenever cfg is updated to exclude a STA.
	 */
	list_for_each_entry(x, &a->cfg.pcfg->steer_excludelist, list) {
		if (!memcmp(s->macaddr, x->macaddr, 6)) {
			loud("STA " MACFMT " in exclude list. Do nothing\n",
					MAC2STR(s->macaddr));
			return 0;
		}
	}

	/* Following if local steering is NOT disallowed */
	if (calculate_steer_verdict(s, nbr) == STEER_OK)
		return 1;

	return 0;
}

static void wifi_sta_steer_timeout(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_steer_timer);
	unsigned int reason;

	if (timestamp_expired(&s->steer_opportunity, s->steer_opportunity_tmo)) {
		/* close steer opportunity window */
		s->steer_policy &= ~STA_STEER_OPPORTUNITY;
		return;
	}

	if (should_steer_sta(s, &reason)) {
		steer_sta(s, NULL);
		/* log_steer_action("Cntlr", s->macaddr, reason, ret); */
	}

	/* check for steer opportunity again in 1 sec */
	timer_set(&s->sta_steer_timer, 1000);
}

static void wifi_sta_steered_deauth(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_steer_deauth_timer);
	struct netif_ap *vif = s->vif;

	warn("%s: Disconnect station: " MACFMT " from interface: %s with reson:%i\n",
		__func__, MAC2STR(s->macaddr), vif->ifname, WIFI_REASON_BSS_TRANSITION_DISASSOC);
	wifi_disconnect_sta(vif->ifname, s->macaddr, WIFI_REASON_BSS_TRANSITION_DISASSOC);
}

static int cond_refresh_sta_neighbor_list(struct agent *a, struct sta *s)
{
	struct netif_ap *vif = s->vif;
#define STA_NBR_REFRESH_BAD_RSSI	-78
	char ev[256] = {0};

	// TODO: use hysteresis
	if (s->rssi_avg[0] > STA_NBR_REFRESH_BAD_RSSI) {
		s->inform_leaving = 0;
		return 0;
	}

#if 0
	s->sta_nbr_invalid %= STA_NBR_REFRESH_CNT * a->cfg.runfreq;
	if (!s->sta_nbr_invalid)
		refresh_sta_neighbor_list(vif, s);

	s->sta_nbr_invalid++;
#endif
#if 0
	if (s->supports_bcnreport && s->sta_nbr_nr == 0)
		refresh_sta_neighbor_list(vif, s);
#endif

	if (!(s->inform_leaving++ % 5)) {
		/* tell cntlr about this bad rssi */
		snprintf(ev, sizeof(ev),
			"{\"macaddr\":\""MACFMT"\""
			",\"vif\":\"%s\""
			",\"action\":\"monitor\""
			",\"lowrssi\":%d}",
			MAC2STR(s->macaddr), vif->ifname, s->rssi_avg[0]);

		agent_notify_event(a, "map.agent", ev);
		/* s->inform_leaving = (s->inform_leaving + 1) % 3; */
		s->wait_for_cntlr_nbr = true;

		/* If no pref_nbr reply from cntlr, and
		 * sta->supports_bcnreport = false, then do not steer.
		 * Either of these is true, then steer based on other
		 * criteria.
		 */
	}

	return 0;
}

#if (EASYMESH_VERSION >= 6)
static void mlsta_update_sta_mld(struct agent *a, struct netif_ap *vif,
			struct wifi_mlsta *mlsta, struct sta *s)
{
	struct sta_mld *mld;
	int num_sta = 0;

	if (!vif || !mlsta)
		return;

	//err("|%s:%d| mldmac:"MACFMT" linkmac:"MACFMT"\n",
	//    __func__, __LINE__,
	//    MAC2STR(vif->mld->macaddr), MAC2STR(mlsta->macaddr));

	mld = agent_get_stamld(a, mlsta->macaddr);
	if (!mld) {
		mld = agent_alloc_stamld(a, vif->mld, mlsta->macaddr);
		if (!mld)
			return;
	}

	memcpy(mld->bssid, mlsta->bssid, 6);
	num_sta = mld->num_affiliated_sta;

	s->mld = mld;
	s->is_affiliated_sta = true;
	s->mlo_link_id = mlsta->sta[0].mlo_link_id;

	if (!agent_is_sta_in_sta_mld(a, mld, mlsta->sta[0].macaddr) &&
		    num_sta < MAX_AFFILIATED_STA_LINKS_NUM) {
		mld->affiliated_sta[num_sta] = s;
		mld->num_affiliated_sta++;
	}

	//err("|%s:%d| num affiliated_sta:%d\n",
	//    __func__, __LINE__,
	//    mld->num_affiliated_sta);
}
#endif

#define STA_STALE_STATS_CNT_MAX 60
static void wifi_sta_periodic_run(atimer_t *t)
{
	struct sta *s = container_of(t, struct sta, sta_timer);
	struct pref_neighbor *pref_nbr = NULL;
	struct netif_ap *vif = s->vif;
	struct wifi_sta sta = {};
	unsigned int reason;
	struct agent *a;
	int ret = 0;

	if (!vif)
		return;

	a = vif->agent;

	/* if (strcmp(vif->ifname, s->vif_name))
	 *	err(stderr, "%s: vif changed under the hood!\n", __func__);
	 */
	if (!vif->cfg || !vif->cfg->enabled)
		return;

	//trace("%s: STA = " MACFMT " ref = %d\n", __func__,
	//					MAC2STR(s->macaddr), s->ref);

	if (s->legacy_steered && s->ref == 2)
		goto rearm_periodic;

	//err("|%s:%d| vif:%s mac:"MACFMT"\n",
	//    __func__, __LINE__,
	//    vif->ifname,MAC2STR(s->macaddr));

#if (EASYMESH_VERSION >= 6)
	if (!vif->is_affiliated_ap) {
#endif
		ret = wifi_get_station(vif, s->macaddr, &sta);
		if (ret)
			goto rearm_periodic;
		update_sta_entry(a, vif, &sta);
#if (EASYMESH_VERSION >= 6)
	} else {
		struct wifi_mlsta mlsta = {0};
		uint8_t *mld_macaddr = NULL;

		if (s->mld)
			mld_macaddr = s->mld->macaddr;

		ret = wifi_get_mld_station(vif, mld_macaddr, s->macaddr, &mlsta);
		if (ret)
			goto rearm_periodic;

		if (hwaddr_is_zero(mlsta.sta[0].macaddr)) {
			dbg("%s: sta:"MACFMT" is no longer connected\n",
			    __func__, MAC2STR(s->macaddr));
			wifi_topology_notification(a, s, vif->ifname,
					0x00 /* left */,
					0x01 /* unspecified reason */);
			wifi_del_sta(a, vif->ifname, s->macaddr);
			return;
		}


		//err("%s: mac:"MACFMT" linkbssid:"MACFMT" mlo_link_id:%d\n",
		//    __func__, MAC2STR(s->macaddr),
		//    MAC2STR(mlsta.sta[0].bssid),
		//    mlsta.sta[0].mlo_link_id);

		if (mlsta.mlo_capable == true
		    && !hwaddr_is_zero(mlsta.sta[0].bssid)
		    && vif->mld) {
			mlsta_update_sta_mld(a, vif, &mlsta, s);
		}
		update_sta_entry(a, vif, &(mlsta.sta[0]));
	}
#endif

	cond_refresh_sta_neighbor_list(a, s);

	if (s->stale_stats_cnt > STA_STALE_STATS_CNT_MAX) {
		dbg("Stale STA " MACFMT ", sending disconnect.\n",
		    MAC2STR(s->macaddr));

		/* Inform driver about missing STA */
		wifi_disconnect_sta(vif->ifname, s->macaddr,
				WIFI_REASON_DISASSOC_STA_HAS_LEFT);

		/* Clean up STA in map */
		wifi_topology_notification(a, s, vif->ifname,
				    0x00 /* left */,
				    0x01 /* unspecified reason */);
		wifi_del_sta(a, vif->ifname, s->macaddr);
		return;
	}

	if (s->rssi_avg[0] && s->rssi_avg[1]) {
		struct agent_config_radio *rcfg =
			get_agent_config_radio(&a->cfg, vif->cfg->device);
		uint8_t curr_rcpi = rssi_to_rcpi(s->rssi_avg[0]);
		uint8_t prev_rcpi = rssi_to_rcpi(s->rssi_avg[1]);
		uint8_t trigger = rcfg->report_rcpi_threshold;

		if ((prev_rcpi <= trigger && curr_rcpi > trigger) ||
		    (prev_rcpi >= trigger && curr_rcpi < trigger))
			timer_set(&a->rcpi_threshold_timer, 0);
	}

	if (sta_steer_allowed(s->steer_policy)
			&& !list_empty(&s->pref_nbrlist)
			&& maybe_steer_sta(s, &reason, &pref_nbr)) {

		steer_sta(s, pref_nbr);
	}

	if (!list_empty(&vif->nbrlist) && list_empty(&s->pref_nbrlist))
		recalc_desired_neighbors(s);

rearm_periodic:
	timer_set(&s->sta_timer, STA_PERIODIC_RUN_INTERVAL);
}

static int wifi_send_sta_report(struct agent *a, const char *ifname,
		uint8_t *macaddr, uint32_t status, uint8_t *bssid)
{
	struct netif_ap *ap = NULL;
	uint8_t ret = 0;

	ap = agent_get_ap_by_ifname(a, ifname);
	if (!ap)
		return -1;


	/* Here we get need to send the steering report */
	ret = send_steer_btm_report(a, a->cntlr_almac, bssid,
			ap->bssid, macaddr, status);
	return ret;
}

static void agent_sta_clean_rssi(struct sta *s)
{
	int i;

	for (i = 0; i < MAX_RSSI_MEAS; i++)
		s->rssi_avg[i] = 0;
	s->num_rssi = 0;
}

static struct sta *wifi_add_sta(struct agent *a, const char *ifname,
				unsigned char *macaddr)
{
	struct netif_ap *ap;
	//struct netif_apcfg *cfg;
	struct sta *sptr, *new;
	struct wifi_assoc_frame *af;

	ap = agent_get_ap_by_ifname(a, ifname);
	if (!ap)
		return NULL;

	sptr = agent_get_sta(a, macaddr);
	if (sptr) {
		if (sptr->vif == ap) {
			dbg("STA " MACFMT "already in list.\n",
					MAC2STR(macaddr));
			if (sptr->legacy_steered) {
				timer_set(&sptr->sta_finalize_timer,
						ap->cfg->steer_legacy_retry_secs * 1000);
			}

			/* In case station is already on stalist, cancel steered station
			 * deauthentication timer if it is stil runnning.
			 */
			if (timer_pending(&sptr->sta_steer_deauth_timer))
				timer_del(&sptr->sta_steer_deauth_timer);

			agent_sta_clean_rssi(sptr);
			return 0;
		} else {
			/* TODO: send topolog notification? */
			clear_sta(sptr);
		}

	}
	/* add new sta */
	new = malloc(sizeof(struct sta));
	if (!new) {
		warn("OOM: new sta malloc failed!\n");
		return NULL;
	}
	memset(new, 0, sizeof(struct sta));
	memcpy(new->macaddr, macaddr, 6);
	new->vif = ap;
	timer_init(&new->sta_timer, wifi_sta_periodic_run);
	timer_init(&new->sta_bcn_req_timer, wifi_sta_bcn_req);
	timer_init(&new->sta_steer_timer, wifi_sta_steer_timeout);
	timer_init(&new->sta_finalize_timer, wifi_sta_finalize);
	timer_init(&new->sta_steer_deauth_timer, wifi_sta_steered_deauth);
	timestamp_update(&new->last_update);
	INIT_LIST_HEAD(&new->sta_nbrlist);
	INIT_LIST_HEAD(&new->pref_nbrlist);
	recalc_desired_neighbors(new);
#if 0
	refresh_sta_neighbor_list(ap, new);
#endif
	new->steer_policy |= STA_STEER_ALLOWED;
	dbg("STA steer policy = 0x%x\n", new->steer_policy);
	af = wifi_get_frame(a, new->macaddr);
	if (af) {
		new->assoc_frame = af;
		list_del(&af->list);
	}

	ap->num_sta++;

	/* INIT_LIST_HEAD(&new->cntlr_preflist); */
	list_add(&new->list, &ap->stalist);

	timer_set(&new->sta_timer, 1 * 1000);
	return new;
}

#if (EASYMESH_VERSION >= 6)
/* remove passed affiliated STA from mld */
int wifi_sta_mld_del_sta(struct agent *a, struct sta *s)
{
	struct sta_mld *mld;
	int i;

	if (!s->mld)
		return -1;

	mld = s->mld;
	for (i = 0; i < mld->num_affiliated_sta; i++) {
		int j;

		if (memcmp(mld->affiliated_sta[i]->macaddr, s->macaddr, 6))
			continue;

		for (j = (i + 1); j < mld->num_affiliated_sta; j++)
			mld->affiliated_sta[j-1] = mld->affiliated_sta[j];

		mld->affiliated_sta[j] = NULL;
		mld->num_affiliated_sta--;
	}

	if (mld->num_affiliated_sta == 0) {
		list_del(&mld->list);
		free(mld);
	}

	return 0;
}

/* remove passed affiliated backhaul STA from mld */
int wifi_sta_mld_del_bsta(struct agent *a, struct netif_bk *bk)
{
	struct bsta_mld *mld;
	int i, j;

	/* bk->mld may be NULL if bSTAMLD is not connected (Eth backhaul) */
	if (!bk || bk->cfg->is_mld_netdev || !bk->mld)
		return -1;

	mld = bk->mld;
	for (i = 0; i < mld->num_affiliated_bsta; i++) {
		if (!memcmp(mld->affiliated_bsta[i]->macaddr, bk->macaddr, 6))
			break;
	}

	if (i == mld->num_affiliated_bsta)
		return -1;

	for (j = (i + 1); j < mld->num_affiliated_bsta; j++)
		mld->affiliated_bsta[j-1] = mld->affiliated_bsta[j];

	mld->affiliated_bsta[j] = NULL;
	mld->num_affiliated_bsta--;

	bk->mld = NULL;
	bk->is_affiliated_sta = false;

	return 0;
}

/* remove passed affiliated AP from mld */
static int wifi_mld_del_ap(struct netif_ap *ap)
{
	struct netif_mld *mld;
	int i;

	if (!ap->is_affiliated_ap)
		return -1;

	if (!ap->mld) {
		warn("|%s:%d| MLD for given netif is not assigned!\n",
		     __func__, __LINE__);
		return -1;
	}

	mld = ap->mld;

	for (i = 0; i < mld->num_affiliated_ap; i++) {
		int j;

		if (memcmp(mld->affiliated_ap[i]->bssid, ap->bssid, 6))
			continue;

		for (j = (i + 1); j < mld->num_affiliated_ap; j++)
			mld->affiliated_ap[j-1] = mld->affiliated_ap[j];

		mld->affiliated_ap[j] = NULL;
		mld->num_affiliated_ap--;
	}
	ap->mld = NULL;
	ap->is_affiliated_ap = false;

	return 0;
}
#endif

static int wifi_del_sta(struct agent *a, const char *ifname,
		uint8_t *macaddr)
{
	struct netif_ap *ap;
	struct sta *s, *tmp;

	ap = agent_get_ap_by_ifname(a, ifname);
	if (!ap)
		return -1;

	list_for_each_entry_safe(s, tmp, &ap->stalist, list) {
		if (!memcmp(s->macaddr, macaddr, 6)) {
			dbg("Delete STA " MACFMT "record\n",
						MAC2STR(macaddr));
			if (ap->num_sta > 0)
				 ap->num_sta--;
			timer_del(&s->sta_finalize_timer);
			timer_del(&s->sta_steer_timer);
			timer_del(&s->sta_bcn_req_timer);
			timer_del(&s->sta_timer);
			timer_del(&s->sta_steer_deauth_timer);
			list_del(&s->list);
			list_flush(&s->pref_nbrlist, struct pref_neighbor, list);
			list_flush(&s->sta_nbrlist, struct sta_neighbor, list);
			if (s->assoc_frame) {
				free(s->assoc_frame->frame);
				free(s->assoc_frame);
			}
#if (EASYMESH_VERSION >= 6)
			if (s->is_affiliated_sta)
				wifi_sta_mld_del_sta(a, s);
#endif
			free(s);
			return 0;
		}
	}
	return -2;
}

int wifi_topology_notification(struct agent *a, struct sta *s, char *ifname, uint8_t assoc_event, uint16_t reason)
{
	trace("%s: --->\n", __func__);
	struct cmdu_buff *cmdu;
	struct netif_ap *ap;

	ap = agent_get_ap_by_ifname(a, ifname);
	if (!ap)
		return -1;


	if (!assoc_event) {
		cmdu = agent_gen_client_disassoc(a, s, ap->bssid, reason);
		if (cmdu) {
			agent_send_cmdu(a, cmdu);
			cmdu_free(cmdu);
		}
	}

	cmdu = agent_gen_topology_notification(a, s, ap->bssid, assoc_event);
	if (!cmdu)
		return -1;

	agent_send_cmdu(a, cmdu);
	cmdu_free(cmdu);

	return 0;
}

static void wifi_chan_change_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = c;
	struct cmdu_buff *cmdu;
	static const struct blobmsg_policy ev_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[ARRAY_SIZE(ev_attr)];
	struct wifi_radio_element *radio;
	char *ifname, *event;
	uint32_t chan = 0;
	uint32_t bw = 0;

	blobmsg_parse(ev_attr, ARRAY_SIZE(ev_attr), tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	ifname = blobmsg_data(tb[0]);
	event = blobmsg_data(tb[1]);

	radio = agent_get_radio_with_ifname(a, ifname);
	if (!radio)
		return;

	if (!strcmp(event, "ap-chan-change")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "target-channel", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "target-width", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "reason", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !data[2]) {
			WARN_ON(1);
			return;
		}

		chan = atoi(blobmsg_data(data[0]));
		bw = atoi(blobmsg_data(data[1]));

		if (!strcmp(blobmsg_data(data[2]), "radar") ||
		    !strcmp(blobmsg_data(data[2]), "move-start"))
			timer_set(&radio->preference_report_timer, 1000);

	} else if (!strcmp(event, "csa-finished")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "channel", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "bandwidth", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !data[2]) {
			WARN_ON(1);
			return;
		}

		chan = atoi(blobmsg_data(data[0]));
		bw = atoi(blobmsg_data(data[1]));
	} else if (!strcmp(event, "ap-csa-finished")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
			[1] = { .name = "bandwidth", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];
		uint32_t freq;

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1]) {
			WARN_ON(1);
			return;
		}

		freq = blobmsg_get_u32(data[0]);
		bw = blobmsg_get_u32(data[1]);
		chan = f2c(freq);
	}

	if (!chan || !bw)
		return;

	if (radio->reported_channel == chan &&
	    radio->reported_bandwidth == bw)
		return;

	radio->current_channel = chan;
	radio->current_bandwidth = bw;

	radio->current_opclass = wifi_opclass_find_id_from_channel(&radio->opclass,
			radio->current_channel, radio->current_bandwidth);

	/* Finally send this to controller */
	dbg("[%s] oper_channel_response chan %d bw %d opclass %d\n",
	    ifname, radio->current_channel, radio->current_bandwidth,
	    radio->current_opclass);

	cmdu = agent_gen_oper_channel_response(a, radio, radio->current_channel,
					       radio->current_bandwidth, 0);
	if (WARN_ON(!cmdu))
		return;

	agent_send_cmdu(a, cmdu);
	cmdu_free(cmdu);

	radio->reported_channel = radio->current_channel;
	radio->reported_bandwidth = radio->current_bandwidth;
}

static void wifi_cac_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = c;
	static const struct blobmsg_policy ev_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[ARRAY_SIZE(ev_attr)];
	struct wifi_radio_element *radio;
	char *ifname, *event;

	if (!a->cfg.ap_follow_sta_dfs)
		return;

	blobmsg_parse(ev_attr, ARRAY_SIZE(ev_attr), tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	ifname = blobmsg_data(tb[0]);
	event = blobmsg_data(tb[1]);

	radio = agent_get_radio_with_ifname(a, ifname);
	if (!radio)
		return;

	if (!strcmp(event, "cac-end")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "success", .type = BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		if (strcmp(blobmsg_data(data[0]), "1"))
			return;

		timer_set(&a->dynbh_scheduler, 1 * 1000);
	}
}

bool is_channel_supported_by_radio(struct wifi_radio_element *re,
				   uint8_t opclass,
				   uint8_t channel)
{
	return wifi_opclass_id_channel_supported(&re->opclass, opclass, channel);
}

struct netif_bk *agent_get_netif_bk_with_channel(struct agent *a,
						 uint8_t opclass,
						 uint8_t channel)
{
	struct wifi_radio_element *re;
	struct netif_bk *best = NULL;
	uint8_t priority = -1;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_bk *bk = &re->bk;

		if (!re->has_bsta)
			continue;

		if (!is_channel_supported_by_radio(re, opclass, channel))
			continue;

		if (best && bk->cfg->priority < priority)
			continue;

		best = bk;
		priority = bk->cfg->priority;
	}

	return best;
}

static int wifi_parse_assoc_frame(struct agent *a, struct json_object *frameobj)
{
	struct wifi_assoc_frame *new, *old;
	struct sta *s;
	const char *framestr, *macaddr;
	int len;
	uint8_t *frame;

	trace("|%s:%d| parsing wifi frame\n", __func__, __LINE__);

	macaddr = json_get_string(frameobj, "macaddr");
	if (!macaddr)
		return -1;

	framestr = json_get_string(frameobj, "raw");
	if (!framestr)
		return -1;

	len = strlen(framestr);
	len = len / 2;
	frame = calloc(len, sizeof(uint8_t));
	if (!frame)
		return -1;

	if (!strtob((char *)framestr, len, frame))
		goto out_frame;

	new = calloc(1, sizeof(*new));
	if (!new)
		goto out_frame;

	new->frame = frame;
	hwaddr_aton(macaddr, new->macaddr);
	new->len = len;

	/* if frame for client exists in list, replace it */
	old = wifi_get_frame(a, new->macaddr);
	if (old) {
		list_del(&old->list);
		free(old->frame);
		free(old);
	}

	/* if sta exists attach to sta directly and exit*/
	s = agent_get_sta(a, new->macaddr);
	if (s) {
		if (s->assoc_frame) {
			free(s->assoc_frame->frame);
			free(s->assoc_frame);
		}
		s->assoc_frame = new;
		return 0;
	}

	list_add(&new->list, &a->framelist);

	return 0;
out_frame:
	free(frame);
	return -1;
}

static int get_frame_type(struct agent *a, struct json_object *frameobj)
{
	const char *framestr;
	int frame_type = -1;

	framestr = json_object_to_json_string(frameobj);
	if (!framestr)
		return -1;

	if (strstr(framestr, "reassoc"))
		frame_type = WIFI_FRAME_REASSOC_REQ;
	else if (strstr(framestr, "assoc"))
		frame_type = WIFI_FRAME_ASSOC_REQ;
	else if (strstr(framestr, "action"))
		frame_type = WIFI_FRAME_ACTION;

	return frame_type;
}

static void check_protocol(const char *framestr, uint8_t *protocol)
{
	uint8_t *frame;
	uint8_t category;
	uint8_t action;
	int offset = 0;
	int len;

	len = strlen(framestr);
	len = (len - 1)/2;
	frame = calloc(len, sizeof(uint8_t));
	if (!frame)
		return;

	if (!strtob((char *)framestr, len, frame))
		goto out;

	if (len < 26)
		goto out;

	/* 2: frame control,
	 * 2: duration,
	 * 6: destination addr,
	 * 6: source addr,
	 * 6: bssid,
	 * 2: seq control
	 */
	offset = 2 + 2 + 6 + 6 + 6 + 2;
	memcpy(&category, frame + offset, 1);
	memcpy(&action, frame + offset + 1, 1);

	if (category == 0x0a) {
		if (action == 0x06)
			*protocol = 0x02;
		else if (action == 0x1a)
			*protocol = 0x03;
	}

	/* TODO:check the action frame category &
	 * type for ANQP.
	 */

out:
	free(frame);
}

static int wifi_parse_all_frame(struct agent *a, const char *ifname,
		struct json_object *frameobj)
{
	trace("%s\n", __func__);
	const char *framestr;
	int stype;
	uint8_t protocol = 0xff;

	stype = get_frame_type(a, frameobj);
	if (stype == -1)
		return -1;

	switch (stype) {
	case WIFI_FRAME_REASSOC_REQ:
		protocol = 0x01;
		framestr = json_get_string(frameobj, "reassoc");
		break;
	case WIFI_FRAME_ASSOC_REQ:
		protocol = 0x00;
		framestr = json_get_string(frameobj, "assoc");
		break;
	case WIFI_FRAME_ACTION:
		framestr = json_get_string(frameobj, "action");
		check_protocol(framestr, &protocol);
		break;
	default:
		framestr = NULL;
		break;
	}

	if (!framestr)
		return -1;

	prepare_tunneled_message((void *)a, ifname, protocol, framestr);

	return 0;
}

static int wifi_process_action_frame(struct agent *a, const char *framestr)
{
	trace("%s\n", __func__);

	int len;
	uint8_t *frame;

	if (!framestr)
		return -1;

	len = strlen(framestr);
	len = len / 2; /* octets */
	frame = calloc(len, sizeof(uint8_t));
	if (!frame)
		return -1;

	if (!strtob((char *)framestr, len, frame))
		goto out_frame;

	if (frame) {
		struct action_frame_container *frame_ptr;
		struct sta *s;

		frame_ptr = (struct action_frame_container *) frame;

		s = agent_get_sta(a, frame_ptr->src);
		if (!s)
			goto out_frame;

		trace("|%s:%d| Action Frame belongs to category %d\n",
				__func__, __LINE__, frame_ptr->category_code);

		if (frame_ptr->category_code == WIFI_ACTION_RADIO_MEASUREMENT) {
			if (frame_ptr->u.rrm.action_code == WIFI_RRM_RADIO_MEASUREMENT_REPORT) {
				uint16_t head_len = frame_ptr->u.rrm.tags - frame;
				uint16_t tags_len = len - head_len;
				uint16_t pos = 0;
				int count = 0;

				if (!s->supports_bcnreport)
					s->supports_bcnreport = true;

				/* Count action frame elements (Measurement Reports) */
				while (pos < tags_len) {
					if (frame_ptr->u.rrm.tags[pos] != 0x27) /* EID 39 - Meas Report */
						warn("|%s:%d| Measurement report frame contains unsupported" \
							 "Element ID\n", __func__, __LINE__);
					/* +2 for tag_no & tag_len */
					pos += (frame_ptr->u.rrm.tags[pos+1] + 2);
					count++;
				}

				trace("|%s:%d| sending beacon metrics response to the cntlr\n",
						__func__, __LINE__);

				send_beacon_metrics_response(a, frame_ptr->src,
						count, frame_ptr->u.rrm.tags, tags_len);

				free(frame);
			} else {
				trace("|%s:%d| Action Frame is not a Radio Measurement Response\n",
					__func__, __LINE__);
			}
		}
	}

	return 0;
out_frame:
	free(frame);
	return -1;
}

static int wifi_process_rx_frame(struct agent *a, struct json_object *frameobj)
{
	trace("%s\n", __func__);
	const char *framestr = NULL;
	int stype;

	stype = get_frame_type(a, frameobj);
	if (stype == -1)
		return -1;

	if (stype == WIFI_FRAME_ACTION) {
		framestr = json_get_string(frameobj, "action");
		if (framestr) {
			dbg("|%s:%d| processing action frame %s\n",
					__func__, __LINE__, framestr);
			return wifi_process_action_frame(a, framestr);
		}
	}

	return -1;
}

#if (EASYMESH_VERSION >= 6)
/* frame starting from payload, skipping header */
int wifi_process_ml_ie(struct agent *a, struct sta_mld *stamld,
		       uint8_t *assocframe, int framelen)
{
	uint8_t *ml;
	uint8_t *ml_ctrl;
	uint8_t type;
	uint8_t presence_bmp[2];

	bool linkid_info_present;
	bool bss_param_chg_present;
	bool medium_sync_present;
	bool eml_caps_present;

	bool mld_caps_ops_present;
	uint8_t *common_info;
	size_t pos;

	ml = wifi_find_ie_ext(assocframe, framelen, IE_EXT_ML);
	if (!ml) {
		dbg("%s: frame did not include ML IE\n", __func__);
		return -1;
	}

	ml_ctrl = ml + 3;
	type = ml_ctrl[0] & 0x7;
	presence_bmp[0] = (ml_ctrl[0] & 0xf0) >> 4;
	presence_bmp[1] = ml_ctrl[1];

	if (type != 0) {
		dbg("%s: ML IE type non-zero\n", __func__);
		return -1;
	}

	linkid_info_present = !!(presence_bmp[0] & 0x1);
	bss_param_chg_present = !!(presence_bmp[0] & 0x2);
	medium_sync_present = !!(presence_bmp[0] & 0x4);
	eml_caps_present = !!(presence_bmp[0] & 0x8);

	mld_caps_ops_present = !!(presence_bmp[1] & 0x1);
	common_info = &ml_ctrl[2];
	pos = 1;

	pos += 6; /* skip apmld macaddr */

	if (linkid_info_present)
		pos += 1;

	if (bss_param_chg_present)
		pos += 1;

	if (medium_sync_present)
		pos += 2;

	if (eml_caps_present) {
		uint8_t *eml_caps = &common_info[pos];

		stamld->emlsr_enabled = !!(eml_caps[0] & BIT(0)) ? true : false;
		stamld->emlmr_enabled = !!(eml_caps[0] & BIT(7)) ? true : false;
		pos += 2;
	}

	if (mld_caps_ops_present) {
		uint8_t *mld_caps = &common_info[pos];
		uint8_t max_simlinks = (mld_caps[0] & 0x0f);

		pos += 2;
		if (max_simlinks >= 1) {
			stamld->nstr_enabled = false;
			stamld->str_enabled = true;
		} else {
			stamld->nstr_enabled = true;
			stamld->str_enabled = false;
		}
	}

	return 0;
}
#endif

static uint16_t wifi_process_pvid_assoc_frame(struct agent *a, uint8_t *frame,
					    int len)
{
	int i;
	uint16_t pvid = 0;

	for (i = 0; i < len; i++) {
		uint8_t *p = frame + i;
		uint8_t elem_len = 0;
		uint8_t wfa_oui[4] = {0x50, 0x6f, 0x9a, 0x1b};

		if (*p++ != 0xdd) /* IEEE 802.1 vendor specific element */
			continue;

		elem_len = *p;

		if (elem_len < 14) { /* At least 14 bytes long */
			i += elem_len;
			continue;
		}

		if (elem_len > len - (i + 2)) {
			dbg("|%s:%d| WFA Element len exceeds raw frame len!\n", __func__, __LINE__);
			break;
		}


		p++; /* len */
		if (memcmp(p, wfa_oui, 4)) {
			i += elem_len;
			continue;
		}

		p += 4; /* wfa oui */

		if (*p++ != 0x06)  { /* multi-ap ext */
			i += elem_len;
			continue;
		}

		p += *p; /* subelem len */
		p++; /* len */


		if (*p++ != 0x07) { /* multi-ap profile */
			i += elem_len;
			continue;
		}

		p += *p; /* subelem len */
		p++; /* len */

		if (*p++ != 0x08) { /* multi-ap def 8021q */
			i += elem_len;
			continue;
		}

		p++; /* len */
		memcpy(&pvid, p, 2);
		dbg("|%s:%d| Found 8021q PVID %d\n", __func__, __LINE__, pvid);
		break;
	}

	return pvid;
}

#define ASSOC_TIMEOUT	60
static void wifi_sta_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = (struct agent *)c;
	char ifname[16] = {0}, event[24] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add = false, del = false;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
/*
{ "wifi.sta": {"ifname":"wlan0","event":"connected","data":{"macaddr":"16:bc:09:81:ec:48","mlo_macaddr":"90:09:df:e9:82:e3","mlo_bssid":"00:03:7f:12:69:69","bssid":"00:03:7f:12:6a:6a","mlo_link_id":1}} }
 */
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
#if (EASYMESH_VERSION >= 6)
			[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "mlo_bssid", .type = BLOBMSG_TYPE_STRING },
			[3] = { .name = "mlo_macaddr", .type = BLOBMSG_TYPE_STRING },
			[4] = { .name = "mlo_link_id", .type = BLOBMSG_TYPE_INT32 },
#endif
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];
		uint8_t macaddr[6] = {0};
		struct sta *s;
#if (EASYMESH_VERSION >= 6)
		uint8_t mlo_macaddr[6] = {0};
		struct netif_ap *ap = NULL;
		uint8_t bssid[6] = {0};
#endif

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		if (!hwaddr_aton(blobmsg_data(data[0]), macaddr))
			return;

#if (EASYMESH_VERSION >= 6)
		if (data[1]) {
			if (!hwaddr_aton(blobmsg_data(data[1]), bssid))
				return;

			ap = agent_get_ap(a, bssid);
			if (!ap) {
				/* could also do link-id check here? */
				return;
			}
			strncpy(ifname,	ap->ifname, sizeof(ifname) - 1);
		}
#endif

#ifndef UNAUTHORIZED_STA_IN_ASSOCLIST
		if (add)
			s = wifi_add_sta(a, ifname, macaddr);
		else
			s = agent_get_sta(a, macaddr);
		if (!s) {
			dbg("%s: Failed to add STA or STA has already disconnected\n", __func__);
			return;
		}
#endif

#if (EASYMESH_VERSION >= 6)
		/* Affiliated STA MAC in-case of MLO client*/
		if (add && data[2] && data[3] && data[4]) {
			uint8_t linkid = 0;

			linkid = (uint8_t) blobmsg_get_u32(data[4]);

			if (ap && ap->mld) {
				uint8_t mlo_bssid[6] = {0};
				struct sta_mld *mld;
				int num_sta = 0;

				if (!hwaddr_aton(blobmsg_data(data[3]), mlo_macaddr))
					return;
				if (!hwaddr_aton(blobmsg_data(data[2]), mlo_bssid))
					return;
				mld = agent_get_stamld(a, mlo_macaddr);
				if (!mld) {
					mld = agent_alloc_stamld(a, ap->mld, mlo_macaddr);
					if (!mld) {
						err("%s: Failed to allocate MLD:"MACFMT"\n", __func__, MAC2STR(mlo_macaddr));
						return;
					}
				}

				memcpy(mld->bssid, mlo_bssid, 6);

				/* TODO: code is copy pasted from mlsta_update_sta_mld() */
				num_sta = mld->num_affiliated_sta;

				s->mld = mld;
				s->is_affiliated_sta = true;
				s->mlo_link_id = linkid;

				if (!agent_is_sta_in_sta_mld(a, mld, macaddr) &&
					num_sta < MAX_AFFILIATED_STA_LINKS_NUM) {
					mld->affiliated_sta[num_sta] = s;
					mld->num_affiliated_sta++;
				}

				if (s->assoc_frame && s->assoc_frame->len > 4) {
					wifi_process_ml_ie(a, mld,
						s->assoc_frame->frame + 4,
						s->assoc_frame->len - 4);
				}
			}
		}
#endif

#ifndef UNAUTHORIZED_STA_IN_ASSOCLIST
		if (add)
			wifi_topology_notification(a, s, ifname, 0x01 /* joined */,
							0x00 /* success */);
		else {
			wifi_topology_notification(a, s, ifname, 0x00 /* left */,
							0x01 /* unspecified reason */);
			wifi_del_sta(a, ifname, macaddr);
		}
#endif

	} else if (!strcmp(event, "deauth")) {
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "reason", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];
		struct agent_config *cfg = &a->cfg;
		struct policy_cfg *pcfg;
		uint16_t reason = 1;
		uint8_t macaddr[6] = {};
		struct sta *s;

		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !cfg)
			return;

		pcfg = cfg->pcfg;
		if (!pcfg)
			return;

		if (!hwaddr_aton(blobmsg_data(data[0]), macaddr))
			return;

		s = agent_get_sta(a, macaddr);
		if (!s) {
			return;
		}


		reason = blobmsg_get_u32(data[1]);

		/* Send failed connection message if the map-agent
		 * has sent fewer than the maximum number of Send Failed
		 * Connection messages (as specified in the
		 * Maximum Reporting Rate element of the Unsuccessful
		 * Association Policy TLV) in the preceding minute
		 */
		if (pcfg->report_sta_assocfails) {
			if (timestamp_expired(&a->last_unassoc,
						ASSOC_TIMEOUT * 1000)) {

				timestamp_update(&a->last_unassoc);
				a->unassoc_cnt = 1;
				send_failed_connection_msg(a, macaddr, 0, reason);
			} else {
				if (a->unassoc_cnt < pcfg->report_sta_assocfails_rate) {
					a->unassoc_cnt++;
					send_failed_connection_msg(a, macaddr, 0, reason);
				}
			}
		}
#ifdef UNAUTHORIZED_STA_IN_ASSOCLIST
		wifi_topology_notification(a, s, ifname, 0, reason);
		wifi_del_sta(a, ifname, macaddr);
#endif
	} else if (!strcmp(event, "btm-resp")) {
		struct blob_attr *data[3];
		static const struct blobmsg_policy data_attr[3] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "target_bssid", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0}, bssid_str[18] = {0};
		uint8_t mac[6] = {0}, bssid[6] = {0};
		int status;
		char ev_data[512] = {0};

		blobmsg_parse(data_attr, 3, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[2])
			return;

		/* macaddr */
		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);
		if (!hwaddr_aton(mac_str, mac))
			return;

		/* target bssid - may be empty in case of failure */
		if (data[1]) {
			strncpy(bssid_str, blobmsg_data(data[1]),
					sizeof(bssid_str) - 1);
			if (!hwaddr_aton(bssid_str, bssid))
				return;
		}

		/* status */
		status = atoi(blobmsg_data(data[2]));

		/* notify event received */
		snprintf(ev_data, sizeof(ev_data),
				"{\"macaddr\":\""MACFMT"\""
				",\"dst_bssid\":\""MACFMT"\"}"
				",\"status\":%d}",
				MAC2STR(mac),
				MAC2STR(bssid),
				status);

		agent_notify_iface_event(a, ifname, "btm-resp", ev_data);

		wifi_send_sta_report(a, ifname, mac, status, bssid);

		trace("%s: btm-resp for "MACFMT" status %d\n",
			  __func__, MAC2STR(mac), status);

		/* Generally after successful steering station is disassociated from source BSSID,
		 * but sometimes it happens that station is not disassociated from source BSSID
		 * and is associated to targed BSSID at the same time. To avoid that situation
		 * we start timer here to disconnect wifi station when it left on source BSSID.
		 * When driver removes station after successful steer then timer will never end
		 * because it will be deleted together with station, but when timer hits timeout
		 * then it means station was not disconnected and funtion wifi_sta_steered_deauth
		 * will do this.
		 */
		if (!status) {
			struct netif_ap *ap = NULL;
			struct sta *s = NULL;

			ap = agent_get_ap_by_ifname(a, ifname);
			if (!ap) {
				warn("%s: Unable to find iface %s\n", __func__, ifname);
				return;
			}

			list_for_each_entry(s, &ap->stalist, list) {
				if (!memcmp(s->macaddr, mac, 6)) {
					timer_set(&s->sta_steer_deauth_timer, STA_STEERED_DEAUTH_INTERVAL * 1000);
					break;
				}
			}
		} else {
			/* TODO:
			 * update reject counter and retry steer later
			 */
		}
	} else if (!strcmp(event, "bcn-report")) {
		struct blob_attr *data[4];
		static const struct blobmsg_policy data_attr[4] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "token", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "mode", .type = BLOBMSG_TYPE_STRING },
			[3] = { .name = "nbr", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0};
		uint8_t sta_addr[6] = {0};
		struct sta *s;
		uint8_t mode = 0;

		blobmsg_parse(data_attr, 4, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1] || !data[2] || !data[3])
			return;

		mode = atoi(blobmsg_data(data[2]));

		if (mode) /* Measurement Report Mode should be 0x00 */
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, sta_addr))
			return;

		s = agent_get_sta(a, sta_addr);
		if (!s)
			return;

		if (!s->supports_bcnreport)
			s->supports_bcnreport = true;
	} else if (!strcmp(event, "wds-station-added") ||
		   !strcmp(event, "wds-station-removed")) {
		struct blob_attr *data[1];
		static const struct blobmsg_policy data_attr[1] = {
			[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		};
		char wdsname[16] = {0};
		struct ts_context *ts = &a->ts;
		struct wifi_radio_element *re;
		bool add = !strcmp(event, "wds-station-added");

		if (a->cfg.brcm_setup)
			return;

		blobmsg_parse(data_attr, 1, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(wdsname, blobmsg_data(data[0]), sizeof(wdsname) - 1);

		if (add)
			netif_alloc_wds(a, wdsname);
		else
			netif_free_wds_by_ifname(a, wdsname);

		list_for_each_entry(re, &a->radiolist, list) {
			struct netif_ap *ap = NULL;

			list_for_each_entry(ap, &re->aplist, list) {
				uint16_t vid;

				if (!ap->cfg)
					continue;

				vid = ap->cfg->vid;
				if (!is_vid_valid(vid))
					vid = ts->primary_vid;

				if (a->cfg.guest_isolation && (vid != 1 && vid != ts->primary_vid)) {
					dbg("/lib/wifi/multiap ts isolate_wds %s %s %s\n", (add ? "add" : "del"), ap->ifname, wdsname);
					runCmd("/lib/wifi/multiap ts isolate_wds %s %s %s", (add ? "add" : "del"), ap->ifname, wdsname);
				}
			}
		}
	} else if (!strcmp(event, "action")) {
		struct blob_attr *data[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "raw", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0};
		uint8_t bbss_addr[6] = {0};
		struct netif_bk *bk;
		char *framestr;
		uint8_t *frame;
		int len = 0;

		blobmsg_parse(data_attr, 2, data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!data[0] || !data[1])
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, bbss_addr))
			return;

		bk = agent_get_bsta_by_ifname(a, ifname);
		if (!bk)
			return;

		framestr = (char *)blobmsg_data(data[1]);

		len = strlen(framestr);
		len = (len - 1) / 2;
		frame = calloc(len, sizeof(uint8_t));
		if (!frame)
			return;

		if (!strtob(framestr, len, frame)) {
			free(frame);
			return;
		}

		if (*frame == 0x09 || *frame == 0x04) {
			struct wifi_radio_element *r;
			uint8_t channel = 0;
			uint8_t opclass = 0;
			uint8_t *ptr = frame;

			ptr++;

			if (*ptr == 0x04) {
				ptr += 2;
				opclass = *ptr;

				ptr++;
				channel = *ptr;

				r = agent_get_radio_with_ifname(a, ifname);
				if (!r) {
					free(frame);
					return;
				}

				if (is_channel_supported_by_radio(r, opclass, channel)) {
					dbg("|%s:%d| bsta %s supports new "\
					     "channel, no action required\n",
					     __func__, __LINE__, bk->ifname);
					free(frame);
					return;
				}

				bk = agent_get_netif_bk_with_channel(a, opclass, channel);
				if (bk) {
					dynbh_bsta_swap_to_link(a, bk->ifname);
					agent_config_reload(a);
				} else
					dynbh_bsta_enable_all(a);
			}
		}

		free(frame);
	} else if (!strcmp(event, "assoc") || !strcmp(event, "reassoc")) {
		struct json_object *jmsg, *data;
		char *str;
		struct blob_attr *bdata[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "raw", .type = BLOBMSG_TYPE_STRING },
		};
		char *framestr;
		uint8_t *frame;
		int len;
		uint16_t pvid;
#ifdef UNAUTHORIZED_STA_IN_ASSOCLIST
		uint8_t macaddr[6] = {0};
		struct sta *s;
#endif

		blobmsg_parse(data_attr, 2, bdata, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!bdata[0] || !bdata[1])
			return;

		framestr = (char *)blobmsg_data(bdata[1]);

		len = strlen(framestr);
		len = (len - 1) / 2;
		frame = calloc(len, sizeof(uint8_t));
		if (!frame)
			return;

		if (!strtob(framestr, len, frame)) {
			free(frame);
			return;
		}

		pvid = wifi_process_pvid_assoc_frame(a, frame, len);
		if (pvid) {
			agent_fill_8021q_setting(a, pvid, 0);
			agent_apply_traffic_separation(a);
		}

		free(frame);

#ifdef UNAUTHORIZED_STA_IN_ASSOCLIST
		hwaddr_aton(blobmsg_get_string(bdata[0]), macaddr);
		s = wifi_add_sta(a, ifname, macaddr);
		if (!s)
			return;
		wifi_topology_notification(a, s, ifname,
					   0x01 /* client joined */,
					   0x00 /* success */);
#endif
		trace("%s: ------------>\n", __func__);

		str = blobmsg_format_json(msg, true);
		if (!str)
			return;

		jmsg = json_tokener_parse(str);
		if (!jmsg)
			goto out_str;

		if (!json_object_is_type(jmsg, json_type_object))
			goto out_json;

		json_object_object_get_ex(jmsg, "data", &data);
		if (!data)
			goto out_json;

		wifi_parse_assoc_frame(a, data);

		/* parse (re-)assoc message
		*/
		wifi_parse_all_frame(a, ifname, data);

out_json:
		json_object_put(jmsg);
out_str:
		free(str);
	} else if (!strcmp(event, "assoc-failure")) {
		/* { "wifi.sta": {"ifname":"wl0","event":"assoc-failure","data":{"macaddr":"00:0a:52:06:2e:b9", "status":12}} } */
		struct blob_attr *bdata[2];
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "status", .type = BLOBMSG_TYPE_INT32 },
		};
		uint8_t macaddr[6] = {0};
		struct agent_config *cfg = &a->cfg;
		struct policy_cfg *pcfg;
		struct sta *s;
		int status, status_code = 0;

		blobmsg_parse(data_attr, 2, bdata, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));

		if (!bdata[0] || !bdata[1] || !cfg)
			return;

		pcfg = cfg->pcfg;
		if (!pcfg)
			return;

		status = blobmsg_get_u32(bdata[1]);
		if (status == 12) /* mac list deny */
			status_code = 33; /* QoS AP lacks sufficient bandwidth */

		hwaddr_aton(blobmsg_get_string(bdata[0]), macaddr);

		s = agent_get_sta(a, macaddr);
		if (!s)
			return;

		if (pcfg->report_sta_assocfails) {
			if (timestamp_expired(&a->last_unassoc,
					      ASSOC_TIMEOUT * 1000)) {
				timestamp_update(&a->last_unassoc);
				a->unassoc_cnt = 1;
				send_failed_connection_msg(a, macaddr, status_code, 0);
			} else {
				if (a->unassoc_cnt < pcfg->report_sta_assocfails_rate) {
					a->unassoc_cnt++;
					send_failed_connection_msg(a, macaddr, status_code, 0);
				}
			}
		}
		wifi_topology_notification(a, s, ifname, 0, 0x01);
		wifi_del_sta(a, ifname, macaddr);
	}
#ifdef EASYMESH_VENDOR_EXT
	else if (!strcmp(event, "probe-req")) {
		struct netif_ap *iface = agent_get_ap_by_ifname(a, ifname);
		struct netif_bk *bk_iface = NULL;
		static const struct blobmsg_policy data_attr[2] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "rssi", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *bdata[2];
		struct cmdu_buff *cmdu;
		char mac_str[18] = {0};
		uint8_t mac[6] = {0};

		if (iface == NULL) {
			bk_iface = agent_get_bsta_by_ifname(a, ifname);
			if (bk_iface == NULL) {
				return;
			}
		}

		blobmsg_parse(data_attr, 2, bdata, blobmsg_data(tb[2]),
			blobmsg_data_len(tb[2]));

		if (!bdata[0] || !bdata[1])
			return;

		strncpy(mac_str, blobmsg_data(bdata[0]), sizeof(mac_str) - 1);
		hwaddr_aton(mac_str, mac);

		/* send vendor probe request to map controller */
		cmdu = agent_gen_vendor_specific_probe_req(a,
			mac, iface != NULL ? iface->bssid : bk_iface->macaddr,
			rssi_to_rcpi((int32_t)blobmsg_get_u32(bdata[1])));
		if (cmdu) {
			agent_send_cmdu(a, cmdu);
			cmdu_free(cmdu);
		}
	}
#endif
}

static void agent_iface_disconnect_all_sta(const struct netif_ap *iface,
					   uint16_t reason)
{
	struct sta *sta;

	list_for_each_entry(sta, &iface->stalist, list) {
		wifi_disconnect_sta(iface->ifname, sta->macaddr, reason);
	}

}

static void agent_init_ifaces_assoc_mode(struct agent *a)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap;

		list_for_each_entry(ap, &re->aplist, list) {
			if (ap->cfg->disallow_assoc) {
				agent_iface_disconnect_all_sta(ap,
					WIFI_REASON_SERVICE_NOT_AUTHORIZED_IN_THIS_LOCATION);
			}

			wifi_set_backhaul_bss_association_mode(ap->ifname,
							ap->cfg->disallow_assoc);
		}
	}
}

static void apply_iface_assocation_mode(struct agent *a, const char *ifname)
{
	struct netif_ap *ap = agent_get_ap_by_ifname(a, ifname);

	if (ap) {
		wifi_set_backhaul_bss_association_mode(ifname,
						       ap->cfg->disallow_assoc);
	}
}


static void wifi_iface_event_handler(void *c, struct blob_attr *msg)
{
	struct agent *a = (struct agent *)c;
	const char *ifname, *event;
	struct json_object *jmsg, *data;
	char *str;

	trace("%s: ------------>\n", __func__);

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	jmsg = json_tokener_parse(str);
	if (!jmsg)
		goto out_str;

	if (!json_object_is_type(jmsg, json_type_object))
		goto out_json;

	json_object_object_get_ex(jmsg, "data", &data);
	if (!data)
		goto out_json;

	ifname = json_get_string(jmsg, "ifname");
	if (!ifname)
		goto out_json;

	event = json_get_string(jmsg, "event");
	if (!event)
		goto out_json;

	if (!strcmp(event, "assoc") || !strcmp(event, "reassoc")) {
		wifi_parse_assoc_frame(a, data);

		/* parse (re-)assoc message
		 */
		wifi_parse_all_frame(a, ifname, data);
	} else if (!strcmp(event, "frame-rx")) {
		wifi_process_rx_frame(a, data);
	} else if (!strcmp(event, "action")) {
		/* parse action frame
		 * BTM, WNM, ANQP
		 */
		wifi_parse_all_frame(a, ifname, data);
	} else if (!strcmp(event, "bss-up")) {
		apply_iface_assocation_mode(a, ifname);
	}

out_json:
	json_object_put(jmsg);
out_str:
	free(str);
}

static void wifi_channel_event_handler(void *c, struct blob_attr *msg)
{
	/* struct agent *a = (struct agent *)c; */
	const char *ifdev, *res;
	int ch = 0;
	struct json_object *jmsg;
	char *str;

	trace("%s: ------------>\n", __func__);
	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	jmsg = json_tokener_parse(str);
	if (!jmsg)
		goto out_str;

	if (!json_object_is_type(jmsg, json_type_object))
		goto out_json;

	ifdev = json_get_string(jmsg, "radio");
	if (!ifdev)
		goto out_json;

	ch = json_get_int(jmsg, "channel");
	UNUSED(ch);

	res = json_get_string(jmsg, "reason");
	if (!res)
		goto out_json;

	if (!strcmp(res, "radar")) {
		fprintf(stderr, "TODO handle ch change due to radar!\n");
		/* wifi_handle_channel_event(a, ifdev, ch, res); */
	}

out_json:
	json_object_put(jmsg);
out_str:
	free(str);
}

uint8_t wifi_noise_to_anpi(int n)
{
	if (!n)
		return 255;
	if (n < -110)
		return 0;
	if (n > 0)
		return 220;
	return (n + 110) * 2;
}

static bool may_enable_fhs(struct agent *a)
{
#ifdef AGENT_ISLAND_PREVENTION
	if (a->cfg.island_prevention)
		return agent_has_active_backhaul(a) ? true : false;
#endif
	return agent_has_active_backhaul(a) ? true : false;
}

static void wifi_radio_handle_survey_evt(struct wifi_radio_element *r,
		struct blob_attr *tb_data)
{
	dbg("%s --->\n", __func__);

	struct blob_attr *data[3];
	static const struct blobmsg_policy supp_attrs[3] = {
		[0] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "noise", .type = BLOBMSG_TYPE_INT32 },
		[2] = { .name = "utilization", .type = BLOBMSG_TYPE_INT32 },
	};
	uint32_t cur_ch;
	int dBm;
	int i;

	if (!tb_data)
		return;

	if (!r) {
		dbg("%s: missing radio_element !\n", __func__);
		return;
	}

	if (!r->scanlist) {
		dbg("%s: missing scanlist!\n", __func__);
		return;
	}

	blobmsg_parse(supp_attrs, 3, data,
		blobmsg_data(tb_data), blobmsg_data_len(tb_data));

	if (!data[0] || !data[1] || !data[2]) {
		dbg("|%s:%d| Wifi event parse error.\n", __func__, __LINE__);
		return;
	}
	cur_ch = (uint8_t)blobmsg_get_u32(data[0]);

	for (i = 0; i < r->scanlist->num_opclass; i++) {
		struct wifi_scanres_opclass_element *op;
		int j;

		op = r->scanlist->opclass_scanlist + i;

		for (j = 0; j < op->num_channels; j++) {
			struct wifi_scanres_channel_element *ch;

			if (op->channel_scanlist[j].channel == cur_ch) {
				uint32_t utilization;

				ch = op->channel_scanlist + j;
				dBm = (int8_t)blobmsg_get_u32(data[1]);
				ch->anpi = wifi_noise_to_anpi(dBm);
				utilization = blobmsg_get_u32(data[2]);
				ch->utilization = (uint8_t)((utilization * 255) / 100);
				return;
			}
		}
	}
}

static void wifi_radio_event_handler(void *c, struct blob_attr *msg)
{
	dbg("%s --->\n", __func__);

	struct agent *a = (struct agent *) c;
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[6] = {
		[0] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	char ifname[18] = {0}, event[64] = {0};

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1])
		return;

	strncpy(ifname, blobmsg_data(tb[1]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[0]), sizeof(event) - 1);

	dbg("%s: Handling %s for %s\n", __func__, event, ifname);

	if (!strcmp(event, "ap-enabled")) {
		/* TODO: do this on event-basis, meaning don't teardown
		* everything and rebuild everything
		*/
		/* completely re-init data */
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
		if (may_enable_fhs(a))
			/* Enable fronthauls */
			timer_set(&a->enable_fhs_scheduler,
					(IFACE_TIMEOUT + 1) * 1000);

	} else if (!strcmp(event, "ap-disabled")) {
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
	} else if (!strcmp(event, "ap-updated")) {
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
	} else if (!strcmp(event, "radar")) {
		struct cmdu_buff *cmdu;

		cmdu = agent_gen_channel_preference_report(a, NULL);
		if (cmdu) {
			agent_send_cmdu(a, cmdu);
			cmdu_free(cmdu);
		}
	} else if (!strcmp(event, "scan_finished")) {
		struct wifi_radio_element *re =
			agent_get_radio_with_ifname(a, ifname);

		if (WARN_ON(!re)) {
			dbg("%s: Current radio_element not found.\n", __func__);
			return;
		}
		re->scan_state = SCAN_DONE;

		handle_wifi_radio_scan_finished(a, re);
	} else if (!strcmp(event, "scan_aborted")) {
		struct wifi_radio_element *re =
			agent_get_radio_with_ifname(a, ifname);

		if (WARN_ON(!re)) {
			dbg("%s: Current radio_element not found.\n", __func__);
			return;
		}
		re->scan_state = SCAN_CANCELED;

		handle_wifi_radio_scan_finished(a, re);
	} else if (!strcmp(event, "survey")) {
		struct wifi_radio_element *re;

		if (!tb[2])
			return;

		re = agent_get_radio_with_ifname(a, ifname);
		if (WARN_ON(!re)) {
			dbg("%s: Current radio_element not found.\n", __func__);
			return;
		}

		wifi_radio_handle_survey_evt(re, tb[2]);
	} else
		dbg("%s: Unhandled wifi event %s!\n", __func__, event);

}

int wifi_mod_bridge(struct agent *a, struct netif_bk *bk, bool add)
{
	int ret = -1;
	char *ifname = bk->ifname;

	/* add wds iface to bridge */

	dbg("|%s:%d| %s interface %s to bridge %s using libeasy API\n", __func__,
	    __LINE__, (add ? "Adding" : "Deleting"),
	    ifname, a->cfg.al_bridge);


	if (add /* && !if_isbridge_interface(ifname) */) {
		dynbh_update_bk_connect_t(a, bk);

		if (!if_isbridge_interface(ifname))
			ret = br_addif(a->cfg.al_bridge, ifname);
	}
	else if(!add /* && if_isbridge_interface(ifname) */) {
		dynbh_update_bk_disconnect_t(a, bk);

		if (if_isbridge_interface(ifname))
			ret = br_delif(a->cfg.al_bridge, ifname);
	}

	/* explicitly disable if removing from bridge */
	if (!ret && !add) {
		int rc;

		rc = wifi_set_4addr(ifname, false);
		dbg("|%s:%d| Disabled 4addr mode for %s (rc:%d)\n", __func__,
				__LINE__, ifname, rc);
	}

	return ret;
}

static void agent_reload_local_cntlr(struct agent *a, bool on)
{
	set_value_by_string("mapcontroller", "controller",
			"enabled", on ? "1" : "0", UCI_TYPE_STRING);
	trace("Reloading mapcontroller\n");

	/* procd is too slow to send signal - manual kill prior to reload
	 * TODO: leverage pidfile
	 */
	runCmd("kill `pidof mapcontroller`");
	uci_reload_services("mapcontroller");

	/* Enable is now set to '1' in config; ensure service is started */
	if (on) {
		runCmd("/etc/init.d/mapcontroller reload");
	}
}

static void agent_enable_local_cntlr(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, cntlr_scheduler);

	dbg("|%s:%d| Attempting to start local controller\n", __func__, __LINE__);

	if (is_local_cntlr_running()) {
		warn("Skip starting local mapcontroller: already running\n");
	} else {
		agent_reload_local_cntlr(a, true);
	}
}

void agent_disable_local_cntlr(struct agent *a)
{
	dbg("|%s:%d| Disable local controller in UCI\n", __func__, __LINE__);
	agent_reload_local_cntlr(a, false);
}

static void agent_schedule_cntlr_start(struct agent *a, int secs)
{
	dbg("|%s:%d| Scheduled controller start in %d sec\n",
			__func__, __LINE__, secs);
	timer_set(&a->cntlr_scheduler, secs * 1000);
}

static void agent_rcpi_thresold_timer_cb(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent,
			rcpi_threshold_timer);
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			struct agent_config_radio *rcfg;
			struct sta *s;

			dbg("%s %d checking clients on %s\n", __func__, __LINE__, ap->ifname);

			if (!ap->cfg)
				continue;

			rcfg = get_agent_config_radio(&a->cfg, ap->cfg->device);
			if (!rcfg)
				continue;

			list_for_each_entry(s, &ap->stalist, list) {
				uint8_t report_rcpi = rcfg->report_rcpi_threshold;
				uint8_t curr_rcpi, prev_rcpi;

				dbg("%s %d found client "MACFMT"\n", __func__, __LINE__, MAC2STR(s->macaddr));

				/* Note: Using average RSSI for all triggers */
				curr_rcpi = rssi_to_rcpi(s->rssi_avg[0]);
				prev_rcpi = rssi_to_rcpi(s->rssi_avg[1]);

				if ((curr_rcpi < report_rcpi) ||
				    (curr_rcpi > report_rcpi && prev_rcpi <= report_rcpi)) {
					struct cmdu_buff *cmdu;
					char ev_data[512] = {0};

					snprintf(ev_data, sizeof(ev_data),
						"{\"macaddr\":\""MACFMT"\""
						",\"report_rcpi_threshold\":%d"
						",\"rcpi\":%d}",
						MAC2STR(s->macaddr),
						rcfg->report_rcpi_threshold, curr_rcpi);

					if (curr_rcpi < report_rcpi)
						agent_notify_iface_event(a, ap->ifname,
								"sta_low_rcpi", ev_data);
					else
						agent_notify_iface_event(a, ap->ifname,
								"sta_high_rcpi", ev_data);

					cmdu = agent_gen_assoc_sta_metric_responsex(a,
						a->cntlr_almac, s, ap);
					if (cmdu) {
						agent_send_cmdu(a, cmdu);
						cmdu_free(cmdu);
					}
				}

			}

		}
	}

/*	curr_rcpi = calculate_radio_rcpi(a, p->ifname);
	prev_rcpi = p->prev_rcpi;
	if (((prev_rcpi > p->cfg->rcpi_threshold) &&
				(curr_rcpi > p->cfg->rcpi_threshold)) ||
			((prev_rcpi < p->cfg->rcpi_threshold) &&
			 (curr_rcpi < p->cfg->rcpi_threshold)))
		goto refresh_interval;
*/
	//prepare_assoc_sta_metric_response(a, p->ifname);

//refresh_interval:
//	p->prev_rcpi = curr_rcpi;
	timer_set(&a->rcpi_threshold_timer, RCPI_THRESHOLD_TIMER);
}

static bool agent_ap_get_state(struct agent *a, struct netif_ap *ap)
{

	if (!ap)
		return -1;

	return ap->enabled ? true : false;
}

static int agent_ap_set_state(struct agent *a, struct netif_ap *ap, bool up)
{
	bool is_enabled;

	if (!ap)
		return -1;

	is_enabled = agent_ap_get_state(a, ap);

	dbg("|%s:%d| setting AP %s %s (current: %s)\n",
	    __func__, __LINE__,
	    ap->ifname, up ? "up" : "down",
	    is_enabled ? "up" : "down");

	if (is_enabled == up)
		return 0;

	if (up && ap->cfg && !ap->cfg->enabled) {
		/* start_disabled set for given ap in cfg */
		dbg("|%s:%d| will not enable ap %s - start_disabled set\n",
		    __func__, __LINE__, ap->ifname);
		return 0;
	}

	if (wifi_ap_set_state(ap->ifname, up)) {
		dbg("|%s:%d| failed to set AP %s %s\n",
		    __func__, __LINE__, ap->ifname, up ? "up" : "down");
		return -1;
	} /* check if the state has changed */ if (agent_ap_get_state(a, ap) != up) {
		dbg("|%s:%d| sheduling iface init\n", __func__, __LINE__);
		timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
	}

	return 0;
}

static void agent_enable_fronthauls(struct agent *a)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		dbg("|%s:%d| enabling all fronthauls\n", __func__, __LINE__);

		list_for_each_entry(ap, &re->aplist, list)
			agent_ap_set_state(a, ap, true);
	}
}

static int agent_req_btm_steer(struct agent *a, const char *ifname, unsigned char *sta_macaddr,
		uint8_t pref_num, struct nbr *pref, uint32_t validity_int)
{
	struct wifi_btmreq req = {};

	dbg("Calling wifi.ap.%s request_btm for sta " MACFMT "\n",
			ifname, MAC2STR(sta_macaddr));

	if (pref_num)
		req.mode |= WIFI_BTMREQ_PREF_INC;

	req.dialog_token = agent_steer_next_dialog_token(a);

	/* Inform remaining clients of imminent disassociation */
	/* disassoc_tmo kept as 0*/
	req.mode |= WIFI_BTMREQ_DISASSOC_IMM;

	/* BTM Abridged bit set to 1 */
	req.mode |= STEER_REQUEST_BTM_ABRIDGED;

	req.validity_int = validity_int / WIFI_BEACON_PERIOD_MS;

	if (wifi_req_btm(ifname, sta_macaddr, pref_num, pref, &req)) {
		err("|%s:%d|Failed calling wifi.ap.%s request_btm for sta " MACFMT "\n",
			__func__, __LINE__, ifname, MAC2STR(sta_macaddr));
		return -1;
	}

	return 0;
}

#ifdef AGENT_ISLAND_PREVENTION
/* Time in seconds until the AP sends a Disassociation frame to the STA */
#define STA_DISCONNECT_TM	30
static int agent_req_btm_bssterm(struct agent *a, const char *ap_name,
				 unsigned char *sta_macaddr,
				 uint32_t disassoc_tmo)
{
	struct wifi_btmreq req = {};

	dbg("Calling wifi.ap.%s request_btm for sta " MACFMT "\n",
			ap_name, MAC2STR(sta_macaddr));

	/* Inform remaining clients of imminent disassociation */
	req.mode |= WIFI_BTMREQ_DISASSOC_IMM;
	/* Inform clients about BSS termination */
	req.mode |= WIFI_BTMREQ_BSSTERM_INC;

	/* Duration - number of minutes for which the BSS is not present.
	 * Value of 0 is reserved. Value 65 535 when the BSS is terminated
	 * for a period longer than or equal to 65 535 minutes.
	 */
	req.bssterm_dur = 1; /* away (min) */
	req.disassoc_tmo = disassoc_tmo / WIFI_BEACON_PERIOD_MS;
	req.dialog_token = agent_steer_next_dialog_token(a);

	if (wifi_req_btm(ap_name, sta_macaddr, 0, NULL, &req)) {
		dbg("|%s:%d| failed to send wifi_req_btm\n", __func__, __LINE__);
		return -1;
	}

	return 0;
}

#define STA_DISCONNECT_TM 30
void agent_schedule_fh_disable(struct agent *a)
{
	trace("%s: --->\n", __func__);
	struct wifi_radio_element *re = NULL;
	struct sta *s;

	if (agent_has_active_backhaul(a))
		/* backhaul regained, don't send BTM */
		return;

	dbg("|%s:%d| will disable aps in %d seconds\n",
	    __func__, __LINE__, STA_DISCONNECT_TM + 2);
	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			/* Try to steer all remaining (b)STAs away */
			list_for_each_entry(s, &ap->stalist, list) {
				agent_req_btm_bssterm(a, ap->ifname, s->macaddr, STA_DISCONNECT_TM * 1000);
			}
		}
	}

	/* Disconnect all STAs in in STA_DISCONNECT_TM seconds */
	timer_set(&a->sta_disconnect_timer, STA_DISCONNECT_TM * 1000);
}

static int agent_sta_disconnect(struct agent *a, struct netif_ap *ap, struct sta *s)
{
	dbg("Calling wifi.ap.%s disconnect for sta " MACFMT "\n",
	    ap->ifname, MAC2STR(s->macaddr));

	return wifi_disconnect_sta(ap->ifname, s->macaddr,
			WIFI_REASON_BSS_TRANSITION_DISASSOC);
}

static void agent_fh_disable_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);
	struct agent *a = container_of(t, struct agent, fh_disable_timer);
	struct wifi_radio_element *re = NULL;

	if (agent_has_active_backhaul(a))
		/* backhaul regained, don't disable FHs */
		return;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;

		list_for_each_entry(ap, &re->aplist, list) {
			agent_ap_set_state(a, ap, false);
		}
	}
}

static void agent_sta_disconnnect_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);
	struct agent *a = container_of(t, struct agent, sta_disconnect_timer);
	struct wifi_radio_element *re = NULL;

	if (agent_has_active_backhaul(a))
		/* backhaul regained, don't disconnect STAs */
		return;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap = NULL;
		struct sta *s;

		list_for_each_entry(ap, &re->aplist, list) {
			list_for_each_entry(s, &ap->stalist, list) {
				agent_sta_disconnect(a, ap, s);
			}
		}
	}

	/* Disable all fronthauls in 2 seconds */
	timer_set(&a->fh_disable_timer, 2 * 1000);
}
#endif /* AGENT_ISLAND_PREVENTION */

int wifiagent_toggle_fh(struct ubus_object *obj, bool isl_prev,
		char *ifname, int enable)
{
	struct agent *a = container_of(obj, struct agent, obj);

#ifdef AGENT_ISLAND_PREVENTION
	if (isl_prev && a->cfg.island_prevention && !enable) {
		/* schedule bringing down all STAs & FHs */
		dbg("|%s:%d| island prevention: schedule FHs down\n",
		    __func__, __LINE__);
		agent_schedule_fh_disable(a);
		return 0;
	}
#else
	UNUSED(isl_prev);
#endif

	if (ifname && ifname[0]) {
		struct netif_ap *ap = NULL;

		/* enable/disable fronthaul(s)  */
		if (strncmp(ifname, "all", 16)) {
			ap = agent_get_ap_by_ifname(a, ifname);
			if (!ap) {
				dbg("interface \'%s\' not found\n", ifname);
				return -1;
			}
			agent_ap_set_state(a, ap, enable ? true : false);
		} else { /* all interfaces */
			struct wifi_radio_element *re = NULL;

			list_for_each_entry(re, &a->radiolist, list) {
				list_for_each_entry(ap, &re->aplist, list) {
					agent_ap_set_state(a, ap, enable ? true : false);
				}
			}
		}
	}

	return 0;
}

static void preference_report_timer_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct wifi_radio_element *re = container_of(t, struct wifi_radio_element, preference_report_timer);
	struct cmdu_buff *cmdu;

	cmdu = agent_gen_channel_preference_report(re->agent, NULL);
	if (cmdu) {
		agent_send_cmdu(re->agent, cmdu);
		cmdu_free(cmdu);
	}
}

bool agent_has_downstream(struct agent *a)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap;

		list_for_each_entry(ap, &re->aplist, list) {
			struct sta *s;

			if (!ap->cfg)
				continue;

			/* only bbss and combined type interfaces may have downstream APs */
			if (ap->cfg->multi_ap != 1 && ap->cfg->multi_ap != 3)
				continue;

			list_for_each_entry(s, &ap->stalist, list)
				return true;
		}
	}

	return false;
}

static void netdev_event_handler(void *agent, struct blob_attr *msg)
{
	char ifname[16] = {0}, link[8] = {0};
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "link", .type = BLOBMSG_TYPE_STRING },
	};
	bool is_ethport = false;
	bool up, down;
	int i;

	if (a->num_ethports <= 0)
		return;

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1]) /* only need ifname and link status */
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(link, blobmsg_data(tb[1]), sizeof(link) - 1);

	for (i = 0; i < a->num_ethports; i++) {
		if (!strncmp(ifname, a->ethports[i], sizeof(ifname))) {
			is_ethport = true;
			break;
		}
	}

	if (!is_ethport)
		return;


	up = !strcmp(link, "up");
	down = !strcmp(link, "down");

	UNUSED(down);

	if (up) {
		dbg("|%s:%d| Scheduling next ACS in 1 second\n", __func__,
				__LINE__);
		timer_set(&a->autocfg_dispatcher, 1 * 1000);
		timestamp_update(&a->eth_connect_t);
	}

	return;
}


static void parse_i1905_info(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct agent *a = (struct agent *)req->priv;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ieee_attrs[2] = {
		[0] = { .name = "ieee1905id", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "interface", .type = BLOBMSG_TYPE_ARRAY }
	};

	blobmsg_parse(ieee_attrs, 2, tb, blob_data(msg), blob_len(msg));

	if (tb[0]) {
		char *mac;

		mac = blobmsg_get_string(tb[0]);
		hwaddr_aton(mac, a->almac);
		dbg("almac = " MACFMT "\n", MAC2STR(a->almac));
	}
}

static void uobj_add_event_handler(void *agent, struct blob_attr *msg)
{
	char path[32] = {0};
	uint32_t id = 0;
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[2];
	static const struct blobmsg_policy ev_attr[2] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "path", .type = BLOBMSG_TYPE_STRING }
	};

	blobmsg_parse(ev_attr, 2, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(path, blobmsg_data(tb[1]), sizeof(path) - 1);
	id = (uint32_t) blobmsg_get_u32(tb[0]);
	dbg("|%s:%d| path = [%s] id = [%d] [%u]\n", __func__, __LINE__, path, id, id);
	if (!strncmp(path, map_plugin, strlen(map_plugin))) {
		/* TODO: how to handle failure? */
		agent_subscribe_for_cmdus(a);
	}
#if 0
	else if (!strncmp(path, "wifi.ap.", 8)) {
		char *ifname = path + 8;
		struct netif_ap *ap;

		dbg("|%s:%d| iface = [%s]\n", __func__, __LINE__, ifname);
		ap = agent_get_ap_by_ifname(a, ifname);
		if (!ap)
			return;
		ap->wifi = id;
		dbg("|%s:%d|wifi.ap.%s object added is [%d] [%u]\n", __func__, __LINE__, ap->ifname, ap->wifi, ap->wifi);
	} else if (!strncmp(path, "wifi.bsta.", 10)) {
		char *ifname = path + 10;
		struct netif_bk *bk;

		warn("|%s:%d| iface = [%s]\n", __func__, __LINE__, ifname);
		bk = agent_get_bsta_by_ifname(a, ifname);
		if (!bk)
			return;
		bk->wifi = id;
		warn("|%s:%d|wifi.bsta.%s object added is [%d] [%u]\n", __func__, __LINE__, bk->ifname, bk->wifi, bk->wifi);
	}
#endif
	else if (!strncmp(path, "ieee1905.al.", 12)) {
		uint32_t ieee1905_obj = 0;
		int ret;

		ret = ubus_lookup_id(a->ubus_ctx, "ieee1905", &ieee1905_obj);
		if (ret)
			return;

		ubus_call_object(a, ieee1905_obj, "info", parse_i1905_info, a);
	} else if (!strncmp(path, "wpa_supplicant.", 15)) {
		char *ifname = path + 15;

		/* re-init blacklist upon supplicant coming back up*/
		backhaul_blacklist_update_ifname(agent, ifname);
	}
#if (EASYMESH_VERSION > 2)
	else if (!strncmp(path, "hostapd.", 8)) {
		/*
		 * QoS rules are volatile and cannot be stored, so they have to be
		 * reapplied after hostapd restarts
		 */
		qos_apply_dscp_rules(agent);

#ifdef USE_LIBDPP
		{
			char *ifname = path + 8;
			struct netif_ap *ap;

			ap = agent_get_ap_by_ifname(a, ifname);
			if (ap && ap->cfg && ap->cfg->advertise_cce)
				dpp_advertise_cce(a, ifname, ap->cfg->advertise_cce);
		}
#endif

	}
#endif
}

void agent_bsta_steer_clear(struct netif_bk *bk)
{
	memset(bk->bsta_steer.target_bssid, 0, 6);
	memset(bk->bsta_steer.prev_bssid, 0, 6);
	bk->bsta_steer.mid = 0;
	memset(bk->bsta_steer.ul_ifname, 0, sizeof(bk->bsta_steer.ul_ifname));
	bk->bsta_steer.expired = false;
}

bool agent_bsta_steer_is_active(struct netif_bk *bk)
{
	return (bk->bsta_steer.mid != 0 && bk->bsta_steer.expired == false);
}

bool agent_bsta_steer_is_active_all(struct agent *a)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->has_bsta && agent_bsta_steer_is_active(&re->bk))
			return true;
	}

	return false;
}


int agent_bsta_steer(struct agent *a, char *ifname, uint8_t *target_bssid,
		     bool force_disassoc)
{
	char fmt[128] = {0};

	UNUSED(a);

	if (force_disassoc) {
		snprintf(fmt, sizeof(fmt), "bsta_steer_with_disassoc %s",
			 ifname);
	} else
		snprintf(fmt, sizeof(fmt), "bsta_steer %s",
			 ifname);

	if (hwaddr_is_zero(target_bssid)|| hwaddr_is_bcast(target_bssid)) {
		snprintf(fmt + strlen(fmt), sizeof(fmt) - strlen(fmt), " any");
	} else {
		snprintf(fmt + strlen(fmt), sizeof(fmt) - strlen(fmt), /* Flawfinder: ignore */
			 " " MACFMT, MAC2STR(target_bssid));
	}

	return agent_exec_platform_scripts(fmt);
}

int agent_bsta_steer_handler(struct agent *a, struct netif_bk *bk, uint8_t *bssid)
{
	uint8_t *target_bssid = bk->bsta_steer.target_bssid;

	if (!memcmp(bssid, target_bssid, 6) ||
	    /* if target was any BSSID, treat connection over correct
	     * bsta as success */
	    hwaddr_is_zero(target_bssid) || hwaddr_is_bcast(target_bssid)) {
		/* success */
		char fmt[64] = {0};

		/* write to config */
		wifi_set_iface_bssid(bk, bssid);
		snprintf(fmt, sizeof(fmt), "write_bsta_config %s", bk->ifname);
		agent_exec_platform_scripts(fmt);
	} else {
		bk->bsta_steer.reason = ERR_REASON_BH_STEERING_NOT_FOUND;
	}

	if (bk->bsta_steer.trigger == BK_STEER_MANDATE)
		send_backhaul_sta_steer_response(a, bk, bk->ifname);

	err("\n\n%s %d bssid = " MACFMT ", target = " MACFMT "\n", __func__,
	    __LINE__, MAC2STR(bssid), MAC2STR(target_bssid));

	timer_del(&bk->steer_timeout);
	agent_bsta_steer_clear(bk);
	return 0;
}

static bool wifi_is_band_onboarded(struct agent *a, enum wifi_band band)
{
	struct wifi_radio_element *re = NULL;

	list_for_each_entry(re, &a->radiolist, list) {
		if (re->has_bsta && re->bk.cfg->onboarded && (re->band & band))
			return true;
	}

	return false;
}

static bool wifi_radio_cac_required(struct agent *a, struct wifi_radio_element *re,
				    int channel, int bandwidth, uint32_t *cac_time)
{
	struct wifi_radio_opclass opclass = {};

	if (!wifi_radio_to_ap(a, re))
		return false;

	/* Get fresh opclass preferences */
	if (WARN_ON(wifi_opclass_preferences(re->name, &opclass))) {
		return false;
	}

	return wifi_opclass_cac_required(&opclass, channel, bandwidth, cac_time);
}

static int wifi_bsta_handle_cac_required(struct agent *a,
					 struct wifi_radio_element *re,
					 struct netif_bk *bk)
{
	struct netif_bk *new_bk = NULL;
	uint32_t cac_time = 0;

	if (!a->cfg.ap_follow_sta_dfs)
		return 0;

	re->cac_required = wifi_radio_cac_required(a, re, re->current_channel,
						   re->current_bandwidth, &cac_time);

	dbg("bsta %s connected on %d/%d cac required %d time %us\n", bk->ifname,
	    re->current_channel, re->current_bandwidth, re->cac_required,
	    cac_time);

	if (!re->cac_required)
		return 0;

	list_for_each_entry(re, &a->radiolist, list) {
		/* Find new bsta */
		if (re->has_bsta && !strcmp(re->bk.ifname, bk->ifname))
			continue;

		new_bk = &re->bk;
		break;
	}

	/* Disable current connection */
	config_disable_bsta(bk->cfg);
	bk->connected = false;


	if (new_bk) {
		dbg("[%s] connect cac required use new link %s\n", bk->ifname, new_bk->ifname);
		config_enable_bsta(new_bk->cfg);
		dynbh_bsta_use_link(a, new_bk->ifname);
	} else {
		dbg("connect cac required disconnect current link %s\n", bk->ifname);
		dynbh_bsta_use_link(a, "none");
	}

	wifi_bsta_disconnect(bk, 0);
	cac_time += 20;
	if (!a->dynbh_upgrade_ongoing)
		timer_set(&a->dynbh_scheduler, cac_time * 1000);

	timestamp_update(&bk->cac_start);
	bk->cac_time = cac_time - 5;

	/* Use 80MHz */
	if (re->current_bandwidth < 80)
		re->current_opclass =
			wifi_opclass_find_id_from_channel(&re->opclass,
					re->current_channel, 80);

	/* Start CAC */
	if (WARN_ON(agent_channel_switch(a, re->macaddr, re->current_channel, re->current_opclass))) {
		bk->cac_time = 0;
		if (!a->dynbh_upgrade_ongoing)
			timer_set(&a->dynbh_scheduler, 5 * 1000);
	}

	/* Don't wait - switch bsta */
	dynbh_bsta_scan_on_enabled(a);

	return 1;
}

void wifi_bsta_check_cac_done(struct agent *a)
{
	struct wifi_radio_element *re = NULL;
	struct netif_bk *bk;
	uint8_t enable_prio = 255;
	bool cac_required;
	uint32_t cac_time;

	if (!a->cfg.ap_follow_sta_dfs)
		return;

	/* lower prio is better */
	list_for_each_entry(re, &a->radiolist, list) {
		bk = &re->bk;

		if (!re->has_bsta)
			continue;

		if (!bk->cfg->enabled)
			continue;

		if (bk->cfg->priority < enable_prio)
			enable_prio = bk->cfg->priority;
	}

	list_for_each_entry(re, &a->radiolist, list) {
		bk = &re->bk;
		if (re->has_bsta)
			continue;

		if (bk->cac_time) {
			cac_required = wifi_radio_cac_required(a, re,
						re->current_channel,
						re->current_bandwidth,
						&cac_time);
			dbg("[%s] %d/%d cac_required %d\n", bk->ifname, re->current_channel,
				re->current_bandwidth, cac_required);

			if (timestamp_expired(&bk->cac_start, bk->cac_time * 1000) || !cac_required) {
				dbg("[%s] connect enable bsta again, cac end (required %d)\n", bk->ifname, cac_required);
				config_enable_bsta(bk->cfg);
				bk->cac_time = 0;
				if (a->dynbh_upgrade_ongoing) {
					a->dynbh_upgrade_ongoing = false;
					timer_set(&a->dynbh_scheduler, 5 * 1000);
				}
			}
		} else {
			/* In case we loose bk->cac_time during agent stop/start */
			if (bk->cfg && !bk->cfg->enabled && bk->cfg->priority < enable_prio) {
				dbg("[%s] connect !cac_time enable higher prio (%d) bsta\n", bk->ifname, enable_prio);
				config_enable_bsta(bk->cfg);
			}
		}
	}
}

static void wifi_bsta_connect(struct agent *a, struct netif_bk *bk,
			      uint8_t *bssid, uint32_t freq)
{
	struct wifi_radio_status radio_status = {};
	struct wifi_radio_element *re = NULL;
	struct wifi_bsta_status status = {};
	uint32_t bandwidth = 0;
	uint32_t channel = 0;
	char fmt[64] = {0};

	/* no operation if no bsta on same band is onboarded! */
	if (!wifi_is_band_onboarded(a, bk->cfg->band)) {
		dbg("|%s:%d| band %d is not onboard\n", __func__, __LINE__,
		    bk->cfg->band);
		return;
	}

	if (!a->cntlr_select.local)
		agent_disable_local_cntlr(a);

	dbg("|%s:%d| connect event received\n", __func__, __LINE__);

	list_for_each_entry(re, &a->radiolist, list)
		re->state = AUTOCFG_ACTIVE;

	agent_enable_fronthauls(a);

	memcpy(bk->bssid, bssid, 6);

	re = agent_get_radio_with_ifname(a, bk->ifname);
	if (!re)
		return;

	/* Assign bssid to configs */
	if (bk->cfg->onboarded
#if (EASYMESH_VERSION >= 6)
	    /* do not assign bssids to MLD netdev or affiliated STAs
	     * TODO: assign MLD bssid once supported by wireless config
	     */
	    && (!bk->is_affiliated_sta && !bk->cfg->is_mld_netdev
	       && re->state == AUTOCFG_HEARTBEAT)
#endif
	    ) {
		wifi_set_iface_bssid(bk, bssid);
		snprintf(fmt, sizeof(fmt), "write_bsta_config %s", bk->ifname);
		agent_exec_platform_scripts(fmt);
	}

	//backhaul_blacklist_clear(a);

	dynbh_update_agent_connect_t(a);
	/* TODO: eth always takes priority */
	if (!agent_is_backhaul_type_eth())
		agent_update_active_uplink(a, bk);
	a->autocfg_interval = 5;
	dbg("|%s:%d| Scheduling next ACS in %u seconds\n", __func__, __LINE__,
			a->autocfg_interval);
	timer_set(&a->autocfg_dispatcher, a->autocfg_interval * 1000);

	/* Get channel/bandwidth from bsta */
	if (!wifi_bsta_status(bk->ifname, &status)) {
		channel = status.channel;
		bandwidth = wifi_bw_to_bw(status.bandwidth);
	}

	/* Update radio channel/bw/opclass */
	re->current_channel = channel;
	re->current_bandwidth = bandwidth;
	re->current_opclass = wifi_opclass_find_id_from_channel(&re->opclass,
				re->current_channel, re->current_bandwidth);

	/* Fall back to radio status */
	if (!re->current_channel) {
		if (!wifi_radio_status(re->name, &radio_status)) {
			channel = radio_status.channel;
			bandwidth = radio_status.bandwidth;

			re->current_channel = radio_status.channel;
			re->current_bandwidth = radio_status.bandwidth;
			re->current_opclass = radio_status.opclass;
		}
	}

	/* Check CAC AP/BSTA action required */
	if (wifi_bsta_handle_cac_required(a, re, bk))
		return;

	if (!channel && freq)
		channel = f2c(freq);

	if (channel) {
		char fmt[512] = {};

		snprintf(fmt, sizeof(fmt), "bsta_set_channel %s %u %u", bk->ifname, channel, bandwidth);
		agent_exec_platform_scripts(fmt);
	}

	dynbh_update_bk_connect_t(a, bk);
	bk->connected = true;
	a->connected = true;
}

static void wifi_bsta_event_scan_failed(struct agent *a, struct netif_bk *bk)
{
	struct wifi_radio_element *re = NULL;
	struct wifi_radio_opclass opclass = {};

	trace("[%s] %s called\n", bk->ifname, __func__);

	if (bk->connected)
		return;

	if (!bk->cfg->enabled)
		return;

	if (bk->cfg->band != BAND_5)
		return;

	re = agent_get_radio_with_ifname(a, bk->ifname);
	if (!re)
		return;

	if (!wifi_radio_to_ap(a, re)) {
		trace("[%s] %s !ap_found\n", bk->ifname, __func__);
		return;
	}

	/* Get fresh opclass preferences */
	if (WARN_ON(wifi_opclass_preferences(re->name, &opclass))) {
		return;
	}

	if (!wifi_opclass_cac_ongoing(&opclass)) {
		trace("[%s] no cac_ongoing\n", bk->ifname);
		return;
	}

	dbg("[%s] cac ongoing skip CAC %s\n", bk->ifname, re->name);
	WARN_ON(wifi_stop_cac(re->name, 36, BW80));
	return;
}

static void wifi_bsta_event_handler(void *agent, struct blob_attr *msg)
{
	char ifname[16] = {0}, event[16] = {0};
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct netif_bk *bk;
	bool add, del;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	bk = agent_get_bsta_by_ifname(a, ifname);
	if (!bk)
		return;

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	/* if scan event and eth backhaul is used - disconnect to prevent loop */
	if ((!strcmp(event, "scan-started") || !strcmp(event, "scan-failed") || add)
			&& agent_is_backhaul_type_eth())
		wifi_bsta_disconnect(bk, 0);

	/** if receive scan event from non-active link, while we are not trying
	 *  to progress or backhaul steer */
	if ((!strcmp(event, "scan-started") || !strcmp(event, "scan-failed"))
			&& ((!bk->connected && !a->dynbh_upgrade_ongoing && a->connected))
			&& !agent_bsta_steer_is_active(bk))
		wifi_bsta_disconnect(bk, 0);

	if (!strcmp(event, "scan-failed")) {
		wifi_bsta_event_scan_failed(a, bk);
		return;
	}

	if (add) {
		uint8_t bssid[6] = {0};
		char bssid_str[18] = {0};
		static const struct blobmsg_policy data_attr[] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[1] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *data[ARRAY_SIZE(data_attr)];
		struct wifi_radio_element *re;
		uint32_t freq = 0;
		int remaining;

		if (!tb[2])
			return;

		/* don't handle event if bk is not enabled */
		if (!bk->cfg->enabled) {
			dbg("|%s:%d| bsta is not enabled - higher prio is "\
					"active, disconnect\n",
					__func__, __LINE__);
			wifi_bsta_disconnect(bk, 0);
			return;
		}


		blobmsg_parse(data_attr, ARRAY_SIZE(data_attr), data, blobmsg_data(tb[2]),
				blobmsg_data_len(tb[2]));
		if (!data[0])
			return;

		strncpy(bssid_str, blobmsg_data(data[0]),
				sizeof(bssid_str) - 1);

		hwaddr_aton(bssid_str, bssid);

		if (data[1])
			freq = blobmsg_get_u32(data[1]);

		wifi_bsta_connect(a, bk, bssid, freq);
		agent_manage_bsta(a, bk);

		timer_del(&a->bh_lost_timer);
		timer_del(&a->bh_reconf_timer);
#ifdef AGENT_ISLAND_PREVENTION
		if (a->cfg.island_prevention) {
			timer_del(&a->sta_disconnect_timer);
			timer_del(&a->fh_disable_timer);
		}
#endif /* AGENT_ISLAND_PREVENTION */

		/* send any pending failed backhaul steer responses */
		list_for_each_entry(re, &a->radiolist, list) {
			if (!re->has_bsta)
				continue;

			if (re->bk.bsta_steer.expired == false ||
			    re->bk.bsta_steer.mid == 0)
				continue;

			/* should always be true from here */
			if (bk->bsta_steer.trigger == BK_STEER_MANDATE) {
				re->bk.bsta_steer.reason = ERR_REASON_BH_STEERING_ASSOC_REJ;
				send_backhaul_sta_steer_response(a, &re->bk, bk->ifname);
			}
			agent_bsta_steer_clear(&re->bk);
		}

		if (agent_bsta_steer_is_active(bk)) {
			/* steer was successful */

			set_value_by_string("mapagent", "agent",
					    "backhaul_override", "1",
					    UCI_TYPE_STRING);
			a->cfg.backhaul_override = 1;

			dynbh_bsta_use_link(a, bk->ifname);
			list_for_each_entry(re, &a->radiolist, list) {
				if (!re->has_bsta)
					continue;

				if (strncmp(re->bk.ifname, bk->ifname, IFNAMSIZ) /* && !agent_bsta_steer_is_active(b) */)
					config_disable_bsta(re->bk.cfg);
			}
			agent_bsta_steer_handler(a, bk, bssid);
			return;
		}

		remaining = timer_remaining_ms(&a->onboarding_scheduler);
		if (remaining == -1 && bk->connected) {
			dynbh_bsta_disable_lower_priority(a, ifname);
			agent_config_reload(a);
			timer_set(&a->disable_unconnected_bstas_scheduler,
					  15 * 1000);
		}
	} else if (del) {
		if (bk->connected)
			dynbh_update_bk_disconnect_t(a, bk);

		bk->connected = false;
		/* Handle link loss if this was supposed to be an active bsta */
		if (bk->cfg->enabled) {
			char ul_ifname[16] = {0};

			dynbh_handle_bh_lost(a, bk);

			if (agent_get_backhaul_ifname(a, ul_ifname)) {
				if (!strncmp(ul_ifname, bk->ifname, IFNAMSIZ)) {
					char ev[512] = {0};

					sprintf(ev,
						"{\"action\":\"unset_uplink\""
						",\"type\":\"wifi\"}");

					agent_notify_event(a, "map.agent", ev);
					agent_exec_platform_scripts("unset_uplink wifi");
				}
			}
		}

		wifi_mod_bridge(a, bk, false);
		a->connected = agent_is_bsta_connected(a);
	} else if (!strcmp(event, "wps-pbc-active")) {
		bk->wps_active = true;
	} else if (!strcmp(event, "wps-timeout")) {
		bk->wps_active = false;
	} else if (!strcmp(event, "csa")) {
		struct wifi_radio_element *re = NULL;
		struct wifi_bsta_status status = {};
		uint32_t cac_time;

		dbg("[%s] connect csa\n", ifname);

		if (a->cfg.ap_follow_sta_dfs) {
			re = agent_get_radio_with_ifname(a, ifname);

			if (re) {
				/* Get channel/bandwidth from bsta */
				if (!wifi_bsta_status(ifname, &status)) {
					re->current_channel = status.channel;
					re->current_bandwidth = wifi_bw_to_bw(status.bandwidth);
					re->current_opclass =
						wifi_opclass_find_id_from_channel(
								&re->opclass,
								re->current_channel,
								re->current_bandwidth);
				}

				if (wifi_radio_cac_required(a, re, re->current_channel,
							    re->current_bandwidth, &cac_time)) {
					/* Simple disconnect, new connect will do required actions */
					wifi_bsta_disconnect(bk, 0);
					dynbh_bsta_scan_on_enabled(a);
				}
			}
		}
	}
}

#if (EASYMESH_VERSION >= 6)
static void wifi_mld_event_handler(void *agent, struct blob_attr *msg)
{
	char ifname[16] = {0}, event[16] = {0};
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add, del;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "mlo-link-added");
	del = !strcmp(event, "mlo-link-remove");

	if (add) {
		/* more actions */
		timer_set(&a->init_ifaces_scheduler,
			  IFACE_TIMEOUT * 1000);
	} else if (del) {
		/* more actions */
		timer_set(&a->init_ifaces_scheduler,
			  IFACE_TIMEOUT * 1000);
	}
}

#endif

#define ONBOARDING_TIMER 15
static void wifi_wps_creds_event_handler(void *agent, struct blob_attr *msg)
{
	char encryption[32] = {0}, ifname[16] = {0}, ssid[33] = {0},
			key[65] = {0};
	static const struct blobmsg_policy ap_attr[5] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "encryption", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "key", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "vlan_id", .type = BLOBMSG_TYPE_INT32 },
	};
	//uint16_t vlan = 0;
	struct agent *a = (struct agent *) agent;
	struct blob_attr *tb[5];
	struct wifi_radio_element *re;
	struct netif_bk *bk;
	int timeout = 0;

	blobmsg_parse(ap_attr, 5, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2] || !tb[3])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(encryption, blobmsg_data(tb[1]), sizeof(encryption) - 1);
	strncpy(ssid, blobmsg_data(tb[2]), sizeof(ssid) - 1);
	strncpy(key, blobmsg_data(tb[3]), sizeof(key) - 1);
	bk = agent_get_bsta_by_ifname(a, ifname);
	if (!bk)
		return;

	bk->wps_active = false;

	/* don't wait for other bands if creds are synced via AP-Autoconfig */
	if (!a->cfg.eth_onboards_wifi_bhs) {
		list_for_each_entry(re, &a->radiolist, list) {
			if (!re->has_bsta)
				continue;

			if (re->bk.wps_active) {
				int remaining = timer_remaining_ms(&a->onboarding_scheduler);

				/* allow up to extra time for the other
				* band(s) to complete
				*/
				timeout = ONBOARDING_TIMER - (remaining == -1 ? 0 : remaining);
			}
		}
	}

	wifi_apply_iface_cfg(bk->ifname, encryption, ssid, key);
	bk->cfg->onboarded = true;

	//uci_reload_services("wireless"); /* move to platform script? */

	re = agent_get_radio_with_ifname(a, ifname);
	if (!re)
		return;

	timer_set(&a->onboarding_scheduler, timeout * 1000);

#if 0 /* Not used at present */
	vlan = 0;
	if (tb[4])
		vlan = (uint16_t) blobmsg_get_u16(tb[4]);

	if (vlan && bk->agent) {
		if (!(a->cfg.pcfg)) {
			a->cfg.pcfg = (struct policy_cfg *)calloc(1, sizeof(struct policy_cfg));
			if (!(a->cfg.pcfg)) {
				err("%s:%d - memory allocation failed\n",
				    __func__, __LINE__);
				return;
			}
		}
		bk->agent->cfg.pcfg->pvid = vlan;

		agent_apply_traffic_separation(bk->agent);
	}
#endif
}

static void agent_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev,
		const char *type, struct blob_attr *msg)
{
	int i;
	char *str;
	struct agent *a = container_of(ev, struct agent, evh);
	struct wifi_ev_handler {
		const char *ev_type;
		void (*handler)(void *ctx, struct blob_attr *ev_data);
	} evs[] = {
		{ "wifi.sta", wifi_sta_event_handler },
		{ "wifi.channel", wifi_channel_event_handler },
		{ "wifi.radio", wifi_radio_event_handler },
		{ "wifi.iface", wifi_iface_event_handler },
		{ "wps_credentials", wifi_wps_creds_event_handler },
		{ "wifi.bsta", wifi_bsta_event_handler },
		{ "network.device", netdev_event_handler },
		{ "ubus.object.add", uobj_add_event_handler },
		{ "wifi.radio", wifi_chan_change_event_handler },
		{ "wifi.radio", wifi_cac_event_handler},
#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
		{ "wifi.sta", dpp_sta_event_handler },
#endif
#endif
#if (EASYMESH_VERSION >= 6)
		{ "wifi.mld", wifi_mld_event_handler },
#endif
	};

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	info("[ &agent = %p ] Received [event = %s]  [val = %s]\n",
			a, type, str);

	for (i = 0; i < ARRAY_SIZE(evs); i++) {
		if (!strcmp(type, evs[i].ev_type)) {
			evs[i].handler(a, msg);
		}
	}

	free(str);
}

static void ieee1905_cmdu_event_handler(void *c, struct blob_attr *msg)
{
	char src[18] = { 0 }, src_origin[18] = { 0 };
	struct agent *a = (struct agent *)c;
	uint8_t srcmac[6], origin[6];
	char in_ifname[16] = {0};
	struct blob_attr *tb[6];
	char *cmdustr = NULL;
	uint8_t *cmdu = NULL;
	int cmdu_strlen = 0;
	uint16_t mid = 0;
	int cmdulen = 0;
	uint16_t type;
	int ret;
	static const struct blobmsg_policy cmdu_attrs[6] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_INT16 },
		[1] = { .name = "mid", .type = BLOBMSG_TYPE_INT16 },
		[2] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "source", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "origin", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "cmdu", .type = BLOBMSG_TYPE_STRING },
	};

	blobmsg_parse(cmdu_attrs, 6, tb, blob_data(msg), blob_len(msg));

	trace("%s: --->\n", __func__);

	if (!tb[0] || !tb[1]) {
		trace("%s: no type or mid.\n", __func__);
		return;
	}

	if (tb[0]) {
		type = (uint16_t)blobmsg_get_u16(tb[0]);
		if (!is_cmdu_for_us(a, type)) {
			trace("%s: type: 0x%x is_cmdu_for_us false.\n", __func__, type);
			return;
		}
	}

	if (tb[1])
		mid = (uint16_t)blobmsg_get_u16(tb[1]);

	if (tb[2])
		strncpy(in_ifname, blobmsg_data(tb[2]), 15);

	if (tb[3]) {
		strncpy(src, blobmsg_data(tb[3]), 17);
		hwaddr_aton(src, srcmac);
	}

	if (tb[4]) {
		strncpy(src_origin, blobmsg_data(tb[4]), 17);
		hwaddr_aton(src_origin, origin);
	}

	if (tb[5]) {
		cmdu_strlen = blobmsg_data_len(tb[5]);
		cmdulen = (cmdu_strlen - 1) / 2;

		cmdustr = calloc(cmdu_strlen + 1, sizeof(char));
		if (!cmdustr)
			return;

		strncpy(cmdustr, blobmsg_data(tb[5]), cmdu_strlen);
		cmdu = calloc(cmdulen, sizeof(uint8_t));
		if (!cmdu) {
			free(cmdustr);
			return;
		}

		strtob(cmdustr, cmdulen, cmdu);
		free(cmdustr);
	}

	ret = CMDU_HOOK(a, type, mid,
			srcmac,
			cmdu, cmdulen,
			CMDU_RX);
	if (ret == CMDU_DROP) {
		warn("[%s]: cmdu 0x%04X dropped by extension at Rx\n", __func__, type);
		goto out;
	}

	agent_handle_map_event(a, type, mid, in_ifname, srcmac, origin, cmdu + 8, cmdulen - 8);
out:
	free(cmdu);
}

static void sync_neighbor_lists(struct netif_ap *ap)
{
	struct nbr nbr[64] = {};
	int nbr_num = 64;
	struct neighbor *n;
	bool found;
	int ret;
	int i;

	/* Update new & changed nbr entries in driver */
	list_for_each_entry(n, &ap->nbrlist, list) {

		if (n->flags & NBR_FLAG_DRV_UPDATED)
			continue;

		/* update nbr in wifi driver */
		dbg("[%s] Update neighbor "
		    MACFMT " entry in driver\n",
		    ap->ifname,
		    MAC2STR(n->nbr.bssid));
		wifi_del_neighbor(ap->ifname, n->nbr.bssid);
		wifi_add_neighbor(ap->ifname, &n->nbr);
		n->flags |= NBR_FLAG_DRV_UPDATED;
	}

	/* Trigger removal of dead entries from list in driver */
	ret = wifi_get_neighbor_list(ap->ifname, nbr, &nbr_num);
	if (ret)
		return;

	for (i = 0; i < nbr_num; i++) {
		found = false;
		list_for_each_entry(n, &ap->nbrlist, list) {
			if (!memcmp(nbr[i].bssid, n->nbr.bssid, 6)) {
				found = true;
				break;
			}
		}

		if (!found) {
			/* Remove dead entry from driver list */
			dbg("[%s] Delete neighbor "
			    MACFMT " entry in driver\n",
			    ap->ifname, MAC2STR(nbr[i].bssid));
			wifi_del_neighbor(ap->ifname, nbr[i].bssid);
		}
	}
}

static void refresh_neighbor_list(atimer_t *t)
{
	struct netif_ap *ap = container_of(t, struct netif_ap, nbr_timer);

	trace("%s: ap = %s\n", __func__, ap->ifname);

	/* First remove aged out entries from FH's nbrlist */
	delete_expired_entries(ap, struct neighbor, &ap->nbrlist, list,
				tsp, NBR_AGEOUT_INTERVAL, NULL,
				ap->nbr_nr);

	/* Now update list in driver so that it matches FH's nbrlist */
	sync_neighbor_lists(ap);

	/* Check again in couple of minutes */
	timer_set(&ap->nbr_timer, NBR_REFRESH_INTERVAL);
}

static void refresh_bssinfo(atimer_t *t)
{
	struct netif_ap *vif = container_of(t, struct netif_ap, bss_timer);
	struct agent *a = vif->agent;
	struct wifi_ap_status ap_status = {};

	int i;

	dbg("%s: vif = %s\n", __func__, vif->ifname);
	if (!wifi_ap_status(vif->ifname, &ap_status)) {
#if (EASYMESH_VERSION >= 6)
		struct wifi_mlsta mlsta[128] = {0};
		int num_mlsta = 128;
#else
		uint8_t sta[128 * 6] = {};
		int num_sta = 128;
#endif
		vif->channel = ap_status.ap.bss.channel;
		memcpy(vif->bssid, ap_status.ap.bss.bssid, 6);
		memcpy(vif->ssid, ap_status.ap.bss.ssid, 32);
		/* others.. */
		if (ap_status.ap.bss.load.utilization != 0xff)
			vif->bssload = ap_status.ap.bss.load.utilization;

		dbg("ap: %s   bssid: " MACFMT
				"  channel: %d   bssload: %d\n",
				vif->ifname, MAC2STR(vif->bssid),
				vif->channel, vif->bssload);
#if (EASYMESH_VERSION >= 6)
		if (!wifi_get_mld_stations(vif, mlsta, &num_mlsta)) {
			for (i = 0; i < num_mlsta; i++) {
				if (!memcmp(mlsta[i].sta->bssid, vif->bssid, 6))
					wifi_add_sta(a, vif->ifname, mlsta[i].sta->macaddr);
			}
		}
#else
		if (!wifi_get_assoclist(vif, sta, &num_sta)) {
			for (i = 0; i < num_sta; i++)
				wifi_add_sta(a, vif->ifname, &sta[i * 6]);
		}
#endif
	}

	timer_set(&vif->bss_timer, BSS_REFRESH_INTERVAL);
}

static void parse_radio_stats(struct ubus_request *req, int type,
			      struct blob_attr *msg)
{
	struct wifi_radio_element *re = (struct wifi_radio_element *)req->priv;
	static const struct blobmsg_policy stats_attr[] = {
		[0] = { .name = "tx_bytes", .type = BLOBMSG_TYPE_INT64 },
		[1] = { .name = "tx_packets", .type = BLOBMSG_TYPE_INT64 },
		[2] = { .name = "tx_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[3] = { .name = "tx_dropped_packets", .type = BLOBMSG_TYPE_INT64 },
		[4] = { .name = "rx_bytes", .type = BLOBMSG_TYPE_INT64 },
		[5] = { .name = "rx_packets", .type = BLOBMSG_TYPE_INT64 },
		[6] = { .name = "rx_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[7] = { .name = "rx_dropped_packets", .type = BLOBMSG_TYPE_INT64 },
		[8] = { .name = "rx_plcp_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[9] = { .name = "rx_fcs_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[10] = { .name = "rx_mac_error_packets", .type = BLOBMSG_TYPE_INT64 },
		[11] = { .name = "rx_unknown_packets", .type = BLOBMSG_TYPE_INT64 },
	};
	struct blob_attr *tb[ARRAY_SIZE(stats_attr)];

	blobmsg_parse(stats_attr, ARRAY_SIZE(stats_attr), tb, blobmsg_data(msg), blob_len(msg));

	if (tb[0])
		re->tx_bytes = blobmsg_get_u64(tb[0]);

	if (tb[1])
		re->tx_packets = blobmsg_get_u64(tb[1]);

	if (tb[2])
		re->tx_error_packets = blobmsg_get_u64(tb[2]);

	if (tb[3])
		re->tx_dropped_packets = blobmsg_get_u64(tb[3]);

	if (tb[4])
		re->rx_bytes = blobmsg_get_u64(tb[4]);

	if (tb[5])
		re->rx_packets = blobmsg_get_u64(tb[5]);

	if (tb[6])
		re->rx_error_packets = blobmsg_get_u64(tb[6]);

	if (tb[7])
		re->rx_dropped_packets = blobmsg_get_u64(tb[7]);

	if (tb[8])
		re->rx_plcp_error_packets = blobmsg_get_u64(tb[8]);

	if (tb[9])
		re->rx_fcs_error_packets = blobmsg_get_u64(tb[9]);

	if (tb[10])
		re->rx_mac_error_packets = blobmsg_get_u64(tb[10]);

	if (tb[11])
		re->rx_unknown_packets = blobmsg_get_u64(tb[11]);
}

static void parse_radio_cac_methods(struct ubus_request *req, int type,
				    struct blob_attr *msg)
{
	struct wifi_radio_element *re = (struct wifi_radio_element *)req->priv;
	struct blob_attr *method;
	int rem_methods;

	blobmsg_for_each_attr(method, msg, rem_methods) {
		if (blobmsg_type(method) != BLOBMSG_TYPE_STRING)
			continue;

		if (!strcmp(blobmsg_data(method), "continous"))
			re->cac_methods |= 1U << WIFI_CAC_CONTINUOUS;
		else if (!strcmp(blobmsg_data(method), "mimo-reduced"))
			re->cac_methods |= 1U << WIFI_CAC_MIMO_REDUCED;
		else if (!strcmp(blobmsg_data(method), "continous-dedicated"))
			re->cac_methods |= 1U << WIFI_CAC_DEDICATED;
		else if (!strcmp(blobmsg_data(method), "time-sliced"))
			re->cac_methods |= 1U << WIFI_CAC_TIME_SLICED;
	}
}

static void parse_radio(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct wifi_radio_element *re = (struct wifi_radio_element *)req->priv;
	static const struct blobmsg_policy radio_attr[] = {
		[0] = { .name = "isup", .type = BLOBMSG_TYPE_BOOL },
		[1] = { .name = "band", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "noise", .type = BLOBMSG_TYPE_INT32 },
		[3] = { .name = "rx_streams", .type = BLOBMSG_TYPE_INT8 },
		[4] = { .name = "tx_streams", .type = BLOBMSG_TYPE_INT8 },
		[5] = { .name = "supp_channels", .type = BLOBMSG_TYPE_ARRAY },
		[6] = { .name = "opclass", .type = BLOBMSG_TYPE_INT32 },
		[7] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[8] = { .name = "regdomain", .type = BLOBMSG_TYPE_STRING },
		[9] = { .name = "txpower", .type = BLOBMSG_TYPE_INT32 },
		[10] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[11] = { .name = "diagnostics", .type = BLOBMSG_TYPE_TABLE },
		[12] = { .name = "bandwidth", .type = BLOBMSG_TYPE_INT32 },
		[13] = { .name = "vendor_id", .type = BLOBMSG_TYPE_STRING },
		[14] = { .name = "stats", .type = BLOBMSG_TYPE_TABLE },
		[15] = { .name = "cac_methods", .type = BLOBMSG_TYPE_ARRAY },
		[16] = { .name = "ap_caps", .type = BLOBMSG_TYPE_TABLE },
		[17] = { .name = "sta_caps", .type = BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[ARRAY_SIZE(radio_attr)];

	dbg("%s ---> %s\n", __func__, re->name);

	blobmsg_parse(radio_attr, ARRAY_SIZE(radio_attr), tb, blob_data(msg), blob_len(msg));

	if (tb[0])
		re->enabled = blobmsg_get_bool(tb[0]);

	if (tb[1]) {
		char *band;

		band = blobmsg_get_string(tb[1]);
		if (!strncmp(band, "2.4GHz", strlen("2.4GHz")))
			re->band = BAND_2;
		else if (!strncmp(band, "5GHz", strlen("5GHz")))
			re->band = BAND_5;
		else if (!strncmp(band, "6GHz", strlen("6GHz")))
			re->band = BAND_6;
		else
			re->band = BAND_UNKNOWN;
	}

	if (tb[2])
		re->anpi = wifi_noise_to_anpi((int)blobmsg_get_u32(tb[2]));

	if (tb[3])
		re->rx_streams = blobmsg_get_u8(tb[3]);

	if (tb[4])
		re->tx_streams = blobmsg_get_u8(tb[4]);
	if (tb[6])
		re->current_opclass = (uint8_t) blobmsg_get_u32(tb[6]);

	if (tb[7])
		re->current_channel = (uint8_t) blobmsg_get_u32(tb[7]);

	if (tb[8])
		memcpy(re->country_code, blobmsg_data(tb[8]), 2);

	if (tb[9])
		re->current_txpower_percent = (uint8_t) blobmsg_get_u32(tb[9]);

	if (tb[10]) {
		char macaddr[18] = {0};

		strncpy(macaddr, blobmsg_data(tb[10]), 17);
		if (!hwaddr_aton(macaddr, re->macaddr))
			return;
	}

	if (tb[11]) {
		struct blob_attr *tb1[4];
		static const struct blobmsg_policy radio_diag_attr[4] = {
			[0] = { .name = "tx_airtime", .type = BLOBMSG_TYPE_INT64 },
			[1] = { .name = "rx_airtime", .type = BLOBMSG_TYPE_INT64 },
			[2] = { .name = "channel_busy", .type = BLOBMSG_TYPE_INT64 },
			[3] = { .name = "obss_airtime", .type = BLOBMSG_TYPE_INT64 },
		};

		blobmsg_parse(radio_diag_attr, 4, tb1, blobmsg_data(tb[11]), blob_len(tb[11]));

		/* get percentage of utilization and scale linearly with 255 */
		if (tb1[0])
			re->tx_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[0]) / (float)1000000) * 255) & 0xff;

		if (tb1[1])
			re->rx_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[1]) / (float)1000000) * 255) & 0xff;

		if (tb1[2])
			re->total_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[2]) / (float)1000000) * 255) & 0xff;

		if (tb1[3])
			re->other_utilization =
				(int)((float)((float)blobmsg_get_u64(tb1[3]) / (float)1000000) * 255) & 0xff;

	}

	if (tb[12])
		re->current_bandwidth = blobmsg_get_u32(tb[12]);

	if (tb[13])
		strncpy(re->vendor, blobmsg_data(tb[13]),
				sizeof(re->vendor) - 1);

	if (tb[14])
		parse_radio_stats(req, type, tb[14]);

	if (tb[15])
		parse_radio_cac_methods(req, type, tb[15]);

	if (tb[16])
		parse_radio_ap_caps(req, type, tb[16]);

	if (tb[17])
		parse_radio_sta_caps(req, type, tb[17]);

	agent_config_opclass(re);
}

struct wifi_scanres_channel_element *wifi_get_scanres_ch_element(
		struct wifi_radio_element *re, uint8_t classid, uint8_t ch)
{
	int i, j;

	if (!re || !re->scanlist)
		return NULL;

	for (i = 0; i < re->scanlist->num_opclass; i++) {
		struct wifi_scanres_opclass_element *op_el;

		op_el = &re->scanlist->opclass_scanlist[i];
		if (op_el->opclass != classid)
			continue;

		for (j = 0; j < op_el->num_channels; j++) {
			if (op_el->channel_scanlist[j].channel == ch)
				return &op_el->channel_scanlist[j];
		}
	}

	dbg("No operating class with channel %d\n", ch);
	return NULL;
}

static void free_scanlist(struct wifi_radio_element *re)
{
	trace("agent: %s: --->\n", __func__);

	struct wifi_scanres_element *scanres_el;
	int j;

	/* return if no scanresults to free */
	if (!re || !re->scanlist)
		return;

	scanres_el = re->scanlist;
	for (j = 0; j < scanres_el->num_opclass; j++) {
		int k;
		struct wifi_scanres_opclass_element *opclass;

		opclass = &scanres_el->opclass_scanlist[j];
		if (!opclass)
			continue;

		for (k = 0; k < opclass->num_channels; k++) {
			struct wifi_scanres_channel_element *ch_el;

			ch_el = &opclass->channel_scanlist[k];
			if (!ch_el)
				continue;

			free(ch_el->nbrlist);
			ch_el->nbrlist = NULL;
		}

		free(opclass->channel_scanlist);
		opclass->channel_scanlist = NULL;
	}

	free(scanres_el->opclass_scanlist);
	scanres_el->opclass_scanlist = NULL;
	free(scanres_el);
	re->scanlist = NULL;
}

#define ADD_NBR_TIMER 10
void reschedule_nbrlist_update(struct netif_ap *ap)
{
	int elapsed = 0;
	int remaining = timer_remaining_ms(&ap->nbr_timer);

	if (remaining < ADD_NBR_TIMER)
		elapsed = ADD_NBR_TIMER - (remaining == -1 ? 0 : remaining);

	timer_set(&ap->nbr_timer, (ADD_NBR_TIMER - elapsed) * 1000);
}

void free_scanlist_neighbors(struct wifi_radio_element *re)
{
	struct wifi_scanres_element *scanres_el;
	int j;

	scanres_el = re->scanlist;
	if (!scanres_el)
		return;

	for (j = 0; j < scanres_el->num_opclass; j++) {
		int k;
		struct wifi_scanres_opclass_element *opclass = &scanres_el->opclass_scanlist[j];

		for (k = 0; k < opclass->num_channels; k++) {
			struct wifi_scanres_channel_element *ch_el = &opclass->channel_scanlist[k];

			if (ch_el->nbrlist)
				free(ch_el->nbrlist);
			ch_el->nbrlist = NULL;
			ch_el->num_neighbors = 0;
		}
	}
}

static int init_scanlist(struct agent *a, const char *radio)
{
	struct wifi_radio_opclass *opclass;
	struct wifi_radio_opclass_entry *entry;
	struct wifi_radio_opclass_channel *channel;
	struct wifi_radio_element *re;
	int num_channels;
	int opclass_supported;
	int i, j, k;

	trace("agent: %s: --->\n", __func__);

	re = agent_get_radio_by_name(a, radio);
	if (WARN_ON(!re))
		return -1;

	opclass = &re->opclass;
	k = 0;

	free_scanlist(re);

	re->scanlist = calloc(1, sizeof(*re->scanlist));
	if (!re->scanlist) {
		dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
		return -1;
	}

	opclass_supported = wifi_opclass_num_supported_bw20(opclass);
	re->scanlist->opclass_scanlist =
		calloc(opclass_supported,
			sizeof(struct wifi_scanres_opclass_element));

	if (!re->scanlist->opclass_scanlist) {
		dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
		goto error;
	}

	re->scanlist->num_opclass = 0;
	for (i = 0; i < opclass->entry_num; i++) {
		entry = &opclass->entry[i];

		if (!wifi_opclass_id_supported(opclass, entry->id)
				|| entry->bandwidth != 20) /* caps */
			continue;

		re->scanlist->opclass_scanlist[k].opclass = entry->id;
		re->scanlist->opclass_scanlist[k].bandwidth = entry->bandwidth;
		/* Store for all supported channels in that opclass */
		num_channels = entry->channel_num;
		re->scanlist->opclass_scanlist[k].num_channels =
				num_channels;

		re->scanlist->opclass_scanlist[k].channel_scanlist =
			calloc(num_channels,
				sizeof(struct wifi_scanres_channel_element));

		if (!re->scanlist->opclass_scanlist[k].channel_scanlist) {
			dbg("|%s:%d| out of memory!\n", __func__, __LINE__);
			goto error;
		}
		/* Store for all supported opclasses */
		re->scanlist->num_opclass++;

		for (j = 0; j < entry->channel_num; j++) {
			channel = &entry->channel[j];
			re->scanlist->opclass_scanlist[k].channel_scanlist[j].channel =
							channel->channel;
		}

		k++;
	}

	return 0;

error:
	free_scanlist(re);
	return -1;
}

static void agent_available_scan_timeout(atimer_t *t)
{
	dbg("[Status code] SCAN NOT COMPLETED (timeout)\n\n");

	struct wifi_radio_element *re;
	int ret;

	re = container_of(t, struct wifi_radio_element, available_scan_timer);

	if (WARN_ON(!re))
		return;

	if (re->scan_state != SCAN_REQUESTED || !re->scan_req.mid)
		return;

	// TODO: revisit - maybe redundant
	re->scan_req.status = CH_SCAN_STATUS_SCAN_NOT_COMPLETED;
	re->scan_state = SCAN_ERROR;

	ret = agent_send_ch_scan_response(re->agent, re, &re->scan_req);
	if (ret)
		return;

	re->scan_req = (const struct wifi_scan_request_radio){ 0 };
}

struct wifi_radio_element *agent_add_radio(struct agent *a, char *ifname)
{
	struct wifi_radio_element *re;

	re = agent_get_radio_by_name(a, ifname);
	if (re)
		return re;

	re = calloc(1, sizeof(struct wifi_radio_element));
	if (!re)
		return NULL;

	a->num_radios++;
	re->agent = a;
	list_add(&re->list, &a->radiolist);
	INIT_LIST_HEAD(&re->aplist);
	strncpy(re->name, ifname, sizeof(re->name));
	timer_init(&re->available_scan_timer,
				agent_available_scan_timeout);
	timer_init(&re->preference_report_timer, preference_report_timer_cb);
	timer_init(&re->bk.steer_timeout, bsta_steer_cb);
	return re;
}

static int init_wifi_radios(struct agent *a)
{
	trace("%s: --->\n", __func__);
	struct agent_config_radio *rcfg;
	int ret = 0;

	list_for_each_entry(rcfg, &a->cfg.radiolist, list) {
		wifi_object_t r_wobj = WIFI_OBJECT_INVALID;
		struct wifi_radio_element *re;
		char r_objname[32] = {0};

		re = agent_add_radio(a, rcfg->name);
		if (!re)
			continue;

		snprintf(r_objname, 31, "wifi.radio.%s", re->name);
		r_wobj = ubus_get_object(a->ubus_ctx, r_objname);
		if (r_wobj == WIFI_OBJECT_INVALID) {
			dbg("%s not present! skipping '%s' from config\n",
					r_objname, re->name);
			continue;
		}

		dbg("%s: getting radio:%s status\n", __func__, re->name);
		ret = ubus_call_object(a, r_wobj, "status", parse_radio, re);

		/* Get fresh opclass preferences after scan */
		wifi_radio_update_opclass_preferences(a, re, 1);
		agent_set_post_scan_action_pref(a, re, true);

		/* Allocate radio_element's scanlist */
		init_scanlist(a, re->name);
	}

	agent_init_wsc_attributes(a);
	return ret;
}

struct netif_ap *netif_alloc_ap(const char *ifname)
{
	struct netif_ap *n = NULL;

	n = calloc(1, sizeof(struct netif_ap));
	if (!n)
		return NULL;

	INIT_LIST_HEAD(&n->stalist);
	INIT_LIST_HEAD(&n->nbrlist);
	INIT_LIST_HEAD(&n->restrict_stalist);
	snprintf(n->ifname, sizeof(n->ifname), "%s", ifname);

	timer_init(&n->nbr_timer, refresh_neighbor_list);
	timer_init(&n->bss_timer, refresh_bssinfo);

	return n;
}

struct netif_wds *netif_alloc_wds(struct agent *a, const char *ifname)
{
	struct netif_wds *n = NULL;

	n = calloc(1, sizeof(struct netif_wds));
	if (!n)
		return NULL;

	snprintf(n->ifname, 15, "%s", ifname);

	list_add(&n->list, &a->wdslist);

	return n;
}

void netif_free_wds(struct netif_wds *n)
{
	list_del(&n->list);
	free(n);
}

void netif_free_wds_by_ifname(struct agent *a, const char *ifname)
{
	struct netif_wds *n = NULL;

	n = agent_get_wds_by_ifname(a, ifname);
	if (!n)
		return;

	netif_free_wds(n);
}

/* get first ap on the radio */
struct netif_ap *wifi_radio_to_ap(struct agent *a, struct wifi_radio_element *re)
{
	struct netif_ap *ap;

	list_for_each_entry(ap, &re->aplist, list) {
		return ap;
	}

	return NULL;
}

static void bk_toggle(struct agent *a, struct netif_bk *bk,
		uint8_t *bssid, uint32_t freq)
{
	/* connect */
	if (!hwaddr_is_zero(bssid) && memcmp(bk->macaddr, bssid, 6)) {
		dynbh_bsta_disable_lower_priority(a, bk->ifname);

		/* checkt previously unconnected */
		if (!bk->connected) {
			wifi_bsta_connect(a, bk, bssid, freq);
			timer_set(&a->disable_unconnected_bstas_scheduler, 0);
			timer_del(&a->bh_lost_timer);
			timer_del(&a->bh_reconf_timer);
		}
	/* disconnect */
	} else {
		bk->connected = false;
		/* check if part of the bridge */
		if (if_isbridge_interface(bk->ifname)) {
			wifi_mod_bridge(a, bk, false);

			if (bk->cfg->enabled) {
				char ul_ifname[16] = {0};

				dynbh_handle_bh_lost(a, bk);

				if (agent_get_backhaul_ifname(a, ul_ifname)
					&& !strncmp(ul_ifname, bk->ifname, IFNAMSIZ))
					agent_exec_platform_scripts("unset_uplink wifi");
			}
		}
	}
}

void parse_bk(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct netif_bk *bk = (struct netif_bk *)req->priv;
	struct agent *a = bk->agent;
	char bssid_str[18] = {0}, macaddr[18] = {0};
	uint8_t bssid[6] = {0};
	uint32_t freq = 0;
	static const struct blobmsg_policy ap_attr[] = {
		[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "frequency", .type = BLOBMSG_TYPE_INT32 },
	};
	struct blob_attr *tb[ARRAY_SIZE(ap_attr)];

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blob_data(msg), blob_len(msg));

	if (!tb[0])
		return;

	/* macaddr */
	strncpy(macaddr, blobmsg_data(tb[0]), 17);
	hwaddr_aton(macaddr, bk->macaddr);

	/* bssid */
	if (tb[1]) {
		strncpy(bssid_str, blobmsg_data(tb[1]), 17);
		hwaddr_aton(bssid_str, bssid);
	}

	/* freq */
	if (tb[2])
		freq = blobmsg_get_u32(tb[2]);

	/* connect / disconnect + dynbh */
	bk_toggle(a, bk, bssid, freq);
}

static void parse_beacon_params(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct wifi_bss_element *bss = (struct wifi_bss_element *)req->priv;
	int rem;
	struct blob_attr *cur;
	struct blob_attr *data[1];
	static const struct blobmsg_policy beacon_attr[1] = {
		[0] = { .name = "beacon-ies", .type = BLOBMSG_TYPE_ARRAY },
	};

	blobmsg_parse(beacon_attr, 1, data, blob_data(msg), blob_len(msg));

	if (!data[0])
		return;

	blobmsg_for_each_attr(cur, data[0], rem) {
		char *iestr = strdup(blobmsg_get_string(cur));

		if (!iestr)
			return;

		/*
		 * esp string having this format: "ff xx 0b"
		 * where xx: 4 or 7 or 10 or 13
		 */
		if ((!strncmp(iestr, "ff", 2)) &&
			(!strncmp(iestr + 4, "0b", 2))) {
			uint8_t *ie, *ie1;
			int i, slen, esp_len;
			int ac;		/* Access Category */

			slen = strlen(iestr);
			ie = calloc(slen/2, sizeof(uint8_t));
			if (!ie) {
				free(iestr);
				return;
			}

			strtob(iestr, slen, ie);
			esp_len = ie[1];
			/*
			 * esp data payload start after 3 bytes;
			 * bypass 3 bytes, (1btye: ff, 1 byte: len, 1byte:0b)
			 */
			ie1 = ie + 3;
			for (i = 1; i < esp_len; i += 3) {
				ac = (*ie1) & 0x03;
				switch (ac) {
				case BK:
					bss->is_ac_bk = 1;
					memcpy(bss->est_wmm_bk, ie1, 3);
					break;
				case BE:
					bss->is_ac_be = 1;
					memcpy(bss->est_wmm_be, ie1, 3);
					break;
				case VI:
					bss->is_ac_vi = 1;
					memcpy(bss->est_wmm_vi, ie1, 3);
					break;
				case VO:
					bss->is_ac_vo = 1;
					memcpy(bss->est_wmm_vo, ie1, 3);
					break;
				}
				ie1 += 3;
			}

			free(ie);
			free(iestr);
			return;
		}
#if (EASYMESH_VERSION >= 6)
		else if ((!strncmp(iestr, "ff", 2)) &&
			(!strncmp(iestr + 4, "6a", 2))) {
			/*
			* esp string having this format: "ff xx 6a"
			* where xx: 9 or 11
			*/
			uint8_t *ie;
			int slen, len;

			slen = strlen(iestr);
			ie = calloc(slen / 2, sizeof(uint8_t));
			if (!ie) {
				free(iestr);
				return;
			}

			strtob(iestr, slen, ie);
			len = ie[1];

			memcpy(bss->eht_oper_caps, ie + 3, len - 1);

			free(ie);
			free(iestr);
			return;
		}
#endif
		free(iestr);
	}
}

static void parse_ap_stats(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct wifi_bss_element *bss = (struct wifi_bss_element *)req->priv;
	struct blob_attr *tb[6];
	static const struct blobmsg_policy ap_stats_attr[6] = {
		[0] = { .name = "tx_unicast_packets", .type = BLOBMSG_TYPE_INT64},
		[1] = { .name = "rx_unicast_packets", .type = BLOBMSG_TYPE_INT64},
		[2] = { .name = "tx_multicast_packets", .type = BLOBMSG_TYPE_INT64},
		[3] = { .name = "rx_multicast_packets", .type = BLOBMSG_TYPE_INT64},
		[4] = { .name = "tx_broadcast_packets", .type = BLOBMSG_TYPE_INT64},
		[5] = { .name = "rx_broadcast_packets", .type = BLOBMSG_TYPE_INT64},
	};


	blobmsg_parse(ap_stats_attr, 6, tb, blob_data(msg), blob_len(msg));

	if (tb[0])
		bss->tx_ucast_bytes = blobmsg_get_u64(tb[0]);

	if (tb[1])
		bss->rx_ucast_bytes = blobmsg_get_u64(tb[1]);

	if (tb[2])
		bss->tx_mcast_bytes = blobmsg_get_u64(tb[2]);

	if (tb[3])
		bss->rx_mcast_bytes = blobmsg_get_u64(tb[3]);

	if (tb[4])
		bss->tx_bcast_bytes = blobmsg_get_u64(tb[4]);

	if (tb[5])
		bss->rx_bcast_bytes = blobmsg_get_u64(tb[5]);
}

#if (EASYMESH_VERSION >= 6)
static int parse_affiliated_ap(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct netif_ap *ap = (struct netif_ap *) req->priv;
	static const struct blobmsg_policy ap_attr[] = {
		[0] = { .name = "id", .type = BLOBMSG_TYPE_INT32 },
		[1] = { .name = "band", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[3] = { .name = "ccfs0", .type = BLOBMSG_TYPE_INT32 },
		[4] = { .name = "ccfs1", .type = BLOBMSG_TYPE_INT32 },
		[5] = { .name = "bandwidth", .type = BLOBMSG_TYPE_INT32 },
		[6] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[7] = { .name = "security", .type = BLOBMSG_TYPE_STRING },
		[8] = { .name = "encryption", .type = BLOBMSG_TYPE_STRING },
		[9] = { .name = "standard", .type = BLOBMSG_TYPE_STRING },
		[10] = { .name = "num_stations", .type = BLOBMSG_TYPE_INT32 },
		[11] = { .name = "max_stations", .type = BLOBMSG_TYPE_INT32 },
		[12] = { .name = "utilization", .type = BLOBMSG_TYPE_INT32 },
		[13] = { .name = "adm_capacity", .type = BLOBMSG_TYPE_INT32 },
		[14] = { .name = "hidden", .type = BLOBMSG_TYPE_BOOL },
		[15] = { .name = "supp_security", .type = BLOBMSG_TYPE_ARRAY },
		[16] = { .name = "capabilities", .type = BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[ARRAY_SIZE(ap_attr)];
	enum wifi_band band;

	if (!ap)
		return -1;

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blobmsg_data(msg), blobmsg_data_len(msg));

	if (!tb[0] || !tb[1] || !tb[6]) {
		return -1;
	}

	band = wifi_bandstr_to_band(blobmsg_get_string(tb[1]));
	if (band != ap->band)
		return -1;

	ap->linkid = blobmsg_get_u32(tb[0]);

	hwaddr_aton(blobmsg_data(tb[6]), ap->bssid);
	if (tb[2])
		ap->channel = blobmsg_get_u32(tb[2]);

	if (tb[5])
		ap->bandwidth = blobmsg_get_u32(tb[5]);

	if (tb[9])
		strncpy(ap->standard, blobmsg_data(tb[9]), sizeof(ap->standard) - 1);

	if (tb[10])
		ap->num_sta = blobmsg_get_u32(tb[10]);

	return 0;
}

static void parse_apmld(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	static const struct blobmsg_policy ap_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "mode", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "links", .type = BLOBMSG_TYPE_ARRAY },
		[5] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
	};
	struct netif_ap *ap = (struct netif_ap *)req->priv;
	struct blob_attr *tb[ARRAY_SIZE(ap_attr)];
	struct agent *a = ap->agent;
	struct netif_mld *mld;
	const char *ifname;
	const char *ssid;
	uint8_t bssid[6] = {0};

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2] || !tb[3] || !tb[4] || !tb[5]) {
		err("|%s:%d| missing some mandatory field!\n", __func__, __LINE__);
		return;
	}

	ifname = blobmsg_get_string(tb[0]);
	if (strlen(ifname) == 0)
		return;

	mld = agent_get_mld_by_ifname(a, ifname);
	if (!mld) {
		mld = agent_alloc_mld(a, ifname);
		if (!mld) {
			err("|%s:%d| failed to alloc MLD\n", __func__, __LINE__);
			return;
		}
	}

	if (!hwaddr_aton(blobmsg_data(tb[1]), mld->macaddr)) {
		err("|%s:%d| MLD mac not valid\n", __func__, __LINE__);
		return;
	}

	if (!hwaddr_aton(blobmsg_data(tb[5]), bssid)) {
		err("|%s:%d| MLD BSSID not valid\n", __func__, __LINE__);
		return;
	}

	/* In case of absence mld interface in the system, macaddres will be zero, use bssid */
	if (hwaddr_is_zero(mld->macaddr) && !hwaddr_is_zero(bssid))
		memcpy(mld->macaddr, bssid, 6);

	ssid = blobmsg_get_string(tb[2]);
	strncpy(mld->ssid, ssid, sizeof(mld->ssid));

	if (tb[4]) {
		struct blob_attr *link = NULL;
		int rem_links = 0;

		blobmsg_for_each_attr(link, tb[4], rem_links) {
			struct wifi_radio_element *re = NULL;
			struct wifi7_radio_capabilities *caps;
			int num_ap = 0;
			int ret;

			if (blobmsg_type(link) != BLOBMSG_TYPE_TABLE) {
				err("|%s:%d| link was not a table\n", __func__, __LINE__);
				continue;
			}

			ret = parse_affiliated_ap(req, type, link);
			if (ret)
				continue;

			strncpy(ap->mld_ifname, ifname, sizeof(ap->mld_ifname) - 1);
			ap->is_affiliated_ap = true;
			strncpy(ap->ssid, ssid, sizeof(ap->ssid));
			ap->enabled = true;

			num_ap = mld->num_affiliated_ap;
			if (!agent_is_ap_in_mld(a, mld, ap->bssid) &&
				num_ap < MAX_AFFILIATED_AP_LINKS_NUM) {
				mld->affiliated_ap[num_ap] = ap;
				mld->num_affiliated_ap++;
			} else {
				err("|%s:%d| link with bssid:"MACFMT" was already in the MLD (or num_ap too high:%d)\n", __func__, __LINE__, MAC2STR(ap->bssid), mld->num_affiliated_ap);
			}

			re = agent_get_radio_with_ifname(a, ap->ifname);
			if (re) {
				caps = &re->wifi7_caps;
				if (caps) {
					/* TODO: revisit: */
					/* Enable based on capabilities */
					if (caps->ap_str_support)
						mld->ap_str_enabled = true;
					if (caps->ap_nstr_support)
						mld->ap_nstr_enabled = true;
					if (caps->ap_emlsr_support)
						mld->ap_emlsr_enabled = true;
					if (caps->ap_emlmr_support)
						mld->ap_emlmr_enabled = true;
				}
			}
			ap->mld = mld;
			break;
		}

		if (mld->num_affiliated_ap == 0) {
			err("|%s:%d| no affiliated APs found in MLD, schedule reload\n", __func__, __LINE__);
			timer_set(&a->init_ifaces_scheduler, IFACE_TIMEOUT * 1000);
		}

	}
}

static void parse_affiliated_bsta(struct agent *a, struct bsta_mld *mld,
				 struct netif_bk *bk, int type,
				 struct blob_attr *msg)
{

	struct blob_attr *link;
	int rem_links;

	/* mlo_links; */
	blobmsg_for_each_attr(link, msg, rem_links) {
		static const struct blobmsg_policy mlo_links[] = {
			[0] = { .name = "mlo_link_id", .type = BLOBMSG_TYPE_INT32 },
			[1] = {	.name = "macaddr", .type = BLOBMSG_TYPE_STRING },
			[2] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
			[3] = { .name = "band", .type = BLOBMSG_TYPE_STRING },
			[4] = { .name = "frequency", .type = BLOBMSG_TYPE_INT32 },
		};
		struct blob_attr *data[ARRAY_SIZE(mlo_links)];
		uint8_t link_id;
		int num_sta;
		enum wifi_band band;

		blobmsg_parse(mlo_links, ARRAY_SIZE(mlo_links), data,
			blobmsg_data(msg), blobmsg_data_len(msg));

		if (!data[0] || !data[1] || !data[2] || !data[3]) {
			err("|%s:%d| missing mlo_links data for ifname:%s\n", __func__, __LINE__, bk->ifname);
			continue;
		}

		link_id = (uint8_t) blobmsg_get_u32(data[0]);

		band = wifi_bandstr_to_band(blobmsg_get_string(data[3]));
		if (band != bk->cfg->band)
			continue;

		hwaddr_aton(blobmsg_data(data[1]), bk->macaddr);
		hwaddr_aton(blobmsg_data(data[2]), bk->bssid);

		num_sta = mld->num_affiliated_bsta;
		bk->mld = mld;
		bk->is_affiliated_sta = true;
		bk->linkid = link_id;

		if (!agent_is_sta_in_bsta_mld(a, mld, bk->macaddr) &&
			num_sta < MAX_AFFILIATED_STA_LINKS_NUM) {
			mld->affiliated_bsta[num_sta] = bk;
			mld->num_affiliated_bsta++;
		}

		break;
	}
}

void parse_bstamld(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	static const struct blobmsg_policy ap_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "4addr", .type = BLOBMSG_TYPE_BOOL },
		[6] = { .name = "mlo_links", .type = BLOBMSG_TYPE_ARRAY },
	};
	struct netif_bk *bk = (struct netif_bk *)req->priv;
	struct blob_attr *tb[ARRAY_SIZE(ap_attr)];
	struct wifi_radio_element *re;
	struct agent *a = bk->agent;
	uint8_t macaddr[6] = {0};
	uint8_t bssid[6] = {0};
	struct bsta_mld *mld = &a->bstamld;
	const char *ifname;
	const char *ssid = NULL;

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2] || !tb[5]) {
		err("|%s:%d| missing some mandatory field(s)!\n", __func__, __LINE__);
		return;
	}

	if (tb[4])
		ssid = blobmsg_get_string(tb[4]);

	if (!hwaddr_aton(blobmsg_data(tb[2]), macaddr)) {
		err("|%s:%d| MAC not valid\n", __func__, __LINE__);
		return;
	}

	if (tb[3] && !hwaddr_aton(blobmsg_data(tb[3]), bssid)) {
		err("|%s:%d| MAC not valid\n", __func__, __LINE__);
		return;
	}

	ifname = blobmsg_get_string(tb[0]);
	re = agent_get_radio_with_ifname(a, bk->ifname);
	if (!re) {
		err("|%s:%d| could not find radio element\n", __func__,
			__LINE__);
		return;
	}

	mld = agent_init_bstamld(a, (char *)ifname, &re->wifi7_caps);
	if (!mld) {
		err("|%s:%d| failed to alloc bSTA MLD ifname:%s\n", __func__, __LINE__, ifname);
		return;
	}

	memcpy(mld->macaddr, macaddr, 6);
	memcpy(mld->bssid, bssid, 6);

	if (bk->cfg->is_mld_netdev && !hwaddr_is_zero(bssid)) {
		bk->mld = mld;
		memcpy(bk->macaddr, macaddr, 6);
		memcpy(bk->bssid, bssid, 6);
	}

	if (tb[6] && !bk->cfg->is_mld_netdev) {
		struct blob_attr *link = NULL;
		int rem_links = 0;

		blobmsg_for_each_attr(link, tb[6], rem_links) {
			if (blobmsg_type(link) != BLOBMSG_TYPE_TABLE) {
				err("|%s:%d| link was not a table\n", __func__, __LINE__);
				continue;
			}

			parse_affiliated_bsta(a, mld, bk, type, link);
			strncpy(bk->mld_ifname, ifname, sizeof(bk->mld_ifname) - 1);
			bk->is_affiliated_sta = true;
			if (ssid)
				strncpy(bk->ssid, ssid, sizeof(bk->ssid));
			bk->enabled = true;
			bk->mld = mld;
		}
	}

	if (!bk->is_affiliated_sta || bk->cfg->is_mld_netdev)
		bk_toggle(a, bk, bssid, 0);
}
#endif

static void parse_ap(struct ubus_request *req, int type,
		struct blob_attr *msg)
{
	struct netif_ap *ap = (struct netif_ap *)req->priv;
	char bssid[18] = { 0 };
	static const struct blobmsg_policy ap_attr[] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "capabilities", .type = BLOBMSG_TYPE_TABLE },
		[3] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
		[4] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
		[5] = { .name = "standard", .type = BLOBMSG_TYPE_STRING },
		[6] = { .name = "bandwidth", .type = BLOBMSG_TYPE_INT32 },
		[7] = { .name = "status", .type = BLOBMSG_TYPE_STRING },
		[8] = { .name = "num_stations", .type = BLOBMSG_TYPE_INT32 },
		[9] = { .name = "enabled", .type = BLOBMSG_TYPE_BOOL },
#if (EASYMESH_VERSION >= 6)
		[10] = { .name = "mld", .type = BLOBMSG_TYPE_STRING },
		[11] = { .name = "mld_ifname", .type = BLOBMSG_TYPE_STRING },
		[12] = { .name = "linkid", .type = BLOBMSG_TYPE_INT32 },
#endif
		[13] = { .name = "band", .type = BLOBMSG_TYPE_STRING },
	};
	struct blob_attr *tb[ARRAY_SIZE(ap_attr)];

	blobmsg_parse(ap_attr, ARRAY_SIZE(ap_attr), tb, blob_data(msg), blob_len(msg));

	if (!tb[1])
		return;

	strncpy(bssid, blobmsg_data(tb[1]), 17);
	if (!hwaddr_aton(bssid, ap->bssid))
		return;

	if (tb[3])
		ap->channel = blobmsg_get_u32(tb[3]);

	if (tb[4])
		strncpy(ap->ssid, blobmsg_data(tb[4]), sizeof(ap->ssid) - 1);

	if (tb[5])
		strncpy(ap->standard, blobmsg_data(tb[5]), sizeof(ap->standard) - 1);

	if (tb[6])
		ap->bandwidth = blobmsg_get_u32(tb[6]);

	if (tb[9])
		ap->enabled = blobmsg_get_bool(tb[9]);

	if (tb[8])
		ap->num_sta = blobmsg_get_u32(tb[8]);


	if (!ap->enabled && may_enable_fhs(ap->agent))
		/* Enable fronthauls */
		timer_set(&ap->agent->enable_fhs_scheduler,
				  (IFACE_TIMEOUT + 1) * 1000);

#if (EASYMESH_VERSION >= 6)
	if (tb[11]) {
		strncpy(ap->mld_ifname, blobmsg_data(tb[11]),
			sizeof(ap->mld_ifname) - 1);
		ap->is_affiliated_ap = true;
	}

	if (tb[12])
		ap->linkid = blobmsg_get_u32(tb[12]);
#endif

	if (tb[13]) {
		char *band;

		band = blobmsg_get_string(tb[13]);
		if (!strncmp(band, "2.4GHz", strlen("2.4GHz")))
			ap->band = BAND_2;
		else if (!strncmp(band, "5GHz", strlen("5GHz")))
			ap->band = BAND_5;
		else if (!strncmp(band, "6GHz", strlen("6GHz")))
			ap->band = BAND_6;
		else
			ap->band = BAND_UNKNOWN;
	}

}

static int agent_radio_update_scanresults_element(struct wifi_radio_element *re, struct wifi_bss *bsss,
		uint32_t classid)
{
	struct wifi_scanres_channel_element *scanres_el;
	int idx;

	dbg("%s: classid = %d\n", __func__, classid);

	scanres_el = wifi_get_scanres_ch_element(re, classid, bsss->channel);
	if (!scanres_el)
		return -1;

	// no neighbors, allocate first
	if (!scanres_el->num_neighbors) {
		scanres_el->nbrlist = calloc(1, sizeof(*(scanres_el->nbrlist)));
		if (!scanres_el->nbrlist)
			return -1;

		scanres_el->num_neighbors = 1;
	}
	// at least one neighbor, allocate one more
	else {
		struct wifi_scanres_neighbor_element *nbr;
		int size;

		size = (scanres_el->num_neighbors + 1) * sizeof(*nbr);
		nbr = (struct wifi_scanres_neighbor_element *)realloc(
					scanres_el->nbrlist, size);

		if (!nbr)
			return -1;

		scanres_el->num_neighbors++;
		scanres_el->nbrlist = nbr;
	}

	idx = scanres_el->num_neighbors - 1; // fill index

	/* Fill in the neighbour data */
	memcpy(scanres_el->nbrlist[idx].bssid, bsss->bssid, 6);

	strncpy(scanres_el->nbrlist[idx].ssid, (char *) bsss->ssid,
			sizeof(scanres_el->nbrlist->ssid) - 1);

	scanres_el->nbrlist[idx].rssi = bsss->rssi;
	scanres_el->nbrlist[idx].bw = wifi_bw_to_bw(bsss->curr_bw);
	scanres_el->nbrlist[idx].utilization = bsss->load.utilization;
	scanres_el->nbrlist[idx].num_stations = bsss->load.sta_count;
	return 0;
}

static int agent_radio_update_scanlist(struct agent *a,
		struct wifi_radio_element *re, struct wifi_bss *bsss, int bss_num)
{
	int i;

	trace("%s: update results for [%s]\n", __func__, re->name);

	for (i = 0; i < bss_num; i++) {
		uint8_t classid = wifi_opclass_find_id_from_channel(
					&re->opclass,
					bsss[i].channel,
					20); /* caps */

		if (agent_radio_update_scanresults_element(re, &bsss[i], classid))
			continue;

		if (classid == OPCLASS_124) {
			dbg("%s: duplicate entries from classid %d to %d\n",
			    __func__, OPCLASS_124, OPCLASS_125);
			if (agent_radio_update_scanresults_element(re, &bsss[i], OPCLASS_125))
				continue;
		}
	}

	return 0;
}

int agent_radio_scanresults(struct agent *a, struct wifi_radio_element *re)
{
	struct wifi_bss bss[128];
	int bss_num = ARRAY_SIZE(bss);
	int ret;

	if (!re || !strlen(re->name))
		return -1;

	trace("[%s] radio scanresults\n", re->name);

	/* Get scan results from the driver */
	ret = wifi_get_scan_results(re->name, bss, &bss_num);

	if (ret) {
		dbg("[%s] failed to get scanresults\n", re->name);
		return -1;
	}

	/* Update scan cache */
	ret = wifi_scanresults_add(&re->scanresults, &re->opclass, bss, bss_num);
	if (ret) {
		dbg("[%s] failed to update scan cache\n", re->name);
		return -1;
	}

	/* TODO: update scanlist using fresh results from the cache */

	/* Empty scanlist & add most recent results for reporting */
	free_scanlist_neighbors(re);
	ret = agent_radio_update_scanlist(a, re, bss, bss_num);

	return ret;
}

void agent_init_interfaces_post_actions(struct agent *a)
{
	trace("%s: --->\n", __func__);
	struct wifi_radio_element *re;
	struct netif_ap *ap = NULL;
	int i;

	list_for_each_entry(re, &a->radiolist, list) {
		list_for_each_entry(ap, &re->aplist, list) {
#if (EASYMESH_VERSION >= 6)
			struct wifi_mlsta mlsta[128] = {0};
			int num_mlsta = 128;

			if (wifi_get_mld_stations(ap, mlsta, &num_mlsta))
				continue;

			for (i = 0; i < num_mlsta; i++) {
				if (!memcmp(mlsta[i].sta->bssid, ap->bssid, 6))
					wifi_add_sta(a, ap->ifname, mlsta[i].sta->macaddr);
			}
#else
			uint8_t sta[128 * 6] = {};
			int num_sta = 128;

			if (wifi_get_assoclist(ap, sta, &num_sta))
				continue;

			for (i = 0; i < num_sta; i++)
				wifi_add_sta(a, ap->ifname, &sta[i * 6]);
#endif
			/* Block STas in WiFi if wifi is restarted at run-time */
			assoc_ctrl_apply_restriction(a, ap);
		}

		/* request scan results from the driver, etc */
		if (re->name[0] != '\0' && strlen(re->name)) {
			/* Make sure scanlist is initialized */
			if (!re->scanlist)
				init_scanlist(a, re->name);
			/* Get & store scan results */
			agent_radio_scanresults(a, re);
			/* Finally update neighbor data */
			update_neighbors_from_scancache(a, &re->scanresults);
		}
	}

	agent_exec_platform_scripts("update_acs");
}

int agent_get_system_info(struct agent *a)
{

	char buf[68] = {0};

	chrCmd(buf, sizeof(buf), "/lib/wifi/multiap db_get device.deviceinfo.SerialNumber");

	strncpy(a->device_inventory.serial_number, buf,
		sizeof(a->device_inventory.serial_number) - 1);

	chrCmd(buf, sizeof(buf), "/lib/wifi/multiap db_get device.deviceinfo.SoftwareVersion");

	strncpy(a->device_inventory.sw_version, buf,
		sizeof(a->device_inventory.sw_version) - 1);

	dbg("|%s:%d| serial number:%s software version:%s\n", __func__, __LINE__, a->device_inventory.serial_number,
		a->device_inventory.sw_version);
	return 0;
}

int agent_init_interfaces(struct agent *a)
{
	struct agent_config *cfg = &a->cfg;
	struct wifi_radio_element *re;
	struct netif_apcfg *f;
	struct netif_bkcfg *b;
	uint32_t ieee1905_obj = 0;
	int ret;
#if (EASYMESH_VERSION > 2)
	struct dpp_chirp *u;
#endif
#if (EASYMESH_VERSION >= 6)
	struct netif_mld *mld, *tmp;
#endif

	ret = ubus_lookup_id(a->ubus_ctx, "ieee1905", &ieee1905_obj);
	if (ret) {
		warn("%s: ieee1905 object missing over UBUS, retry in 10s\n", __func__);
		timer_set(&a->init_ifaces_scheduler, 10 * 1000);
		return -1;
	}

	ubus_call_object(a, ieee1905_obj, "info", parse_i1905_info, a);

	ret = init_wifi_radios(a);
	if (ret || list_empty(&a->radiolist)) {
		warn("%s: Failure initializing radios, retry in 10s\n", __func__);
		timer_set(&a->init_ifaces_scheduler, 10 * 1000);
	}

	/* mark APs as torndown before fetching fresh data */
	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap;

		re->has_bsta = false;
		memset(&re->bk, 0, sizeof(re->bk));
		list_for_each_entry(ap, &re->aplist, list) {
			ap->torndown = true;
			ap->enabled = false;
#if (EASYMESH_VERSIN >= 6)
			if (ap->is_affiliated_ap)
				wifi_mld_del_ap(ap);
#endif
		}
	}

#if (EASYMESH_VERSION >= 6)
	/* mark AP MLDs as disabled before fetching fresh data*/
	list_for_each_entry(mld, &a->apmldlist, list)
		mld->enabled = false;
#endif
	a->connected = false;

	list_for_each_entry(f, &cfg->aplist, list) {
		void (*fn_parser)(struct ubus_request *, int, struct blob_attr *) = parse_ap;
		wifi_object_t uobj_id = WIFI_OBJECT_INVALID;
		struct wifi_bss_element *bss;
		struct netif_ap *fn = NULL;
		char objname[32] = {0};
		struct subscribe_fr {
			uint8_t type;
			uint8_t stype;
		} subfr[] = {{0, 0}, {0, 2}, {0, 13}};
		int num_subfr = ARRAY_SIZE(subfr);
		int ret;
		int k;

		snprintf(objname, 31, "wifi.ap.%s", f->name);

		re = agent_get_radio_by_name(a, f->device);
		if (!re)
			continue;

		fn = agent_get_ap_by_ifname(a, f->name);
		if (!fn) {
			fn = netif_alloc_ap(f->name);
			if (fn) {
				/* subscribe following frames
				* WIFI_FRAME_ASSOC_REQ,
				* WIFI_FRAME_REASSOC_REQ,
				* WIFI_FRAME_ACTION
				*/
				// TODO: call unsubscribe_frame method on cleanup.

				fn->cfg = f;
				fn->agent = a;
				strncpy(fn->radio_name, re->name, IFNAMSIZ-1);

				list_add(&fn->list, &re->aplist);
			} else
				continue;
		}

		fn->band = re->band;
		fn->torndown = false;

#if (EASYMESH_VERSION >= 6)
		if (f->mld_id) {
			struct mld_credential *mldcred;

			mldcred = agent_get_mld_credential_by_id(&a->cfg,
								 f->mld_id);
			if (mldcred) {
				snprintf(objname, 31, "wifi.apmld.%s", mldcred->ifname);
				fn_parser = parse_apmld;
			}
		}
#endif

		uobj_id = ubus_get_object(a->ubus_ctx, objname);
		if (uobj_id == WIFI_OBJECT_INVALID) {
			err("%s not present! skipping '%s' from config\n",
							objname, f->name);
			continue;
		}

		ret = ubus_call_object(a, uobj_id, "status", fn_parser, fn);
		if (ret) {
			err("Failed to get '%s' (ret = %d) name:%s\n", "status", ret, objname);
			timer_set(&a->init_ifaces_scheduler,
				IFACE_TIMEOUT * 1000);
		}

		bss = agent_get_wifi_bss(a, fn->bssid);
		if (!bss) {
			struct wifi_bss_element *bsslist;

			bsslist = (struct wifi_bss_element *)realloc(re->bsslist,
					(re->num_bss + 1) * sizeof(struct wifi_bss_element));
			if (!bsslist)
				continue;

			re->bsslist = bsslist;
			re->num_bss++;
			bss = &re->bsslist[re->num_bss - 1];
			memset(bss, 0, sizeof(struct wifi_bss_element));
		}

		memcpy(bss->bssid, fn->bssid, 6);
		strncpy(bss->ssid, fn->ssid, sizeof(bss->ssid) - 1);
		bss->enabled = fn->enabled;

#if 1 /* TODO: remove once support is added from wifi.mld object */

		snprintf(objname, 31, "wifi.ap.%s", f->name);
		uobj_id = ubus_get_object(a->ubus_ctx, objname);
		if (uobj_id == WIFI_OBJECT_INVALID) {
			err("%s not present! skipping '%s' from config\n",
					objname, f->name);
			continue;
		}
#endif

		for (k = 0; k < num_subfr; k++)
			wifi_subscribe_frame(f->name, subfr[k].type, subfr[k].stype);

		ubus_call_object(a, uobj_id, "stats", parse_ap_stats, bss);
		ubus_call_object(a, uobj_id, "dump_beacon", parse_beacon_params, bss);
		/* TODO: FIXME: we need to read stations here! */
	}

	/** if AP is gone from config free it altogether
	 *  if AP is disabled, cancel all timers/data
	 */
	list_for_each_entry(re, &a->radiolist, list) {
		struct netif_ap *ap, *tmp;

		list_for_each_entry_safe(ap, tmp, &re->aplist, list) {
			if (ap->torndown)
				netif_free(ap);
			else if (!ap->enabled)
				netif_reset(ap);
		}

	}

	list_for_each_entry(b, &cfg->bklist, list) {
		void (*bk_parser)(struct ubus_request *, int, struct blob_attr *) = parse_bk;
		wifi_object_t uobj_id = WIFI_OBJECT_INVALID;
		struct wifi_radio_element *re;
		struct netif_bk *bn;
		char objname[32] = {0};


		snprintf(objname, 31, "wifi.bsta.%s", b->name);

		re = agent_get_radio_by_name(a, b->device);
		if (re) {
			bn = &re->bk;
			re->has_bsta = true;
		}
#if (EASYMESH_VERSION >= 6)
		else if (b->is_mld_netdev) {
			bn = &a->bstamld.bk;
			a->has_bstamld = true;
		}
#endif
		else
			continue;

		snprintf(bn->ifname, sizeof(bn->ifname), "%s", b->name);

		bn->cfg = b;
		bn->agent = a;
		memset(bn->macaddr, 0, 6);

#if (EASYMESH_VERSION >= 6)
		if (b->mld_id) {
			struct mld_credential *mldcred;

			mldcred = agent_get_mld_credential_by_id(&a->cfg,
								 b->mld_id);
			if (mldcred) {
				snprintf(objname, 31, "wifi.bstamld.%s", mldcred->ifname);
				bk_parser = parse_bstamld;
			}
		}
#endif

		uobj_id = ubus_get_object(a->ubus_ctx, objname);
		if (uobj_id == WIFI_OBJECT_INVALID) {
			dbg("%s not present! skipping '%s' from config\n",
							objname, b->name);
			continue;
		}
		ret = ubus_call_object(a, uobj_id, "status", bk_parser, bn);
		if (ret) {
			err("Failed to get '%s' (ret = %d) name:%s\n", "status", ret, objname);
			timer_set(&a->init_ifaces_scheduler,
					IFACE_TIMEOUT * 1000);
		}
		if (bn->connected)
			a->connected = true;


	}

#if (EASYMESH_VERSION >= 6)
	/* if an MLD is gone from config free it */
	list_for_each_entry_safe(mld, tmp, &a->apmldlist, list) {
		if (!agent_get_mld_credential_by_ifname(&a->cfg, mld->ifname))
			clear_apmld(mld);
	}

	if (a->has_bstamld) {
		/* if bsta config has gone away free bstamld */
		if (!agent_get_bsta_mld_credential(&a->cfg))
			agent_clear_bsta_mld(a);
	}
#endif

#if (EASYMESH_VERSION > 2)
	list_for_each_entry(u, &cfg->chirplist, list) {
		struct netif_bk *bn;

		bn = agent_get_bsta_by_ifname(a, u->ifname);
		if (bn) {
			bn->dpp_cfg = u;

			if (timer_remaining_ms(&bn->dpp_do_chirp) == -1) {
				timer_set(&bn->dpp_do_chirp, 0);
			}
		}
	}
#endif

	/* TODO: move this to backhaul.c and post_actions */
	if (!agent_has_active_backhaul(a)
	    && timer_remaining_ms(&a->bh_lost_timer) == -1
	    && !list_empty(&cfg->bklist)) {
		if (a->cfg.ap_follow_sta_dfs) {
			struct dyn_bh_cfg *dbcfg = &a->cfg.dbhcfg;
			struct wifi_radio_element *re;
			struct netif_bk *sta = NULL;

			list_for_each_entry(re, &a->radiolist, list) {
				if (!re->has_bsta)
					continue;

				sta = &re->bk;
				dbg("[%s] connect %s bsta\n", sta->ifname, sta->cac_time ? "disable" : "enable");
				if (sta->cac_time)
					config_disable_bsta(sta->cfg);
				else
					config_enable_bsta(sta->cfg);
			}

			dynbh_bsta_scan_on_enabled(a);
			if (dbcfg->bh_miss_tmo)
				/* Trigger bh link loss timer */
				timer_set(&a->bh_lost_timer, dbcfg->bh_miss_tmo * 1000);

			timestamp_update(&a->disconnect_t);
		} else {
			struct dyn_bh_cfg *dbcfg = &a->cfg.dbhcfg;

			dynbh_bsta_enable_all(a);
			if (dbcfg) {
				timer_set(&a->bh_reconf_timer,
					  dbcfg->bh_reconf_tmo * 1000);
				timestamp_update(&a->disconnect_t);
			}
		}
	} else if (!list_empty(&cfg->bklist) && !a->dynbh_upgrade_ongoing) {
		timer_set(&a->dynbh_scheduler, 30 * 1000);
	}

	agent_config_reload(a);

	agent_init_interfaces_post_actions(a);
	if (cfg && cfg->pcfg && is_vid_valid(cfg->pcfg->pvid)) {
		a->reconfig_reason |= AGENT_RECONFIG_REASON_VLAN_SETUP;
		timer_set(&a->reload_scheduler, 5 * 1000);
	}

	return 0;
}

static void agent_enable_fhs_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, enable_fhs_scheduler);

	agent_enable_fronthauls(a);
}

static void agent_boot_scan_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, boot_scan_scheduler);
	struct wifi_radio_element *re;
	bool radio_missing = false;

	if (!a)
		return; /* err */

	dbg("|%s:%d| boot scan %s", __func__, __LINE__,
	    a->cfg.scan_on_boot_only ? "enabled" : "disabled");

	if (!a->cfg.scan_on_boot_only)
		/* On boot scan disabled */
		return;

	list_for_each_entry(re, &a->radiolist, list) {
		if (!re->enabled) {
			radio_missing = true;
			continue;
		}

		if (re->scan_state != SCAN_DONE)
			wifi_radio_scan_req_all(a);
	}

	if (a->boot_scan_tries++ < BOOT_UP_SCAN_MAX_TRY
			&& radio_missing)
		/* one or more radio not available yet, re-try again later */
		timer_set(&a->boot_scan_scheduler, BOOT_UP_SCAN_ITV * 1000);
}


#define OPER_CHAN_REP_INTERVAL 60
static void agent_periodic_run(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, agent_periodic_run);
	sigset_t waiting_mask;

	a->uptime += 1;

	sigpending(&waiting_mask);

	if (sigismember(&waiting_mask, SIGHUP)) {
		dbg("|%s:%d| Received SIGHUP, reload config\n", __func__, __LINE__);
		signal(SIGHUP, SIG_IGN);

		agent_config_reload(a);
	}

	trace("|%s:%d| periodic time (elapsed:%"PRIu64")\n", __func__, __LINE__,
	      a->uptime);

	if ((a->uptime % OPER_CHAN_REP_INTERVAL) == 0)
		send_oper_channel_report(a, NULL);

	timer_set(&a->agent_periodic_run, 1 * 1000);
}


static void agent_radio_stats_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);

	struct agent *a = container_of(t, struct agent, radio_stats_scheduler);
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		wifi_object_t r_wobj = WIFI_OBJECT_INVALID;
		char r_objname[32] = {0};

		/* mark outdated scanresults */
		wifi_scanresults_mark_expired(&re->scanresults);

		/* get radio status and update stats */
		snprintf(r_objname, 31, "wifi.radio.%s", re->name);
		r_wobj = ubus_get_object(a->ubus_ctx, r_objname);

		if (r_wobj == WIFI_OBJECT_INVALID)
			continue;

		ubus_call_object(a, r_wobj, "status", parse_radio, re);
	}

	timer_set(&a->radio_stats_scheduler, RADIO_STATS_TIMER);
}

static void agent_init_ifaces_cb(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, init_ifaces_scheduler);

	if (!a)
		return;

	agent_init_interfaces(a);
}


#define RELOAD_TIMEOUT 10

static void agent_reload_cb(atimer_t *t)
{
	trace("%s: --->\n", __func__);
	struct agent *a = container_of(t, struct agent, reload_scheduler);

	if (a->cfg.eth_onboards_wifi_bhs) {
		/* re-sync any credentials passed with AP-Autoconfig */
		agent_exec_platform_scripts("bsta_to_wireless");
	}

	agent_exec_platform_scripts("sync_credentials");

	dbg("|%s:%d| reconfig_reason = %d\n", __func__, __LINE__, a->reconfig_reason);

	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_VLAN_TEARDOWN)) {
		dbg("|%s:%d| ts teardown\n", __func__, __LINE__);
		agent_clear_traffic_sep(a);
	}

	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_VLAN_SETUP)) {
		dbg("|%s:%d| ts setup\n", __func__, __LINE__);
		agent_apply_traffic_separation(a);
	}

	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_AP_CRED)) {
		struct wifi_radio_element *re = NULL;

		dbg("|%s:%d| apconf reload APs credentials\n", __func__, __LINE__);
		list_for_each_entry(re, &a->radiolist, list) {
			struct netif_ap *ap = NULL;

			list_for_each_entry(ap, &re->aplist, list) {
				char cmd[512] = {};

				if (!ap->cfg)
					continue;

				if (ap->reconfig_reason & AGENT_AP_RECONFIG_SSID) {
					dbg("%s apconf reconfig ssid %s\n", ap->ifname, ap->cfg->ssid);
					snprintf(cmd, sizeof(cmd), "set_ap_ssid %s %s", ap->ifname, ap->cfg->ssid);
					agent_exec_platform_scripts(cmd);
				}

				if (ap->reconfig_reason & AGENT_AP_RECONFIG_PSK) {
					dbg("%s apconf reconfig psk %s\n", ap->ifname, ap->cfg->key);
					snprintf(cmd, sizeof(cmd), "set_ap_psk %s %s", ap->ifname, ap->cfg->key);
					agent_exec_platform_scripts(cmd);
				}

				ap->reconfig_reason = 0;
			}
		}
	}

#if (EASYMESH_VERSION >= 6)
	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_AP_AUTOCONF) ||
	    !!(a->reconfig_reason & AGENT_RECONFIG_REASON_PUNCT_BITMAP) ||
	    !!(a->reconfig_reason & AGENT_RECONFIG_REASON_MLD_ID)) {
		dbg("|%s:%d| reload wireless\n", __func__, __LINE__);
#else
	if (!!(a->reconfig_reason & AGENT_RECONFIG_REASON_AP_AUTOCONF)) {
		dbg("|%s:%d| apconf reload wireless\n", __func__, __LINE__);
#endif

		/*
		 * Allow to change order in config/wireless.
		 * Some platforms require this before wifi
		 * reload/restart.
		 */
		agent_exec_platform_scripts("config_reorder");

		uci_reload_services("wireless");
	}

	a->reconfig_reason = 0;
}

/* send the steering compled message this function also resets the value of
 * steering opportunity
 */
static void agent_steering_opp_timeout(atimer_t *t)
{
	trace("agent: %s: --->\n", __func__);
	struct agent *a = container_of(t, struct agent, sta_steer_req_timer);
	uint8_t origin[6] = { 0 };


	memcpy(origin, a->cntlr_almac, 6);
	send_sta_steer_complete((void *)a, origin);
}

#define max_val(a,b) \
	({ typeof(a) _a = (a); \
	typeof(b) _b = (b); \
	_a > _b ? _a : _b; })

static void agent_start_controller_conditionally(struct agent *a)
{
	uint8_t max_retries = a->cntlr_select.retry_int;
	uint16_t probe_int = a->cntlr_select.probe_int;
	int try = a->cntlr_miss_count + 1;
	int max_misstime = max_val((max_retries * probe_int),
				HEARTBEAT_AUTOCFG_INTERVAL + 5);
	bool start_controller = false;

	trace("|%s:%d| Entry max_misstime %d\n", __func__, __LINE__, max_misstime);

	if (!timestamp_expired(&a->observed_time, (try * probe_int) * 1000)
			|| is_local_cntlr_running()) {
		dbg("|%s:%d| Active controller available\n",
			__func__, __LINE__);
		a->cntlr_miss_count = 0;
		a->active_cntlr = true;
		return;
	}

	if (!timestamp_expired(&a->observed_time, max_misstime * 1000)) {
		a->cntlr_miss_count++;
		return;
	}

	a->active_cntlr = false;

	memset(a->cntlr_almac, 0, sizeof(a->cntlr_almac));

	dbg("|%s:%d| No controller observed for over %d seconds\n",
	    __func__, __LINE__, max_misstime);

#ifdef AGENT_DYNAMIC_CNTRL
	if (a->cntlr_select.autostart)
		start_controller = true;
#endif

	if (start_controller) {
		if (is_local_cntlr_available()) {
			time_t now;
			/* Schedule start of local Controller after
			 * random time chosen from a 10s time window
			 * to reduce chance of multiple MAP-Agents
			 * starting own controller at the same time.
			 */
			srand(time(&now));
			agent_schedule_cntlr_start(a, rand() % 10);
		} else {
			warn("Local cntlr unavailable: skip start\n");
		}
	} else if (a->cntlr_miss_count) {
		/* Report controller absence to the user first time
		 * it goes missing
		 */
		wifiagent_log_cntlrinfo(a);
	}

	a->cntlr_miss_count = 0;
}


static void agent_dispatch_autoconfig(atimer_t *t)
{
	struct agent *a = container_of(t, struct agent, autocfg_dispatcher);
	struct wifi_radio_element *re;

	trace("|%s:%d| Triggering AP-Autoconfig Search\n", __func__, __LINE__);

	agent_start_controller_conditionally(a);
	a->autocfg_interval = a->cntlr_select.probe_int;

	agent_check_bsta_connections(a);

	list_for_each_entry(re, &a->radiolist, list) {
		struct cmdu_buff *cmdu;

		if (!a->active_cntlr) {
			trace("|%s:%d| No active Controller present!\n", __func__, __LINE__);
			re->state = AUTOCFG_ACTIVE;
		}

		cmdu = agent_gen_ap_autoconfig_search(a, re, 0x02);
		if (!cmdu)
			continue;
		trace("|%s:%d| Sending Autoconfig Search for radio %s(%s)\n",
				__func__, __LINE__, re->name,
				(re->band == BAND_2 ? "2.4GHz" : "5GHz"));
		re->mid = agent_send_cmdu(a, cmdu);

		cmdu_free(cmdu);
	}

	dbg("|%s:%d| Scheduling next autoconfig search in %u seconds\n",
				__func__, __LINE__, a->autocfg_interval);
	timer_set(&a->autocfg_dispatcher, a->autocfg_interval * 1000);
}

static void agent_metric_report_timer_cb(atimer_t *t)
{
	struct agent_config *cfg =
		container_of(t, struct agent_config, metric_report_timer);
	struct agent *a = container_of(cfg, struct agent, cfg);
	struct cmdu_buff *cmdu;
	char *ifname = NULL;
	struct node *n;

	n = agent_find_node(a, a->almac);
	if (!n)
		goto refresh_interval;

	cmdu = prepare_ap_metrics_query((void *)a, ifname);
	if (!cmdu)
		goto refresh_interval;

	send_ap_metrics_response(a, cmdu, n);
	cmdu_free(cmdu);
refresh_interval:
	timer_set(&cfg->metric_report_timer,
			cfg->pcfg->report_interval * 1000);
}

/* remove passed interface from mapagent and wireless config, if present */
static int wifi_clean_iface(const char *ifname)
{
	char fmt[128] = {0};

	snprintf(fmt, sizeof(fmt), "clean_stale_ifaces %s", ifname);
	agent_exec_platform_scripts(fmt);

	dbg("|%s:%d| Cleaned interface %s\n", __func__, __LINE__, ifname);
	return 0;
}

/* remove all torndown interfaces on radio from config/memory */
int wifi_clean_torndown_ifaces(struct agent *a, struct wifi_radio_element *re)
{
	struct netif_ap *ap = NULL, *ap_tmp;

	list_for_each_entry_safe(ap, ap_tmp, &re->aplist, list) {
		if (!ap->torndown)
			continue;

		wifi_clean_iface(ap->ifname);
		clean_ap(ap->cfg);
		netif_free(ap);
	}

	return 0;
}

void clear_sta(struct sta *s)
{
	dbg("Delete STA " MACFMT "\n", MAC2STR(s->macaddr));
	timer_del(&s->sta_finalize_timer);
	timer_del(&s->sta_steer_timer);
	timer_del(&s->sta_bcn_req_timer);
	timer_del(&s->sta_timer);
	timer_del(&s->sta_steer_deauth_timer);
	list_del(&s->list);
	list_flush(&s->pref_nbrlist, struct pref_neighbor, list);
	list_flush(&s->sta_nbrlist, struct sta_neighbor, list);
	if (s->assoc_frame) {
		free(s->assoc_frame->frame);
		free(s->assoc_frame);
	}
#if (EASYMESH_VERSION >= 6)
	if (s->is_affiliated_sta)
		wifi_sta_mld_del_sta(s->vif->agent, s);
#endif
	free(s);
}

void clear_stalist(struct netif_ap *p)
{
	struct sta *s, *tmp;

	list_for_each_entry_safe(s, tmp, &p->stalist, list) {
		clear_sta(s);
	}
}

void clear_nbrlist(struct netif_ap *p)
{
	struct neighbor *n, *tmp;

	list_for_each_entry_safe(n, tmp, &p->nbrlist, list) {
		dbg("Delete NBR " MACFMT "\n", MAC2STR(n->nbr.bssid));
		list_del(&n->list);
		free(n);
	}
}

void clear_restrict_stalist(struct netif_ap *p)
{
	struct restrict_sta_entry *rs, *tmp;

	list_for_each_entry_safe(rs, tmp, &p->restrict_stalist, list) {
		list_del(&rs->list);
		free(rs);
	}
}

static void netif_reset(struct netif_ap *ap)
{
	/* clear stalist */
	clear_stalist(ap);

	/* clear nbrlist */
	clear_nbrlist(ap);

	/* clear restricted sta list */
	clear_restrict_stalist(ap);

	/* cancel timers */
	timer_del(&ap->nbr_timer);
	timer_del(&ap->bss_timer);
	timer_del(&ap->util_threshold_timer);
	timer_del(&ap->una_sta_meas_timer);

#if (EASYMESH_VERSION >= 6)
	if (ap->is_affiliated_ap)
		wifi_mld_del_ap(ap);
#endif
}

static void netif_free(struct netif_ap *ap)
{
	netif_reset(ap);
	list_del(&ap->list);
	free(ap);
}

void clear_aplist_in_radio(struct wifi_radio_element *re)
{
	struct netif_ap *ap = NULL, *tmp;

	list_for_each_entry_safe(ap, tmp, &re->aplist, list) {
		netif_free(ap);
	}
}

void clear_aplist(struct agent *a)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list)
		clear_aplist_in_radio(re);
}

void clear_bk(struct agent *a, struct netif_bk *bk)
{
#if (EASYMESH_VERSION >= 6)
	if (bk->is_affiliated_sta)
		wifi_sta_mld_del_bsta(a, bk);
#endif
	memset(bk, 0, sizeof(*bk));
}

void clear_bklist(struct agent *a)
{
	struct wifi_radio_element *re;

	list_for_each_entry(re, &a->radiolist, list) {
		if (!re->has_bsta)
			continue;
		clear_bk(a, &re->bk);
		re->has_bsta = false;
	}
}

void clear_wdslist(struct agent *a)
{
	struct netif_wds *p, *tmp;

	list_for_each_entry_safe(p, tmp,  &a->wdslist, list) {
		netif_free_wds(p);
	}
}

#if (EASYMESH_VERSION >= 6)

void agent_clean_affiliated_sta_mld(struct sta_mld *mld)
{
	int i;

	for (i = (mld->num_affiliated_sta - 1); i >= 0; i--) {
		struct sta *s = mld->affiliated_sta[i];

		if (!s)
			continue;

		s->mld = NULL;
		s->is_affiliated_sta = false;
		mld->affiliated_sta[i] = NULL;
	}

	mld->num_affiliated_sta = 0;
}


void agent_free_sta_mld(struct sta_mld *mld)
{
	agent_clean_affiliated_sta_mld(mld);

	list_del(&mld->list);
	free(mld);
}

/*
void clear_stamldlist(struct agent *a)
{
	struct netif_mld *mld = NULL;

	list_for_each_entry_safe(mld, tmp,  &a->apmldlist, list) {
		struct sta_mld *smld = NULL, *tmp;

		list_for_each_entry_safe(smld, tmp,  &mld->stamldlist, list)
			agent_free_sta_mld(mld);
	}
}
*/

void agent_clean_affiliated_bsta_mld(struct bsta_mld *mld)
{
	int i;

	for (i = 0; i < mld->num_affiliated_bsta; i++) {
		struct netif_bk *bk = mld->affiliated_bsta[i];

		if (!bk)
			continue;

		bk->mld = NULL;
		bk->is_affiliated_sta = false;
		mld->affiliated_bsta[i] = NULL;
	}

	mld->num_affiliated_bsta = 0;
}


void agent_clear_bsta_mld(struct agent *a)
{
	struct bsta_mld *mld = &a->bstamld;

	agent_clean_affiliated_bsta_mld(mld);

	a->has_bstamld = false;
}

void agent_clean_affiliated_ap_mld(struct netif_mld *mld)
{
	int i;

	for (i = (mld->num_affiliated_ap - 1); i >= 0; i--) {
		struct netif_ap *ap = mld->affiliated_ap[i];

		if (!ap)
			continue;

		ap->mld = NULL;
		ap->is_affiliated_ap = false;
		mld->affiliated_ap[i] = NULL;
	}

	mld->num_affiliated_ap = 0;
}


void agent_free_ap_mld(struct netif_mld *mld)
{
	agent_clean_affiliated_ap_mld(mld);

	list_del(&mld->list);
	free(mld);
}

/* TODO: fix mld free funcs... too many and they are all different */

static void clear_apmld(struct netif_mld *mld)
{
	struct sta_mld *smld = NULL, *stmp;

	list_for_each_entry_safe(smld, stmp,  &mld->stamldlist, list)
		agent_free_sta_mld(smld);

	agent_free_ap_mld(mld);
}

void clear_apmldlist(struct agent *a)
{
	struct netif_mld *mld = NULL, *tmp;

	list_for_each_entry_safe(mld, tmp,  &a->apmldlist, list) {
		clear_apmld(mld);
	}

	a->num_ap_mld = 0;
}
#endif

int agent_map_sub_cb(void *bus, void *priv, void *data)
{
	struct blob_attr *msg = (struct blob_attr *)data;
	char *str;


	str = blobmsg_format_json(msg, true);
	trace("Received notification '%s'\n", str);
	free(str);

	ieee1905_cmdu_event_handler(priv, msg);

	return 0;
}

int agent_map_del_cb(void *bus, void *priv, void *data)
{
	struct agent *a = (struct agent *)priv;
	uint32_t *obj = (uint32_t *)data;

	a->subscribed = false;
	fprintf(stdout, "Object 0x%x no longer present\n", *obj);

	return 0;
}

static int agent_subscribe_for_cmdus(struct agent *a)
{
	mapmodule_cmdu_mask_t cmdu_mask = {0};
	uint32_t map_id;
	int ret;


	map_prepare_cmdu_mask(cmdu_mask,
			CMDU_TYPE_TOPOLOGY_DISCOVERY,
			CMDU_TYPE_TOPOLOGY_NOTIFICATION,
			CMDU_TYPE_TOPOLOGY_QUERY,
			CMDU_TYPE_TOPOLOGY_RESPONSE,
			CMDU_TYPE_VENDOR_SPECIFIC,
			CMDU_TYPE_LINK_METRIC_QUERY,
			CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE,
			CMDU_TYPE_AP_AUTOCONFIGURATION_WSC,
			CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW,
			CMDU_TYPE_HIGHER_LAYER_QUERY,
			CMDU_1905_ACK,
			CMDU_BEACON_METRICS_QUERY,
			CMDU_BACKHAUL_STEER_REQUEST,
			CMDU_AP_METRICS_QUERY,
			CMDU_ASSOC_STA_LINK_METRICS_QUERY,
			CMDU_UNASSOC_STA_LINK_METRIC_QUERY,
			CMDU_CHANNEL_SCAN_REQUEST,
			CMDU_CHANNEL_SCAN_REPORT,
			CMDU_POLICY_CONFIG_REQ,
			CMDU_BACKHAUL_STA_CAPABILITY_QUERY,
			CMDU_BACKHAUL_STA_CAPABILITY_REPORT,
			CMDU_CHANNEL_PREFERENCE_QUERY,
			CMDU_CHANNEL_SELECTION_REQ,
#ifdef OPER_CHAN_CHANGE_RELAY_MCAST
			CMDU_OPERATING_CHANNEL_REPORT,
#endif
			CMDU_CLIENT_STEERING_REQUEST,
			CMDU_CLIENT_ASSOC_CONTROL_REQUEST,
			CMDU_AP_CAPABILITY_QUERY,
			CMDU_AP_CAPABILITY_REPORT,
			CMDU_HIGHER_LAYER_DATA,
			CMDU_CLIENT_CAPABILITY_QUERY,
			CMDU_CAC_REQUEST,
			CMDU_CAC_TERMINATION,
#if (EASYMESH_VERSION > 2)
			CMDU_BSS_CONFIG_RESPONSE,
			CMDU_DPP_CCE_INDICATION,
			CMDU_PROXIED_ENCAP_DPP,
			CMDU_DPP_BOOTSTRAPING_URI,
			CMDU_DIRECT_ENCAP_DPP,
			CMDU_AGENT_LIST,
			CMDU_SERVICE_PRIORITIZATION_REQUEST,
#endif
#if (EASYMESH_VERSION >= 6)
			CMDU_AP_MLD_CONFIG_REQUEST,
			CMDU_BSTA_MLD_CONFIG_REQUEST,
#endif
			-1);

	memcpy(a->cmdu_mask, cmdu_mask, sizeof(a->cmdu_mask));

	trace("<----------------------------------- %s\n", __func__);
	for (;;) {
		ret = ubus_lookup_id(a->ubus_ctx, map_plugin, &map_id);
		if (!ret)
			break;

		trace("ieee1905.map not up yet, sleeping for 2s!\n");
		sleep(2);
	}

	a->map_oid = map_id;

	ret = map_subscribe(a->ubus_ctx,
			    &a->map_oid,
			    "mapagent", &cmdu_mask, a,
			    agent_map_sub_cb,
			    agent_map_del_cb,
			    &a->subscriber);
	if (!ret) {
		a->subscribed = true;
	} else {
		trace("mapagent: Failed to 'register' with %s (err = %s)\n",
		      map_plugin, ubus_strerror(ret));
	}

	return ret;
}

int agent_init_defaults(struct agent *a)
{
	INIT_LIST_HEAD(&a->radiolist);
	INIT_LIST_HEAD(&a->framelist);
	INIT_LIST_HEAD(&a->nodelist);
	INIT_LIST_HEAD(&a->wdslist);
#if (EASYMESH_VERSION > 2)
	INIT_LIST_HEAD(&a->qos_rules);
#endif
#if (EASYMESH_VERSION >= 6)
	INIT_LIST_HEAD(&a->apmldlist);
#endif

	a->cntlr_select.local = CONTROLLER_SELECT_LOCAL;
	a->cntlr_select.auto_detect = CONTROLLER_SELECT_AUTODETECT;
	a->cntlr_select.probe_int = CONTROLLER_SELECT_PROBE_INT;
	a->cntlr_select.retry_int = CONTROLLER_SELECT_RETRY_INT;
#ifdef AGENT_DYNAMIC_CNTRL
	a->cntlr_select.autostart = CONTROLLER_SELECT_AUTOSTART;
#endif

	/* don't try to start cntlr right away */
	timestamp_update(&a->observed_time);
	timestamp_update(&a->last_unassoc);

	a->cntlr_miss_count = 0;
	a->active_cntlr = false;
	a->multiple_cntlr = false;
	a->autocfg_interval = CONTROLLER_SELECT_PROBE_INT;


	a->is_sta_steer_start = 0;
	a->sta_steerlist_count = 0;
	timer_init(&a->sta_steer_req_timer, agent_steering_opp_timeout);

	return 0;
}

static int agent_ackq_timeout_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *e)
{
	struct agent *a = container_of(q, struct agent, cmdu_ack_q);
	struct cmdu_buff *cmdu = (struct cmdu_buff *) e->cookie;
	int ret;
	uint16_t msgid;

	trace("%s: ---> cmdu = %04x to "MACFMT" \n", __func__,
		cmdu_get_type(cmdu), MAC2STR(cmdu->origin));

	if (e->resend_cnt-- > 0) {
		ret = ieee1905_ubus_send_cmdu(a->ubus_ctx, cmdu, &msgid, a->pvid);
		if (ret)
			err("%s fail to send cmdu\n", __func__);
		dbg("%s CMDU sent, msgid = %d\n", __func__, msgid);
		return CMDU_ACKQ_TMO_REARM;
	}

	return CMDU_ACKQ_TMO_DELETE;
}

static void agent_ackq_delete_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *e)
{
	struct cmdu_buff *cmdu = (struct cmdu_buff *) e->cookie;

	trace("%s: ---> cmdu = %04x to "MACFMT" \n", __func__,
		cmdu_get_type(cmdu), MAC2STR(cmdu->origin));

	cmdu_free(cmdu);
}

void run_agent(void)
{
	struct agent *w;
	struct ubus_context *ctx;
	sigset_t base_mask;

	sigemptyset(&base_mask);
	sigaddset(&base_mask, SIGHUP);

	sigprocmask(SIG_SETMASK, &base_mask, NULL);
	set_sighandler(SIGPIPE, SIG_IGN);

	w = calloc(1, sizeof(*w));
	if (!w)
		return;

	this_agent = w;
	dbg("Starting wifi_agent... (&agent = %p)\n", w);

	agent_init_defaults(w);

	cmdu_ackq_init(&w->cmdu_ack_q);
	w->cmdu_ack_q.timeout_cb = agent_ackq_timeout_cb;
	w->cmdu_ack_q.delete_cb = agent_ackq_delete_cb;

	uloop_init();
	ctx = ubus_connect(NULL);
	if (!ctx) {
		err("Failed to connect to ubus\n");
		free(w);
		return;
	}

	w->ubus_ctx = ctx;
	w->evh.cb = agent_event_handler;

#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
	timer_init(&w->dpp_legacy_timeout, dpp_legacy_timeout_cb);
	INIT_LIST_HEAD(&w->conf_resplist);
	if (dpp_init(&w->dpp, 0, NULL)) {
		fprintf(stderr, "Failed to init dpp context\n");
		goto out_ubus;
	}

	allmac_init_table(&w->peer_table);

	dpp_set_ctx_private_data(w->dpp, w);
	dpp_register_cb(w->dpp, dpp_frame_handler);
	w->own_peer = dpp_create_responder_instance(w->dpp);
	if (!w->own_peer) {
		fprintf(stderr, "Failed to create enrollee instance\n");
		goto out_ubus;
	}

	strcpy(w->own_peer->e_req.tech, "map");
	strcpy(w->own_peer->e_req.dpp_name, "iopsys_map_dpp");
	w->own_peer->e_req.name_len = strlen(w->own_peer->e_req.dpp_name);
	w->own_peer->e_req.netrole = DPP_NETROLE_MAPAGENT;

	dpp_print_bootstrap_info(w->own_peer->own_bi);

	int ret;
	uint8_t bcast[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

	ret = dpp_build_presence_announcement(w->dpp, bcast, &w->pa_frame, &w->pa_framelen);
	if (ret) {
		warn("Failed to build presence announcement frame!\n");
		goto out_ubus;
	}

	dump(w->pa_frame, w->pa_framelen, "Presence Annoucement");
	timer_init(&w->dpp_periodic_pa, dpp_send_chirp);
	timer_set(&w->dpp_periodic_pa, 10 * 1000);
#endif
#endif

	agent_config_init(w, &w->cfg);
#ifdef VENDOR_EXTENSION
	agent_load_extensions(w);
#endif

	w->ts.nl_main_sk = nl_init_main_sock(w);
	if (!w->ts.nl_main_sk)
		goto out_ubus;

	w->ts.nl_sk.cb = nl_event_uloop_cb;
	w->ts.nl_sk.fd = nl_socket_get_fd(w->ts.nl_main_sk);
	uloop_fd_add(&w->ts.nl_sk, ULOOP_READ);
#if (EASYMESH_VERSION > 2)
#if 0
	if (w->cfg.persist_uri) {
		/* attempt to read ec private key */
		dpp_get_keys(w);

		/* if no found, generate one and write to the file */
		if (!w->dpp_ctx.ec_key)
			dpp_gen_keypair(w);
	}
#endif
#endif

	ubus_add_uloop(ctx);

	/* used by some ts script handling */
	setenv("AL_BRIDGE", w->cfg.al_bridge, 1);

	agent_init_wsc_attributes(w);

	//memcpy(w->cntlr_almac, w->cfg.cntlr_almac, 6);

	//agent_config_dump(&w->cfg);

	get_registered_steer_rules();	/* TODO: return rule list and improve */

	//agent_switch_according_to_pref(w); /*switch to the channel according to the prefrence*/

	ubus_register_event_handler(ctx, &w->evh, "network.device");
	ubus_register_event_handler(ctx, &w->evh, "wifi.*");
	ubus_register_event_handler(ctx, &w->evh, "wps_credentials");
	ubus_register_event_handler(ctx, &w->evh, "ubus.object.*");

	timer_init(&w->autocfg_dispatcher, agent_dispatch_autoconfig);
	timer_init(&w->reload_scheduler, agent_reload_cb);
	timer_init(&w->cntlr_scheduler, agent_enable_local_cntlr);
	timer_init(&w->rcpi_threshold_timer, agent_rcpi_thresold_timer_cb);
	timer_init(&w->onboarding_scheduler, agent_trigger_bsta_sync);
	timer_init(&w->bh_lost_timer, dynbh_bh_lost_cb);
	timer_init(&w->bh_reconf_timer, dynbh_bh_reconf_cb);
#ifdef AGENT_ISLAND_PREVENTION
	timer_init(&w->sta_disconnect_timer, agent_sta_disconnnect_cb);
	timer_init(&w->fh_disable_timer, agent_fh_disable_cb);
#endif /* AGENT_ISLAND_PREVENTION */
	timer_init(&w->disable_unconnected_bstas_scheduler, dynbh_disable_unconnected_bsta_cb);
	timer_init(&w->dynbh_scheduler, dynbh_upgrade_backhaul_cb);
	timer_init(&w->init_ifaces_scheduler, agent_init_ifaces_cb);
	timer_init(&w->enable_fhs_scheduler, agent_enable_fhs_cb);
	timer_init(&w->radio_stats_scheduler, agent_radio_stats_cb);
	timer_init(&w->boot_scan_scheduler, agent_boot_scan_cb);
	timer_init(&w->agent_periodic_run, agent_periodic_run);

	timer_init(&w->cfg.metric_report_timer, agent_metric_report_timer_cb);

	get_ethportslist(&w->num_ethports, w->ethports);
	agent_init_interfaces(w);
	agent_init_ifaces_assoc_mode(w);
	/* switch to the channel according to the prefrence */
	//agent_switch_according_to_pref(w);

	agent_get_system_info(w);

	/* w->cfg.enabled */
	agent_publish_object(w, MAPAGENT_OBJECT);
	agent_publish_dbg_object(w, MAPAGENT_DBG_OBJECT);

	//run_agent(w);

	timer_set(&w->rcpi_threshold_timer, RCPI_THRESHOLD_TIMER);
	timer_set(&w->autocfg_dispatcher, 0 * 1000);
	timer_set(&w->radio_stats_scheduler, RADIO_STATS_TIMER);
	w->boot_scan_tries = 0;
	timer_set(&w->boot_scan_scheduler, BOOT_UP_SCAN_TIME * 1000);
	timer_set(&w->agent_periodic_run, 1 * 1000);

	agent_subscribe_for_cmdus(w);

	nl_wds_start_all(w);

	timestamp_update(&w->start_t);

	uloop_run();

/* out_and_exit: */
	map_unsubscribe(w->ubus_ctx, w->subscriber);
	agent_free_wifi_assoc_frames(w);
	agent_free_radios(w);
	agent_remove_object(w, MAPAGENT_OBJECT);
	agent_remove_dbg_object(w, MAPAGENT_DBG_OBJECT);
	agent_config_clean(&w->cfg);
	agent_free_cntlr_sync(w);
	cmdu_ackq_free(&w->cmdu_ack_q);
	clear_aplist(w);
	clear_bklist(w);
	clear_wdslist(w);
#if (EASYMESH_VERSION > 2)
#ifdef USE_LIBDPP
	allmac_clean_table(&w->peer_table);
	dpp_free_peer(w->dpp, w->own_peer);
	free(w->pa_frame);
	free(w->dpp);
#endif
	qos_rule_free_all(&w->qos_rules);
#endif
#if (EASYMESH_VERSION >= 6)
	agent_clear_bsta_mld(w);
	clear_apmldlist(w);
#endif
	ubus_unregister_event_handler(ctx, &w->evh);
#ifdef VENDOR_EXTENSION
	agent_unload_extensions(w);
#endif
	uloop_done();
	nl_free_main_sock(w->ts.nl_main_sk);
	agent_free_nodes(w);
out_ubus:
	ubus_free(ctx);
	free(w);
}

struct wsc_data *agent_free_wsc_data(struct wsc_data *wsc)
{
	if (wsc->m1_frame)
		free(wsc->m1_frame);
	if (wsc->key) {
		free(wsc->key->pub);
		free(wsc->key->key);
		free(wsc->key);
	}

	return NULL;
}

static void agent_free_nodes(struct agent *a)
{
	struct node *n, *tmp;

	list_for_each_entry_safe(n, tmp, &a->nodelist, list) {
		list_del(&n->list);
		free(n);
	}
}

static void agent_free_wifi_assoc_frames(struct agent *a)
{
	struct wifi_assoc_frame *n = NULL, *tmp;

	list_for_each_entry_safe(n, tmp, &a->framelist, list) {
		list_del(&n->list);
		free(n->frame);
		free(n);
	}
}

static void agent_reset_radio(struct wifi_radio_element *re)
{
	free_scanlist(re);
	clear_aplist_in_radio(re);
	free(re->bsslist);
	re->bsslist = NULL;
	re->num_bss = 0;
	free(re->unassoc_stalist);
	re->unassoc_stalist = NULL;
	agent_free_wsc_data(&re->autconfig);
	memset(&re->autconfig, 0, sizeof(re->autconfig));
	timer_del(&re->preference_report_timer);
}

void agent_free_radios(struct agent *a)
{
	struct wifi_radio_element *re, *tmp;

	trace("%s: a->num_radios to free = %d\n", __func__, a->num_radios);
	list_for_each_entry_safe(re, tmp, &a->radiolist, list) {
		agent_reset_radio(re);
		list_del(&re->list);
		free(re);
	}
	a->num_radios = 0;
}

void agent_free_cntlr_sync(struct agent *a)
{
	if (a->sync_config_req)
		free(a->sync_config_req);

	if (a->privkey) {
		struct wsc_key *k = (struct wsc_key *) a->privkey;

		free(k->pub);
		free(k->key);
		free(k);
	}
}

/* following implements 'map.agent' ubus methods */
int agent_assoc_control_sta(uint8_t *bssid, uint8_t mode, int tmo,
		int num_sta, uint8_t stalist[][6])
{
	trace("%s: --->\n", __func__);

	struct agent *a = this_agent;
	struct netif_ap *ap;
	int ret = 0;

	if (!a) {
		dbg("this wifiagent is NULL\n");
		return -1;
	}

	if (num_sta == 0 || !stalist)
		return -1;

	ap = agent_get_ap(a, bssid);

	if (!ap || !ap->cfg)
		return -1;

	/* Start / stop validity timer or update config */
	if (mode == ASSOC_CTRL_UNBLOCK)
		ret = assoc_ctrl_unblock_sta(a, ap, num_sta, stalist);
	else
		ret = assoc_ctrl_block_sta(a, ap, tmo, mode,
					num_sta, stalist);

	return ret;
}

int wifiagent_steer_sta(struct ubus_context *ctx,
			char *ifname,
			unsigned char *sta,
			int prefcnt, struct pref_neighbor *pref, int optime)
{
	struct agent *a = this_agent;
	struct sta *s;
	struct nbr *wifi_nbr_list = NULL;
	int i;
	int ret = 0;

	UNUSED(ctx);

	if (!a) {
		dbg("this wifiagent is NULL\n");
		return -1;
	}

	s = agent_get_sta(a, sta);
	if (!s) {
		dbg("STA not found!\n");
		return -1;
	}

	if (optime) {
		s->steer_policy |= STA_STEER_OPPORTUNITY;
		if (!timer_pending(&s->sta_steer_timer)) {
			rebuild_cntlr_preflist(a, s, prefcnt, pref);
			/* (Re)start steer opportunity timer.
			 * In the opportunity time window constantly check
			 * every second for this sta's steering opportinity.
			 */
			s->steer_opportunity_tmo = optime * 1000;
			clock_gettime(CLOCK_MONOTONIC, &s->steer_opportunity);
			timer_set(&s->sta_steer_timer, 1000);
			return 0;
		}
	}

	/* steer mandate */
	s->steer_policy &= ~STA_STEER_OPPORTUNITY;
	if (/* !(s->caps & STA_CAP_11V_BSS_TRANS) || */   /* TODO */
			!(s->steer_policy & STEER_BTM)) {
		return steer_sta_legacy(s);
	}

	/* Translate internal 'pref_neighbor' to wifi's 'nbr' struct */
	wifi_nbr_list = calloc(prefcnt, sizeof(struct nbr));
	if (WARN_ON(!wifi_nbr_list))
		return -ENOMEM;

	for (i = 0; i < prefcnt; i++) {
		agent_pref_neighbor_to_nbr(&pref[i], &wifi_nbr_list[i]);
	}

	ret = agent_req_btm_steer(a, ifname, s->macaddr, prefcnt, wifi_nbr_list, 0);

	if (ret)
		dbg("Failed to send BTM request to " MACFMT "\n",
		    MAC2STR(s->macaddr));

	free(wifi_nbr_list);

	return ret;
}

int wifiagent_get_status(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	struct agent *agent = this_agent;
	struct wifi_radio_element *re;
	struct netif_ap *p = NULL;
	struct neighbor *n;
	struct blob_buf bb;
	void *a, *b, *c, *d, *t;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	a = blobmsg_open_array(&bb, "radios");
	list_for_each_entry(re, &agent->radiolist, list) {
		t = blobmsg_open_table(&bb, "");
		blobmsg_add_string(&bb, "name", re->name);
		blobmsg_add_macaddr(&bb, "macaddr", re->macaddr);
		blobmsg_add_u32(&bb, "channel", re->current_channel);
		blobmsg_add_u32(&bb, "bandwidth", re->current_bandwidth);
		blobmsg_add_u32(&bb, "opclass", re->current_opclass);
		b = blobmsg_open_table(&bb, "autoconfig");
		blobmsg_add_string(&bb, "state", (re->state ? "ACTIVE":"HEARTBEAT"));
		c = blobmsg_open_table(&bb, "message_identifier");
		blobmsg_add_u32(&bb, "search", re->mid);
		blobmsg_add_u32(&bb, "wsc", re->wsc_mid);
		blobmsg_add_u32(&bb, "renew", re->renew_mid);
		blobmsg_close_table(&bb, c);
		blobmsg_close_table(&bb, b);

		d = blobmsg_open_array(&bb, "fronthaul");
		list_for_each_entry(p, &re->aplist, list) {
			struct sta *s;
			void *tt;

			if (!p->cfg)
				continue;

			tt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "name", p->ifname);
	#if (EASYMESH_VERSION >= 6)
			blobmsg_add_u8(&bb, "is_affiliated_ap", p->is_affiliated_ap);
			if (p->is_affiliated_ap && p->mld) {
				blobmsg_add_string(&bb, "mld_ifname", p->mld->ifname);
				blobmsg_add_u32(&bb, "mlo_linkid", p->linkid);
			}
	#endif
			if (p->band == BAND_2)
				blobmsg_add_string(&bb, "band", "2.4GHz");
			else if (p->band == BAND_5)
				blobmsg_add_string(&bb, "band", "5GHz");
			else if (p->band == BAND_6)
				blobmsg_add_string(&bb, "band", "6GHz");
			else
				blobmsg_add_string(&bb, "band", "Unknown");

			blobmsg_add_u8(&bb, "enabled", p->enabled);
			blobmsg_add_macaddr(&bb, "bssid", p->bssid);
			blobmsg_add_string(&bb, "ssid", p->ssid);
			blobmsg_add_u32(&bb, "channel", p->channel);
			blobmsg_add_u32(&bb, "load", p->bssload);
			blobmsg_add_u32(&bb, "num_sta", p->num_sta);

			b = blobmsg_open_array(&bb, "neighbor");
			list_for_each_entry(n, &p->nbrlist, list) {
				void *ttt;

				ttt = blobmsg_open_table(&bb, "");
				blobmsg_add_macaddr(&bb, "bssid", n->nbr.bssid);
				blobmsg_add_u32(&bb, "channel", n->nbr.channel);
				blobmsg_close_table(&bb, ttt);
			}
			blobmsg_close_array(&bb, b);

			b = blobmsg_open_array(&bb, "stations");
			list_for_each_entry(s, &p->stalist, list) {
				void *aa, *ttt, *tttt;
				struct pref_neighbor *pn;

				ttt = blobmsg_open_table(&bb, "");
				blobmsg_add_macaddr(&bb, "macaddr", s->macaddr);
	#if (EASYMESH_VERSION >= 6)
				blobmsg_add_u8(&bb, "is_affiliated_sta", s->is_affiliated_sta);
				if (s->is_affiliated_sta && s->mld) {
					blobmsg_add_macaddr(&bb, "mld_macaddr", s->mld->macaddr);
					blobmsg_add_macaddr(&bb, "mld_bssid", s->mld->bssid);
					blobmsg_add_u32(&bb, "mlo_linkid", s->mlo_link_id);
				}
	#endif
				/* blobmsg_add_u32(&bb, "conn_time", s->connected_ms / 1000); */
				blobmsg_add_u32(&bb, "rssi", s->rssi_avg[0]);
				blobmsg_add_u32(&bb, "bcnreport_capable",
						s->supports_bcnreport);
				blobmsg_add_u32(&bb, "btmreq_steers",
						s->steer_btm_cnt);
				blobmsg_add_u32(&bb, "legacy_steers",
						s->legacy_steered);
				aa = blobmsg_open_array(&bb, "neighbor");
				list_for_each_entry(pn, &s->pref_nbrlist, list) {
					char pn_bssidstr[18] = {0};

					hwaddr_ntoa(pn->bssid, pn_bssidstr);
					tttt = blobmsg_open_table(&bb, "");
					blobmsg_add_string(&bb, "bssid", pn_bssidstr);
					blobmsg_add_u32(&bb, "rssi", pn->rssi);
					blobmsg_close_table(&bb, tttt);
				}
				blobmsg_close_array(&bb, aa);
				blobmsg_close_table(&bb, ttt);
			}
			blobmsg_close_array(&bb, b);
			blobmsg_close_table(&bb, tt);
		}
		blobmsg_close_array(&bb, d);
		d = blobmsg_open_table(&bb, "backhaul");
		if (re->has_bsta && re->bk.cfg) {
			struct netif_bk *bk = &re->bk;
			void *ttt;

			ttt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "name", bk->ifname);
			blobmsg_add_u8(&bb, "enabled", bk->enabled);
			blobmsg_add_u8(&bb, "connected", bk->connected);
			blobmsg_add_macaddr(&bb, "macaddr", bk->macaddr);
			blobmsg_add_macaddr(&bb, "bssid", bk->bssid);
			blobmsg_add_string(&bb, "ssid", bk->ssid);
			blobmsg_add_u32(&bb, "channel", bk->channel);

			b = blobmsg_open_table(&bb, "cfg");
			blobmsg_add_string(&bb, "name", bk->cfg->name);
			blobmsg_add_u8(&bb, "enabled", bk->cfg->enabled);
			blobmsg_add_u32(&bb, "band", bk->cfg->band);
			blobmsg_add_string(&bb, "device", bk->cfg->device);
			blobmsg_add_string(&bb, "ssid", bk->cfg->ssid);
			blobmsg_add_string(&bb, "key", bk->cfg->key);
			blobmsg_add_string(&bb, "encryption", bk->cfg->encryption);
			blobmsg_add_u8(&bb, "onboarded", bk->cfg->onboarded);
			blobmsg_add_macaddr(&bb, "bssid", bk->cfg->bssid);
			blobmsg_add_u16(&bb, "priority", (uint16_t) bk->cfg->priority);
			blobmsg_close_table(&bb, b);
			blobmsg_close_table(&bb, ttt);
		}
		blobmsg_close_table(&bb, d);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

#if (EASYMESH_VERSION >= 6)
	struct netif_mld *mld;

	a = blobmsg_open_array(&bb, "ap_mld");
	list_for_each_entry(mld, &agent->apmldlist, list) {
		void *aa, *aaa;
		int i;

		t = blobmsg_open_table(&bb, "");

		blobmsg_add_string(&bb, "ifname", mld->ifname);
		blobmsg_add_macaddr(&bb, "macaddr", mld->macaddr);
		blobmsg_add_string(&bb, "ssid", mld->ssid);

		aa = blobmsg_open_array(&bb, "affiliated_ap");
		for (i = 0; i < mld->num_affiliated_ap; i++) {
			void *tt;

			p = mld->affiliated_ap[i];

			if (!p)
				break;

			tt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "ifname", p->ifname);
			blobmsg_close_table(&bb, tt);
		}
		blobmsg_close_array(&bb, aa);

		struct sta_mld *smld;

		aaa = blobmsg_open_array(&bb, "mld_stations");
		list_for_each_entry(smld, &mld->stamldlist, list) {
			void *ttt, *tttt;

			ttt = blobmsg_open_table(&bb, "");
			blobmsg_add_macaddr(&bb, "mld_macaddr", smld->macaddr);
			blobmsg_add_macaddr(&bb, "mld_bssid", smld->bssid);
			blobmsg_add_u32(&bb, "num_affiliated_sta", smld->num_affiliated_sta);
			tttt = blobmsg_open_array(&bb, "stations");
			for (i = 0; i < smld->num_affiliated_sta; i++) {
				void *ttttt;
				struct sta *s;
				ttttt = blobmsg_open_table(&bb, "");
				s = smld->affiliated_sta[i];
				blobmsg_add_macaddr(&bb, "macaddr", s->macaddr);
				blobmsg_add_macaddr(&bb, "bssid", s->bssid);
				blobmsg_add_u32(&bb, "linkid", s->mlo_link_id);
				blobmsg_close_table(&bb, ttttt);
			}
			blobmsg_close_array(&bb, tttt);
			blobmsg_close_table(&bb, ttt);
		}
		blobmsg_close_array(&bb, aaa);

		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	a = blobmsg_open_table(&bb, "bsta_mld");
	if (agent->has_bstamld) {
		void *aa;
		int i;

		t = blobmsg_open_table(&bb, "");

		blobmsg_add_string(&bb, "ifname", agent->bstamld.bk.ifname);
		blobmsg_add_macaddr(&bb, "macaddr", agent->bstamld.bk.macaddr);
		blobmsg_add_macaddr(&bb, "bssid", agent->bstamld.bk.bssid);
		blobmsg_add_u32(&bb, "num_affiliated_bsta", agent->bstamld.num_affiliated_bsta);

		aa = blobmsg_open_array(&bb, "affiliated_sta");
		for (i = 0; i < agent->bstamld.num_affiliated_bsta; i++) {
			struct netif_bk *bk;
			void *tt;

			bk = agent->bstamld.affiliated_bsta[i];

			if (!bk)
				break;

			tt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "ifname", bk->ifname);
			blobmsg_add_macaddr(&bb, "macaddr", bk->macaddr);
			blobmsg_add_macaddr(&bb, "bssid", bk->bssid);
			blobmsg_close_table(&bb, tt);
		}
		blobmsg_close_array(&bb, aa);

		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);
#endif
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int wifiagent_get_nodes(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	struct agent *agent = this_agent;
	struct node *n;
	struct blob_buf bb;
	void *a;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "num_nodes", agent->num_nodes);

	a = blobmsg_open_array(&bb, "nodes");
	list_for_each_entry(n, &agent->nodelist, list) {
		char aladdrstr[18] = {0};
		void *t;

		hwaddr_ntoa(n->alid, aladdrstr);
		t = blobmsg_open_table(&bb, "");

		blobmsg_add_string(&bb, "almac", aladdrstr);
		blobmsg_add_u16(&bb, "profile", n->map_profile);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int wifiagent_get_info(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	char almac_str[18] = {};
	struct blob_buf bb;
	struct agent *a = this_agent;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "active_cntlr", a->active_cntlr);
	hwaddr_ntoa(a->cntlr_almac, almac_str);
	blobmsg_add_string(&bb, "cntlr_almac", almac_str);
	blobmsg_add_u32(&bb, "local_cntlr", is_local_cntlr_running());
	blobmsg_add_u32(&bb, "multiple_cntlr_found", a->multiple_cntlr);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int wifiagent_get_bk_info(struct ubus_context *ctx,
		struct ubus_request_data *req)
{
	struct blob_buf bkfile = { 0 }, bb = { 0 };
	struct netif_bk *bk = NULL;
	struct agent *a = this_agent;
	struct blob_attr *tb[5];
	static const struct blobmsg_policy bk_attr[5] = {
		[0] = { .name = "type", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		[3] = { .name = "backhaul_device_id", .type = BLOBMSG_TYPE_STRING },
		[4] = { .name = "backhaul_macaddr", .type = BLOBMSG_TYPE_STRING },
	};
	int ret;
	void *t, *tt;
	uint32_t wifi_bk_uptime = a->wifi_backhaul_time;
	struct wifi_radio_element *re;

	blob_buf_init(&bkfile, 0);

	if (!blobmsg_add_json_from_file(&bkfile, MAP_UPLINK_PATH)) {
		dbg("%s: Failed to parse %s\n", __func__, MAP_UPLINK_PATH);
		goto out;
	}

	ret = blobmsg_parse(bk_attr, 5, tb, blob_data(bkfile.head), blob_len(bkfile.head));
	if (ret)
		goto out;

	blob_buf_init(&bb, 0);
	t = blobmsg_open_table(&bb, "active_backhaul");
	if (tb[0])
		blobmsg_add_blob(&bb, tb[0]);
	if (tb[1])
		blobmsg_add_blob(&bb, tb[1]);
	if (tb[2])
		blobmsg_add_blob(&bb, tb[2]);
	if (tb[3])
		blobmsg_add_blob(&bb, tb[3]);
	if (tb[4])
		blobmsg_add_blob(&bb, tb[4]);
	blobmsg_close_table(&bb, t);

	bk = dynbh_get_best_connected_bsta(a);
	if (bk)
		wifi_bk_uptime += timestamp_elapsed_sec(&bk->connect_t);

	blobmsg_add_u32(&bb, "agent_uptime", timestamp_elapsed_sec(&a->start_t));
	blobmsg_add_u32(&bb, "wifi_backhaul_uptime", wifi_bk_uptime);
	tt = blobmsg_open_array(&bb, "backhauls");

	list_for_each_entry(re, &a->radiolist, list) {
		void *ttt;
		uint32_t tot_uptime = bk->total_connect_time;

		if (!re->has_bsta)
			continue;

		bk = &re->bk;

		if (!bk->cfg)
			continue;

		ttt = blobmsg_open_table(&bb, "");
		blobmsg_add_string(&bb, "name", bk->ifname);
		blobmsg_add_u8(&bb, "connected", bk->connected);
		blobmsg_add_macaddr(&bb, "macaddr", bk->macaddr);
		blobmsg_add_macaddr(&bb, "bssid", bk->bssid);
		if (bk->connected)
			tot_uptime += timestamp_elapsed_sec(&bk->connect_t);
		blobmsg_add_u32(&bb, "total_uptime", tot_uptime);
		blobmsg_close_table(&bb, ttt);
	}

	blobmsg_close_array(&bb, tt);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bkfile);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
out:
	blob_buf_free(&bkfile);
	return UBUS_STATUS_UNKNOWN_ERROR;
}

int agent_switch_according_to_pref(struct agent *a)
{
	uint32_t channel = 0;
	uint32_t opclass = 0;
	int ret = 0;
	struct wifi_radio_element *re;

	trace("agent: %s: --->\n", __func__);

	list_for_each_entry(re, &a->radiolist, list) {
		ret = agent_get_highest_preference(re, re->current_opclass, &channel,
			&opclass);

		/* The operating class channel preference has been set
		 * now we want to switch the channel to the max preference */
		trace("agent|%s: %d|: opclass is %d channel is %d--->\n",
				__func__, __LINE__, opclass, channel);

		if ((re->current_opclass == opclass) && (re->current_channel == channel))
			continue;

		if (channel != 0 && opclass != 0) {
			ret = agent_channel_switch(a, re->macaddr, channel, opclass);
			if (ret == -1) {
				/* Here we need to set the default preference for all the
				 * channels in that operating class also set the response
				 * code as rejected*/
				agent_set_channel_preference_to_default(re);
			}
		}
	}
	return 0;
}

void agent_set_post_scan_action_pref(struct agent *a, struct wifi_radio_element *re, bool opclass_preferences)
{
	re->post_scan_action.opclass_preferences = opclass_preferences;
}

/* Allocates channel scan report CMDU and puts Timestamp TLV into it */
struct cmdu_buff *agent_prepare_scan_cmdu(struct agent *a, struct tlv *tsp)
{
	trace("%s --->\n", __func__);

	struct cmdu_buff *cmdu_data;

	/* Allocate new fragment CMDU */
	cmdu_data = cmdu_alloc_frame(CH_SCAN_RESP_CMDU_MAX_LEN);
	if (!cmdu_data) {
		dbg("%s: -ENOMEM\n", __func__);
		return NULL;
	}
	cmdu_set_type(cmdu_data, CMDU_CHANNEL_SCAN_REPORT);
	memcpy(cmdu_data->origin, a->cntlr_almac, 6);

	/* Put timestamp TLV to the fragment CMDU */
	if (cmdu_copy_tlvs(cmdu_data, &tsp, 1)) {
		dbg("%s: error: cmdu_put_tlv()\n", __func__);
		return NULL;
	}
	dbg("|%s:%d| added MAP_TLV_TIMESTAMP\n", __func__, __LINE__);

	return cmdu_data;
}

/* Splits Channel Scan Report TLVs into fragment CMDUs.
 * Sends fragmet CMDUs to controller.
 */
static int agent_send_ch_scan_rsp_frag(struct agent *a, struct cmdu_buff *cmdu)
{
	trace("agent: %s: --->\n", __func__);

	struct tlv_policy d_policy_scan[] = {
		[0] = {
			.type = MAP_TLV_CHANNEL_SCAN_RES,
			.present = TLV_PRESENT_MORE,
			.minlen = 9
		}
	};
	/* Timestamp TLV */
	struct tlv *tv[CHAN_SCAN_REPORT_NUM_OF_TLV_TYPES][TLV_MAXNUM] = {0};
	struct tlv *tv_scan[256] = {0};	/* Channel Scan Results TLVs */
	int i;
	int num_tlv = 256, num_tlv_copied = 0;
	int ret = 0;
	struct cmdu_buff *frag_cmdu = NULL;

	if (cmdu->datalen < CH_SCAN_RESP_CMDU_MAX_LEN)
		/* No fragmentation required, just send the CMDU as is */
		return agent_send_cmdu(a, cmdu);

	/* If the number of neighbors detected during a channel scan would
	 * mean that the channel scan report message would not fit within
	 * one 1905 CMDU, the Multi-AP Agent shall split the channel scan
	 * report across multiple Channel Scan Result TLVs by splitting
	 * the information related to sets of neighbor BSSs into separate
	 * Channel Scan Result TLVs and setting the NumberofNeighbors field
	 * to the number of neighbors contained in the corresponding TLV.
	 */

	/* Note: assuming neighbors are already evenly split between TLVs */

	ret = map_cmdu_validate_parse(cmdu, tv, ARRAY_SIZE(tv), a->cfg.map_profile);
	if (!ret) {
		dbg("%s: map_cmdu_validate_parse failed, err = (%d) '%s'\n", __func__,
		    map_error, map_strerror(map_error));
		return -1;
	}

	if (!tv[CHAN_SCAN_REPORT_TIMESTAMP_IDX][0]) {
		dbg("%s: Missing TIMESTAMP_TLV!\n", __func__);
		return -1;
	}

	ret = cmdu_parse_tlv_single(cmdu, tv_scan, d_policy_scan, &num_tlv);
	if (ret) {
		dbg("%s: cmdu_parse_tlv_single failed,  err = (%d) '%s'\n",
		    __func__, map_error, map_strerror(map_error));
		return -1;
	}

	if (!tv_scan[0]) {
		dbg("%s: Missing CHANNEL SCAN RESULT_TLV!\n", __func__);
		return -1;
	}

	tv[CHAN_SCAN_REPORT_TIMESTAMP_IDX][0]->len =
		tlv_length(tv[CHAN_SCAN_REPORT_TIMESTAMP_IDX][0]); // FIXME
	frag_cmdu = agent_prepare_scan_cmdu(a, tv[CHAN_SCAN_REPORT_TIMESTAMP_IDX][0]);
	if (WARN_ON(!frag_cmdu))
		return -1;

	for (i = 0; i < num_tlv; i++) {
		uint16_t tlv_len = tlv_length(tv_scan[i]);

		if (tlv_total_length(tv_scan[i]) > frag_cmdu->end - frag_cmdu->tail) {
			/* No space left for current TLV in CMDU buffer */
			cmdu_put_eom(frag_cmdu);
			agent_send_cmdu(a, frag_cmdu);
			cmdu_free(frag_cmdu);

			num_tlv_copied = 0;

			/* Create next fragment and put next TLVs into it */
			frag_cmdu = agent_prepare_scan_cmdu(a, tv[CHAN_SCAN_REPORT_TIMESTAMP_IDX][0]);
		}

		/* Add TLV to CMDU & continue */
		tv_scan[i]->len = tlv_len; // FIXME: assign len smwhr else
		if (cmdu_copy_tlvs(frag_cmdu, &tv_scan[i], 1)) {
			dbg("%s:%d copy TLVs failed, end = %p, tail = %p, len = %d\n",
			    __func__, __LINE__,
			    frag_cmdu->end, frag_cmdu->tail, tv_scan[i]->len);
			return -1;
		}
		num_tlv_copied++;
	}

	if (num_tlv_copied) { /* avoid sending timestamp alone */
		cmdu_put_eom(frag_cmdu);
		agent_send_cmdu(a, frag_cmdu);
	}

	cmdu_free(frag_cmdu);

	return 0;
}

int agent_send_ch_scan_response(struct agent *a, struct wifi_radio_element *re,
		struct wifi_scan_request_radio *req)
{
	struct cmdu_buff *cmdu_data = NULL;
	int ret = 0;

	dbg("%s: called.\n", __func__);

	/* Generate the response cmdu */
	if (WARN_ON(!re))
		return -1;

	if (!req)
		/* Independent channel scan response */
		cmdu_data = agent_gen_independent_ch_scan_response(a, re);
	else
		cmdu_data = agent_gen_ch_scan_response_radio(a,
						re, req, req->status);

	if (!cmdu_data)
		return -1;

	/* Status will be SUCCESS once results available */
	if (req && req->status == CH_SCAN_STATUS_SCAN_NOT_COMPLETED) {
		/* Replace old request data with the new one */
		re->scan_req = *req;
		/* Reset the timer of available time (5min) */
		timer_set(&re->available_scan_timer, 300000);
	}

	/* Send the response cmdu */
	ret = agent_send_ch_scan_rsp_frag(a, cmdu_data);

	cmdu_free(cmdu_data);

	return ret;
}

bool agent_ch_scan_succesful(struct agent *a)
{
	if (a->scan_status_code == CH_SCAN_STATUS_SUCCESS)
		return true;

	return false;
}
